/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** \mainpage

    \section part_ug_section User's guide

    - \ref part_section
    - \ref nuclei_section
    - \ref todo
    - \ref bug
    - \ref partref_section

    \htmlonly
    <a href="../latex/refman.pdf">PDF documentation</a>
    \endhtmlonly

    \hline
    \section part_section Particles

    These classes in the library \c o2scl_part calculate the
    thermodynamic properties of interacting and non-interacting
    quantum and classical particles.

    The class \ref part is the basic structure for a particle:
    - m - mass
    - g - degeneracy factor (i.e. \f$2j+1\f$)
    - n - number density
    - ns - scalar number density 
    - ed - energy density
    - pr - pressure
    - en - entropy
    - ms - effective mass
    - nu - effective chemical potential
    - bool inc_rest_mass - True if the rest mass is included
    - bool non_interacting - False if the particle includes interactions
    - std::string name - the name

    The data members part::ms and part::nu allow one to specify
    modifications to the mass and the chemical potential due to
    interactions. This allows one to calculate the properties of
    particle due to interactions so long as the basic form of the
    free-particle dispersion relation is unchanged, i.e.
    \f[
    \sqrt{k^2+m^2} - \mu \rightarrow \sqrt{k^2+m^{* 2}} - \nu 
    \f]
    Typically, if the particle is
    non-interacting, then \c mu and \c m are copied to \c nu and
    \c ms, computations are performed with \c nu and \c ms, and
    then, if necessary, the result for \c nu is copied back to 
    \c mu.

    If part::inc_rest_mass is \c true (as is the default), then all
    functions include the rest mass energy density in the energy
    density, the "mu" functions expect that the rest mass is included
    in part::mu or part::nu as input and the "density" functions
    output part::mu or part::nu including the rest mass.
    
    When inc_rest_mass is true, antiparticles are implemented by
    choosing the antiparticle chemical potential to be \f$ - \mu \f$,
    and when inc_rest_mass is false, antiparticles are implemented by
    choosing the chemical potential of the antiparticles to be \f$ -
    \mu - 2 m \f$.

    The thermodynamic identity used to compute the pressure for
    interacting particles is
    \f[
    P = -\varepsilon + s T + \nu n
    \f]
    where \c nu is used. This way, the particle class doesn't need to
    know about the structure of the interactions to ensure that the
    thermodynamic identity is satisfied. Note that in the \c o2scl_eos
    library, where in the equations of state the normal thermodynamic
    identity is used
    \f[
    P = -\varepsilon + s T + \mu n
    \f]
    Frequently, the interactions which create an effective chemical
    potential which is different than \c mu thus create extra terms in
    the pressure and the energy density for the given equation of
    state.

    At zero temperature, fermions and bosons can be treated exactly in
    the classes \ref fermion and \ref boson. \ref quark is a
    descendant of the \ref fermion class which contains extra data
    members for the quark condensate and the contribution to the bag
    constant. \ref classical is a descendant of both fermion and boson
    and calculates everything in the classical limit.

    At finite temperature, there are a couple different approaches.  The
    approximation scheme from \ref Johns96 is used in eff_boson,
    eff_fermion, and eff_quark. An exact method is used in rel_boson
    and rel_fermion, but these are necessarily quite a bit slower.

    The class nonrel_fermion can be used to assume a non-relativistic
    dispersion relation for fermions and includes zero-temperature
    methods and an exact method for finite temperatures.

    \hline

    <b>Units:</b>

    Factors of \f$ \hbar, c \f$ and \f$ k_B \f$ have been removed
    everywhere, so that mass, energy, and temperature all have the
    same units.  Number densities have units of mass cubed (or energy
    cubed), and entropy is unitless.

    \hline
    \section nuclei_section Atomic nuclei

    \b Nuclei

    Atomic nuclei, class \ref nucleus, are implemented as descendants of
    \ref classical. This class sets the value of nucleus::inc_rest_mass to
    false by default. 

    Nuclear mass formulas are given as children of \ref nuclear_mass. 
    The class mnms95_mass provides the mass formula from \ref Moller95,
    and the class ame_mass provides the mass formula from \ref Audi95
    or \ref Audi03. 
    
    The class \ref nuclear_dist provides an experimental generic base
    class for a collection of several nuclei with an STL-like
    iterator.  The sole descendant, simple_dist, provides a simple
    collection of nuclei.

    \hline
    \section parttodo_section Other Todos and Bugs

    Only more general items which aren't particular to a specific
    class are listed here. See the full lists at \ref todo and \ref bug. 

    \bug 
    - Most of the boson classes don't work right now. 

    \hline
    \section partref_section Bibliography

    Some of the references which contain links should direct you to
    the work referred to directly through dx.doi.org.

    \anchor Audi95 Audi95: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/0375-9474(95)00445-9">
    G. Audi and A. H. Wapstra</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/0375-9474(95)00445-9}{
    G. Audi and A. H. Wapstra},
    \endlatexonly
    Nucl. Phys. A \b 595 (1995) 409-480.

    \anchor Audi03 Audi03: 
    \htmlonly
    <a href="http://dx.doi.org/10.1016/j.nuclphysa.2003.11.003">
    G.Audi, A. H. Wapstra and C. Thibault</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1016/j.nuclphysa.2003.11.003}{
    G.Audi, A. H. Wapstra and C. Thibault} ,
    \endlatexonly
    Nucl. Phys. A729 (2003) 337.

    \anchor Eggleton73 Eggleton73:
    P.P. Eggleton, J. Faulkner, and B.P. Flannery,
    Astron. and Astrophys. 23 (1973) 325.

    \anchor Johns96 Johns96:
    \htmlonly
    <a href="http://dx.doi.org/10.1086/178212">
    Johns, P.J. Ellis, and J.M. Lattimer</a>, 
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1086/178212}{
    Johns, P.J. Ellis, and J.M. Lattimer}, 
    \endlatexonly
    Astrophys. J. \b 473, (1996) 1020.

    \anchor Moller95 Moller95:
    \htmlonly
    <a href="http://dx.doi.org/10.1006/adnd.1995.1002">
    P. Moller, J.R. Nix, W.D. Myers, and W.J. Switecki</a>,
    \endhtmlonly
    \latexonly
    \href{http://dx.doi.org/10.1006/adnd.1995.1002}{
    P. Moller, J.R. Nix, W.D. Myers, and W.J. Switecki},
    \endlatexonly
    At. Data Nucl. Data Tables \b 59 (1995) 185.

*/
