/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SIM_ANNEAL_H
#define O2SCL_SIM_ANNEAL_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/multi_min.h>
#include <o2scl/tptr_schedule.h>
#include <o2scl/rnga.h>
#include <o2scl/gsl_rnga.h>
#include <o2scl/tptr_geoseries.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Simulated annealing base

      The seed of the generator is not fixed initially by calls to
      mmin(), so if successive calls should reproduce the same
      results, then the random seed should be set by the user before
      each call.
      
      For the algorithms here, it is important that all of the inputs
      x[i] to the function are scaled similarly relative 
      to the temperature. For example, if the inputs x[i] are all
      of order 1, one might consider a temperature schedule which begins
      with \f$ T=1 \f$ . 

      The number of iterations at each temperature is controlled by
      minimize::ntrial which defaults to 100.

  */
  template<class param_t, class func_t, class vec_t=ovector_view,
    class rng_t=gsl_rnga> class sim_anneal : 
  public multi_min<param_t,func_t,func_t,vec_t> {
    
    public:
  
    sim_anneal() {
      tp=&def_schedule;
      
      this->ntrial=100;
    }

    virtual ~sim_anneal() {}

    /** \brief Calculate the minimum \c fmin of \c func w.r.t the 
	array \c x of size \c nvar.
    */
    virtual int mmin(size_t nvar, vec_t &x, double &fmin, param_t &pa,
		     func_t &func) 
    {
      set_err_ret("Empty base function in sim_anneal::mmin()",
		  gsl_nobase);
    }
      
    /** \brief Specify the temperature schedule */
    int set_tptr_schedule(tptr_schedule<vec_t> &tr) {
      tp=&tr;
      return 0;
    }
    
    /** 
	\brief Print out iteration information.

	Depending on the value of the variable verbose, this prints out
	the iteration information. If verbose=0, then no information is
	printed, while if verbose>1, then after each iteration, the
	present values of x and y are output to std::cout along with the
	iteration number. If verbose>=2 then each iteration waits for a
	character.  
    */
    virtual int print_iter(size_t nv, vec_t &x, double y, int iter,
			   double tptr, std::string comment) 
    {
      if (this->verbose<=0) return 0;
	
      size_t i;
      char ch;
	
      std::cout << comment << " Iteration: " << iter << std::endl;
      text_out_file outs(&std::cout,79);
      outs.word_out("x:");
      for(i=0;i<nv;i++) outs.double_out(x[i]);
      outs.end_line();
      std::cout << "y: " << y << " Tptr: " << tptr << std::endl;
      if (this->verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }
	
      return 0;
    }
      
    /// The default random number generator
    rng_t def_rng;

    /// Return string denoting type, \c "sim_anneal".
    virtual const char *type() { return "sim_anneal"; }

    /// The default temperature schedule
    tptr_geoseries<vec_t> def_schedule;

#ifndef DOXYGEN_INTERNAL

    protected:
    
    /// Pointer to the temperature annealing schedule
    tptr_schedule<vec_t> *tp;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif

