/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_OVECTOR_TLATE_H
#define O2SCL_OVECTOR_TLATE_H

/** \file ovector_tlate.h
    \brief File for definitions of vectors
*/

#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
#include <gsl/gsl_vector.h>
#include <o2scl/err_hnd.h>
#include <o2scl/string_conv.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/array.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief A vector view with finite stride

  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_view_tlate : 
  public vparent_t {
    
  public:

    /// \name Copy constructors
    //@{
    /// Shallow copy constructor - create a new view of the same vector
    ovector_view_tlate(const ovector_view_tlate &v) {
      vparent_t::block=NULL;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=v.stride();
      vparent_t::owner=0;      
    }
    
    /// Shallow copy constructor - create a new view of the same vector
    ovector_view_tlate& operator=(const ovector_view_tlate &v) {
      vparent_t::block=NULL;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=v.stride();
      vparent_t::owner=0;      

      return *this;
    }

    /// Shallow copy constructor - view a unit-stride vector
    ovector_view_tlate(const uvector_view_tlate<data_t> &v) {
      vparent_t::block=NULL;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=1;
      vparent_t::owner=0;      
    }
    
    /// Shallow copy constructor - view a unit-stride vector
    ovector_view_tlate& operator=(const uvector_view_tlate<data_t> &v) {
      vparent_t::block=NULL;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=1;
      vparent_t::owner=0;      

      return *this;
    }
    //@}

    ~ovector_view_tlate() {};

    /// \name Get and set methods
    //@{
    /** 
	\brief Array-like indexing 
    */
    data_t &operator[](size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    data_t &operator()(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator()(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
    /** \brief Get (with optional range-checking) */
    data_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::get(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
    /** \brief Get pointer (with optional range-checking) */
    data_t *get_ptr(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::get_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data;
      }
#endif
      return vparent_t::data+i*vparent_t::stride;
    }
    
    /** \brief Get pointer (with optional range-checking) */
    const data_t *get_const_ptr(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_view_tlate::get_const_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data;
      }
#endif
      return (const data_t *)(vparent_t::data+i*vparent_t::stride);
    }
    
    /** \brief Set (with optional range-checking) */
    int set(size_t i, data_t val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err_ret((((std::string)"Array index ")+itos(i)+" out of bounds"
		     +" in ovector_view_tlate::set(). Size: "+
		     itos(vparent_t::size)+
		     " (index should be less than size).").c_str(),gsl_index);
      }
#endif
      vparent_t::data[i*vparent_t::stride]=val;
      return 0;
    }

    /** \brief Set all of the value to be the value \c val */
    int set_all(double val) {
      for(size_t i=0;i<vparent_t::size;i++) {
	vparent_t::data[i*vparent_t::stride]=val;
      }
      return 0;
    }

    /** 
	\brief Method to return vector size 
	
	If no memory has been allocated, this will quietly 
	return zero.
    */
    size_t size() const {
      return vparent_t::size;
    }

    /** 
	\brief Method to return capacity

	Analogous to \c std::vector<>.capacity()
    */
    size_t capacity() const {
      if (vparent_t::block) return vparent_t::block->size;
      return 0;
    }

    /** 
	\brief Method to return vector stride
	
	If no memory has been allocated, this will quietly 
	return zero.
    */
    size_t stride() const {
      return vparent_t::stride;
    }
    //@}

    /// \name Arithmetic
    //@{
    /** \brief operator+= */
    ovector_view_tlate<data_t,vparent_t,block_t> &operator+=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      size_t lsize=x.size();
      if (lsize>this->size()) lsize=this->size();
      for(size_t i=0;i<lsize;i++) (*this)[i]+=x[i];
      
      return *this;
    }
    
    /** \brief operator-= */
    ovector_view_tlate<data_t,vparent_t,block_t> &operator-=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      size_t lsize=x.size();
      if (lsize>this->size()) lsize=this->size();
      for(size_t i=0;i<lsize;i++) (*this)[i]-=x[i];
      
      return *this;
    }

    /** \brief operator+= */
    ovector_view_tlate<data_t,vparent_t,block_t> &operator+=(data_t &x) {
      for(size_t i=0;i<this->size();i++) (*this)[i]+=x;
      return *this;
    }
    
    /** \brief operator-= */
    ovector_view_tlate<data_t,vparent_t,block_t> &operator-=(data_t &x) {
      for(size_t i=0;i<this->size();i++) (*this)[i]-=x;
      return *this;
    }
    
    /** \brief operator*= */
    ovector_view_tlate<data_t,vparent_t,block_t> &operator*=(const data_t &y) {
      for(size_t i=0;i<this->size();i++) (*this)[i]*=y;

      return *this;
    }

    /** \brief Norm */
    data_t norm() const {
      data_t result=0;
      for(size_t i=0;i<this->size;i++) {
	result+=(*this)[i]*(*this)[i];
      }
      return sqrt(result);
    }
    //@}
    
    /// \name Other methods
    //@{
    /** \brief Swap vectors */
    int swap(ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      size_t t1, t2;
      double *t3;
      block_t *t4;
      int t5;
      
      t1=vparent_t::size;
      t2=vparent_t::stride;
      t3=vparent_t::data;
      t4=vparent_t::block;
      t5=vparent_t::owner;
      
      vparent_t::size=x.size;
      vparent_t::stride=x.stride;
      vparent_t::data=x.data;
      vparent_t::block=x.block;
      vparent_t::owner=x.owner;

      x.size=t1;
      x.stride=t2;
      x.data=t3;
      x.block=t4;
      x.owner=t5;

      return 0;
    }

    /** 
	\brief Return true if this object owns the data it refers to

	This can be used to determine if an object is a "vector_view",
	or a legitmate "vector". If is_owner() is true, then it is 
	an \ref ovector_tlate object.

	If any O2scl class creates a \ref ovector_tlate object in
	which \ref is_owner() returns false, then it is a bug!
    */
    bool is_owner() const {
      if (vparent_t::owner==1) return true;
      return false;
    }

    /** \brief Exhaustively look through the array for a
	particular value 

	This can only fail if \em all of the entries in the array are
	not finite, in which case it calls set_err() and returns
	0. The error handler is reset at the beginning of lookup().
    */
    size_t lookup(const data_t x0) const {
      err_hnd->reset();
      const ovector_view_tlate<data_t,vparent_t,block_t> *a=this;
      size_t row=0, i=0, nvar=size();
      while(!finite((*a)[i]) && i<nvar-1) i++;
      if (i==nvar-1) {
	set_err("Array not finite in ovector_view_tlate::lookup()",1);
	return 0;
      }
      data_t best=(*a)[i], bdiff=fabs((*a)[i]-x0);
      for(;i<nvar;i++) {
	if (finite((*a)[i]) && fabs((*a)[i]-x0)<bdiff) {
	  row=i;
	  best=(*a)[i];
	  bdiff=fabs((*a)[i]-x0);
	}
      }
      return row;
    }

    /** \brief Find the maximum element */
    data_t max() const {
      data_t maxval=0;
      if (vparent_t::size>0) {
	bool setb=false;
	for(size_t i=0;i<vparent_t::size;i++) {
	  double v=vparent_t::data[i];
	  if (finite(v)) {
	    if (setb==false) {
	      maxval=v;
	      setb=true;
	    } else if (v>maxval) {
	      maxval=v;
	    }
	  }
	}
	if (setb==false) {
	  set_err("No finite values in ovector_view_tlate::max().",
		  gsl_efailed);
	  return 0.0;
	}
      } else {
	return 0.0;
      }
      return maxval;
    }

    /** \brief Find the location of the maximum element */
    size_t max_index() const {
      data_t maxval;
      size_t loc;
      if (vparent_t::size>0) {
	bool setb=false;
	for(size_t i=0;i<vparent_t::size;i++) {
	  double v=vparent_t::data[i];
	  if (finite(v)) {
	    if (setb==false) {
	      maxval=v;
	      loc=i;
	      setb=true;
	    } else if (v>maxval) {
	      maxval=v;
	      loc=i;
	    }
	  }
	}
	if (setb==false) {
	  set_err("No finite values in ovector_view_tlate::max().",
		  gsl_efailed);
	  return 0;
	}
      } else {
	return 0;
      }
      return loc;
    }

    /** \brief Find the minimum element */
    data_t min() const {
      data_t minval=0;
      if (vparent_t::size>0) {
	bool setb=false;
	for(size_t i=0;i<vparent_t::size;i++) {
	  double v=vparent_t::data[i];
	  if (finite(v)) {
	    if (setb==false) {
	      minval=v;
	      setb=true;
	    } else if (v<minval) {
	      minval=v;
	    }
	  }
	}
	if (setb==false) {
	  set_err("No finite values in ovector_view_tlate::min().",
		  gsl_efailed);
	  return 0.0;
	}
      } else {
	return 0.0;
      }
      return minval;
    }

    /** \brief Find the location of the minimum element */
    size_t min_index() const {
      data_t maxval;
      size_t loc;
      if (vparent_t::size>0) {
	bool setb=false;
	for(size_t i=0;i<vparent_t::size;i++) {
	  double v=vparent_t::data[i];
	  if (finite(v)) {
	    if (setb==false) {
	      maxval=v;
	      loc=i;
	      setb=true;
	    } else if (v<maxval) {
	      maxval=v;
	      loc=i;
	    }
	  }
	}
	if (setb==false) {
	  set_err("No finite values in ovector_view_tlate::max().",
		  gsl_efailed);
	  return 0;
	}
      } else {
	return 0;
      }
      return loc;
    }

    /** \brief Return a gsl vector */
    vparent_t *get_gsl_vector() { return this; };

    /** \brief Return a \c const gsl vector */
    const vparent_t *get_gsl_vector_const() const { return this; };

    //@}

#ifndef DOXYGEN_INTERNAL

  protected:

    /** \brief Empty constructor provided for use by 
	ovector_tlate(const ovector_tlate &v)
    */
    ovector_view_tlate() {};

#endif
    
  };

  /** 
      \brief A vector with finite stride


      There are several global binary operators associated with
      objects of type \ref uvector_tlate. The are documented in the
      "Functions" section of \ref ovector_tlate.h.
  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    
    /// \name Standard constructor
    //@{
    /** \brief Create an ovector of size \c n with owner as 'true' */
    ovector_tlate(size_t n=0) {
      
      vparent_t::data=0;
      vparent_t::size=0;
      vparent_t::stride=0;
      vparent_t::owner=0;
      vparent_t::block=0;
      
      // This must be set to 1 even if n=0 so that future calls to
      // operator= (and other functions) work properly
      vparent_t::owner=1;

      if (n>0) {
	vparent_t::block=(block_t *)malloc(sizeof(block_t));
	if (vparent_t::block) {
	  vparent_t::block->data=(data_t *)malloc(n*sizeof(data_t));
	  if (vparent_t::block->data) {
	    vparent_t::block->size=n;
	    vparent_t::data=vparent_t::block->data;
	    vparent_t::size=n;
	    vparent_t::stride=1;
	  } else {
	    std::free(vparent_t::block);
	    set_err("No memory for data in ovector constructor",
		    gsl_enomem);
	  }
	} else {
	  set_err("No memory for block in ovector contructor",
		  gsl_enomem);
	}
      }
    }
    //@}
    
    /// \name Copy constructors
    //@{
    /// Deep copy constructor, allocate new space and make a copy
    ovector_tlate(const ovector_tlate &v) : 
      ovector_view_tlate<data_t,vparent_t,block_t>() {
	
      vparent_t::data=0;
      vparent_t::size=0;
      vparent_t::stride=0;
      vparent_t::owner=0;
      vparent_t::block=0;
	
      size_t n=v.size();
      if (n>0) {
	vparent_t::block=(block_t *)malloc(sizeof(block_t));
	if (vparent_t::block) {
	  vparent_t::block->data=(data_t *)malloc(n*sizeof(data_t));
	  if (vparent_t::block->data) {
	    vparent_t::block->size=n;
	    vparent_t::data=vparent_t::block->data;
	    vparent_t::size=n;
	    vparent_t::stride=1;
	    vparent_t::owner=1;
	    for(size_t i=0;i<n;i++) {
	      vparent_t::data[i*vparent_t::stride]=v[i];
	    }
	  } else {
	    std::free(vparent_t::block);
	    set_err("No memory for data in ovector constructor",
		    gsl_enomem);
	  }
	} else {
	  set_err("No memory for block in ovector contructor",
		  gsl_enomem);
	}
      } else {
	vparent_t::size=0;
	vparent_t::stride=0;
      }
	
    }
      
#ifdef O2SCL_NEVER_DEFINED
  } 
  {
#endif  

    /// Deep copy constructor, allocate new space and make a copy
    ovector_tlate(const ovector_view_tlate<data_t,vparent_t,block_t> &v) : 
      ovector_view_tlate<data_t,vparent_t,block_t>() {
      
      vparent_t::data=0;
      vparent_t::size=0;
      vparent_t::stride=0;
      vparent_t::owner=0;
      vparent_t::block=0;
      
      size_t n=v.size();
      if (n>0) {
	vparent_t::block=(block_t *)malloc(sizeof(block_t));
	if (vparent_t::block) {
	  vparent_t::block->data=(data_t *)malloc(n*sizeof(data_t));
	  if (vparent_t::block->data) {
	    vparent_t::block->size=n;
	    vparent_t::data=vparent_t::block->data;
	    vparent_t::size=n;
	    vparent_t::stride=1;
	    vparent_t::owner=1;
	    for(size_t i=0;i<n;i++) {
	      vparent_t::data[i*vparent_t::stride]=v[i];
	    }
	  } else {
	    std::free(vparent_t::block);
	    set_err("No memory for data in ovector constructor",
		    gsl_enomem);
	  }
	} else {
	  set_err("No memory for block in ovector contructor",
		  gsl_enomem);
	}
      } else {
	vparent_t::size=0;
	vparent_t::stride=0;
      }
	
    }
	
#ifdef O2SCL_NEVER_DEFINED
  } 
  {
#endif  

    /// Deep copy constructor, allocate new space and make a copy
    ovector_tlate(const uvector_view_tlate<data_t> &v) : 
      ovector_view_tlate<data_t,vparent_t,block_t>() {
	  
      vparent_t::data=0;
      vparent_t::size=0;
      vparent_t::stride=0;
	  
      size_t n=v.size();
	  
      if (n>0) {
	vparent_t::block=(block_t *)malloc(sizeof(block_t));
	if (vparent_t::block) {
	  vparent_t::block->data=(data_t *)malloc(n*sizeof(data_t));
	  if (vparent_t::block->data) {
	    vparent_t::block->size=n;
	    vparent_t::data=vparent_t::block->data;
	    vparent_t::size=n;
	    vparent_t::stride=1;
	    vparent_t::owner=1;
	    for(size_t i=0;i<n;i++) {
	      vparent_t::data[i*vparent_t::stride]=v[i];
	    }
	  } else {
	    std::free(vparent_t::block);
	    set_err("No memory for data in ovector constructor",
		    gsl_enomem);
	  }
	} else {
	  set_err("No memory for block in ovector contructor",
		  gsl_enomem);
	}
      } else {
	vparent_t::size=0;
	vparent_t::stride=0;
      }

    }
    
#ifdef O2SCL_NEVER_DEFINED
  } 
  {
#endif  

    /** \brief Deep copy constructor, if owner is true, allocate
	space and make a new copy, otherwise, just copy into the
	view
    */
    ovector_tlate& operator=(const ovector_tlate &v) {

      size_t size2=v.size();
      if (this->owner) {
	allocate(size2);
      } else {
	if (this->size()!=size2) {
	  set_err("Sizes don't match in ovector::operator=(ovector)",
		  gsl_ebadlen);
	  return *this;
	}
      }
      for(size_t i=0;i<size2;i++) {
	vparent_t::data[i*vparent_t::stride]=v[i];
      }

      return *this;
    }

    /** \brief Deep copy constructor, if owner is true, allocate
	space and make a new copy, otherwise, just copy into the
	view
    */
    ovector_tlate& operator=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &v) {

      size_t size2=v.size();
      if (this->owner) {
	allocate(size2);
      } else {
	if (this->size()!=size2) {
	  set_err
	    ("Sizes don't match in ovector::operator=(ovector_view)",
	     gsl_ebadlen);
	  return *this;
	}
      }
      for(size_t i=0;i<size2;i++) {
	vparent_t::data[i*vparent_t::stride]=v[i];
      }

      return *this;
    }

    /** \brief Deep copy constructor, if owner is true, allocate
	space and make a new copy, otherwise, just copy into the
	view
    */
    ovector_tlate& operator=
      (const uvector_view_tlate<data_t> &v) {

      size_t size2=v.size();
      if (this->owner) {
	allocate(size2);
      } else {
	if (this->size()!=size2) {
	  set_err
	    ("Sizes don't match in ovector::operator=(uvector_view)",
	     gsl_ebadlen);
	  return *this;
	}
      }
      for(size_t i=0;i<size2;i++) {
	vparent_t::data[i*vparent_t::stride]=v[i];
      }

      return *this;
    }
    //@}

    ~ovector_tlate() {
      if (vparent_t::size>0) {
	if (vparent_t::owner==1) {
	  if (vparent_t::block->size>0) {
	    std::free(vparent_t::block->data);
	  }
	  std::free(vparent_t::block);
	  vparent_t::size=0;
	}
      }
    }
    
    /// \name Memory allocation
    //@{
    /** 
	\brief Allocate memory for size \c n after freeing any memory
	presently in use
    */
    int allocate(size_t nsize) {
      if (vparent_t::size>0) free();

      if (nsize>0) {
	vparent_t::block=(block_t *)malloc(sizeof(block_t));
	if (vparent_t::block) {
	  vparent_t::block->data=(data_t *)malloc(nsize*sizeof(data_t));
	  if (vparent_t::block->data) {
	    vparent_t::block->size=nsize;
	    vparent_t::data=vparent_t::block->data;
	    vparent_t::size=nsize;
	    vparent_t::stride=1;
	    vparent_t::owner=1;
	  } else {
	    std::free(vparent_t::block);
	    set_err_ret("No memory for data in ovector::allocate()",
			gsl_enomem);
	  }
	} else {
	  set_err_ret("No memory for block in ovector::allocate()",
		      gsl_enomem);
	}
      } else {
	set_err_ret("Zero size in ovector::allocate()",gsl_einval);
      }
      return 0;
    }

    /** 
	\brief Free the memory 
    
	This function will safely do nothing if used without first
	allocating memory or if called multiple times in succession.
    */
    int free() {
      if (vparent_t::size>0) {
	if (vparent_t::owner==1) {
	  if (vparent_t::block->size>0) {
	    std::free(vparent_t::block->data);
	  }
	  std::free(vparent_t::block);
	}
	vparent_t::size=0;
	vparent_t::stride=0;
      }
      return 0;
    }
    //@}

    /// \name Stack-like operations (very experimental)
    //@{

    /// Add a value to the end of the vector
    int push_back(data_t val) {

      if (vparent_t::owner==0) {

	set_err_ret
	  ("Cannot push on a vector view in ovector::push_back().",
	   gsl_einval);

      } else {

	if (vparent_t::size==0) {

	  // Empty, so make a 1-element vector
	  allocate(1);
	  vparent_t::data[0]=val;
	  
	} else if (vparent_t::block->size==vparent_t::size) {

	  // Allocate new memory in vparent_t::block
	  vparent_t::block->data=(data_t *)malloc
	    (2*vparent_t::block->size*sizeof(data_t));
	  vparent_t::block->size*=2;
		
	  // Copy the data to the new memory
	  for(size_t i=0;i<vparent_t::size;i++) {
	    vparent_t::block->data[i]=vparent_t::data[i];
	  }
	  
	  // Delete the old memory
	  std::free(vparent_t::data);
	  
	  // Reset data to the new memory
	  vparent_t::data=vparent_t::block->data;

	  // Add the new value to the end of the array and 
	  // increment size
	  vparent_t::block->data[vparent_t::size]=val;
	  vparent_t::size++;

	} else {

	  // Add the new value to the end of the array and 
	  // increment size
	  vparent_t::block->data[vparent_t::size]=val;
	  vparent_t::size++;

	}

      }

      return 0;
    }

    /** 
	\brief Reserve memory by increasing capacity
    
	Increase the maximum capacity of the vector so that calls
	to push_back() do not need to automatically increase the
	capacity.

	This function quietly does nothing if \c cap is smaller than
	the present vector size given by \ref size().
    */
    int reserve(size_t cap) {
	    
      if (vparent_t::owner==0) {

	set_err_ret
	  ("Can't reserve on a vector view in ovector::reserve().",
	   gsl_einval);
	      
      } else {

	// Do nothing if we are reserving memory for less than the
	// current size
	      
	if (cap>vparent_t::size) {

	  // If it's empty, we need to allocate a block
	  if (!vparent_t::block) {
	    vparent_t::block=(block_t *)malloc(sizeof(block_t));
	  }
		
	  // Allocate new memory in vparent_t::block
	  vparent_t::block->data=(data_t *)malloc(cap*sizeof(data_t));
	  vparent_t::block->size=cap;

	  // Copy the data to the new memory
	  for(size_t i=0;i<vparent_t::size;i++) {
	    vparent_t::block->data[i]=vparent_t::data[i];
	  }

	  // Delete the old memory
	  std::free(vparent_t::data);

	  // Reset data to the new memory
	  vparent_t::data=vparent_t::block->data;
	}
      }
      return 0;
    }

    /// Return the last value and shrink the vector size by one
    data_t pop() {
      if (vparent_t::owner==0) {
	set_err("Cannot pop() on a vector view in ovector::pop().",
		gsl_einval);
	return 0.0;
      } else if (vparent_t::size==0) {
	set_err("Attempted to pop an empty vector in ovector::pop().",
		gsl_einval);
	return 0.0;
      }
      vparent_t::size--;
      return vparent_t::data[vparent_t::size];
    }
    //@}

    /// \name Other methods
    //@{
    int erase(size_t ix) {
      if (vparent_t::owner==0) {
	set_err_ret("Vector view sent to ovector::erase().",
		    gsl_einval);
      }
      if (ix<vparent_t::size) {
	for(size_t i=ix+1;ix<vparent_t::size;i++) {
	  vparent_t::data[ix-1]=vparent_t::data[ix];
	}
	vparent_t::size--;
      } else {
	set_err_ret("Tried to erase() past end in ovector::erase().",
		    gsl_einval);
      }
    }
    //@}
    
  };

  /** 
      \brief Create a vector from an array 
  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_array_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n */
    ovector_array_tlate(size_t n, data_t *dat) {
      if (n>0) {
	vparent_t::block=NULL;
	vparent_t::data=dat;
	vparent_t::size=n;
	vparent_t::stride=1;
	vparent_t::owner=0;
      }
    }
  };

  /** 
      \brief Create a vector from an array with a stride
  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_array_stride_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n and stride \c s 
     */
    ovector_array_stride_tlate(size_t n, data_t *dat, size_t s) {
      if (n>0) {
	vparent_t::block=NULL;
	vparent_t::data=dat;
	vparent_t::size=n;
	vparent_t::stride=s;
	vparent_t::owner=0;
      }
    }
  };

  /** 
      \brief Create a vector from a subvector of another


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_subvector_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c orig */
    ovector_subvector_tlate
      (ovector_view_tlate<data_t,vparent_t,block_t> &orig, 
       size_t offset, size_t n) {
      if (offset+n-1<orig.size()) {
	vparent_t::block=NULL;
	vparent_t::data=orig.data+offset*orig.stride();
	vparent_t::size=n;
	vparent_t::stride=orig.stride();
	vparent_t::owner=0;
      } else {
	vparent_t::size=0;
	set_err("Subvector failed in ovector_subvector() constructor.",
		gsl_efailed);
      }
    }
  };

  /** 
      \brief Create a const vector from an array 


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_const_array_tlate :
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n */
    ovector_const_array_tlate(size_t n, const data_t *dat) {
      if (n>0) {
	vparent_t::block=NULL;
	// We have to do an explicit cast here, but we prevent the
	// user from changing the data.
	vparent_t::data=(data_t *)dat;
	vparent_t::size=n;
	vparent_t::stride=1;
	vparent_t::owner=0;
      }
    }
    
    ~ovector_const_array_tlate() {};
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /** \name Ensure \c const by hiding non-const members
     */
    //@{
    data_t &operator[](size_t i) { return vparent_t::data[0]; }
    data_t &operator()(size_t i) { return vparent_t::data[0]; }
    data_t *get_ptr(size_t i) { return NULL; }
    int set(size_t i, data_t val) { return 0; }
    int swap(ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return 0; 
    }
    int set_all(double val) { return 0; }
    vparent_t *get_gsl_vector() { return NULL; }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator+=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator-=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator*=(const data_t &y) {
      return *this;
    }
    //@}
      
#endif

  };

  /** 
      \brief Create a const vector from an array with a stride 


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_const_array_stride_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n */
    ovector_const_array_stride_tlate(size_t n, const data_t *dat, size_t s) {
      if (n>0) {
	vparent_t::block=NULL;
	// We have to do an explicit cast here, but we prevent the
	// user from changing the data.
	vparent_t::data=(data_t *)dat;
	vparent_t::size=n;
	vparent_t::stride=s;
	vparent_t::owner=0;
      }
    }

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /** \name Ensure \c const by hiding non-const members
     */
    //@{
    data_t &operator[](size_t i) { return vparent_t::data[0]; }
    data_t &operator()(size_t i) { return vparent_t::data[0]; }
    data_t *get_ptr(size_t i) { return NULL; }
    int set(size_t i, data_t val) { return 0; }
    int swap(ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return 0; 
    }
    int set_all(double val) { return 0; }
    vparent_t *get_gsl_vector() { return NULL; }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator+=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator-=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator*=(const data_t &y) {
      return *this;
    }
    //@}

#endif

  };

  /** 
      \brief Create a const vector from a subvector of another vector


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_const_subvector_tlate :
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c orig 
     */
    ovector_const_subvector_tlate
      (const ovector_view_tlate<data_t,vparent_t,block_t> &orig, 
       size_t offset, size_t n) {
      if (offset+n-1<orig.size()) {
	vparent_t::block=NULL;
	vparent_t::data=orig.data+offset*orig.stride();
	vparent_t::size=n;
	vparent_t::stride=orig.stride();
	vparent_t::owner=0;
      } else {
	vparent_t::block=0;
	vparent_t::data=0;
	vparent_t::size=0;
	vparent_t::stride=0;
	vparent_t::owner=0;
	set_err("Failed in ovector_subvector() constructor.",gsl_efailed);
      }
    }

    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_const_subvector_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
      }
#endif
      return vparent_t::data[i*vparent_t::stride];
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /** \name Ensure \c const by hiding non-const members
     */
    //@{
    data_t &operator[](size_t i) { return vparent_t::data[0]; }
    data_t &operator()(size_t i) { return vparent_t::data[0]; }
    data_t *get_ptr(size_t i) { return NULL; }
    int set(size_t i, data_t val) { return 0; }
    int swap(ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return 0; 
    }
    int set_all(double val) { return 0; }
    vparent_t *get_gsl_vector() { return NULL; }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator+=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator-=
      (const ovector_view_tlate<data_t,vparent_t,block_t> &x) {
      return *this;
    }
    ovector_view_tlate<data_t,vparent_t,block_t> &operator*=(const data_t &y) {
      return *this;
    }
    //@}

#endif

  };
  
  /** 
      \brief Reversed view of a vector



      Note that you cannot reverse a reversed vector and expect to get
      the original vector back.
  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_reverse_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n and stride \c s 
     */
    ovector_reverse_tlate(ovector_view_tlate<data_t,vparent_t,block_t> &v) {
      vparent_t::block=v.block;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=v.stride();
      vparent_t::owner=0;
    }

    /// \name Get and set methods
    //@{
    /** 
	\brief Array-like indexing 
    */
    data_t &operator[](size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    data_t &operator()(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator()(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get (with optional range-checking) */
    data_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::get(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get pointer (with optional range-checking) */
    data_t *get_ptr(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::get_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data;
      }
#endif
      return vparent_t::data+(vparent_t::size-1-i)*vparent_t::stride;
    }
    
    /** \brief Get pointer (with optional range-checking) */
    const data_t *get_const_ptr(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_reverse_tlate::get_const_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return (const data_t *)vparent_t::data;
      }
#endif
      return (const data_t *)(vparent_t::data+
			      (vparent_t::size-1-i)*vparent_t::stride);
    }
    
    /** \brief Set (with optional range-checking) */
    int set(size_t i, data_t val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err_ret((((std::string)"Array index ")+itos(i)+" out of bounds"
		     +" in ovector_reverse_tlate::set(). Size: "+
		     itos(vparent_t::size)+
		     " (index should be less than size).").c_str(),gsl_index);
      }
#endif
      vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride]=val;
      return 0;
    }

    /** \brief Set all of the value to be the value \c val */
    int set_all(double val) {
      for(size_t i=0;i<vparent_t::size;i++) {
	vparent_t::data[i*vparent_t::stride]=val;
      }
      return 0;
    }

  };

  /** 
      \brief Reversed view of a vector


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_const_reverse_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n and stride \c s 
     */
    ovector_const_reverse_tlate
      (const ovector_view_tlate<data_t,vparent_t,block_t> &v) {
      vparent_t::block=v.block;
      vparent_t::data=v.data;
      vparent_t::size=v.size();
      vparent_t::stride=v.stride();
      vparent_t::owner=0;
    }

    /// \name Get and set methods
    //@{
    /** 
	\brief Array-like indexing 
    */
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_const_reverse_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator()(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_const_reverse_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get (with optional range-checking) */
    data_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_const_reverse_tlate::get(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get pointer (with optional range-checking) */
    const data_t *get_const_ptr(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_const_reverse_tlate::get_const_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return (const data_t *)vparent_t::data;
      }
#endif
      return (const data_t *)(vparent_t::data+
			      (vparent_t::size-1-i)*vparent_t::stride);
    }
    
  };

  /** 
      \brief Reversed view of a subvector



      Note that you cannot reverse a reversed vector and expect to get
      the original vector back.
  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_subvector_reverse_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {
  public:
    /** \brief Create a vector from \c dat with size \c n and stride \c s 
     */
    ovector_subvector_reverse_tlate
      (ovector_view_tlate<data_t,vparent_t,block_t> &v, size_t offset, 
       size_t n) {
      vparent_t::data=0;
      vparent_t::size=0;
      vparent_t::stride=0;
      if (offset+n-1<v.size()) {
	vparent_t::block=NULL;
	vparent_t::data=v.data+offset*v.stride();
	vparent_t::size=n;
	vparent_t::stride=v.stride();
	vparent_t::owner=0;
      } else {
	set_err("Failed in ovector_sub_reverse() constructor.",gsl_einval);
      }
    }

    /// \name Get and set methods
    //@{
    /** 
	\brief Array-like indexing 
    */
    data_t &operator[](size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::operator[]. Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    data_t &operator()(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator()(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::operator(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get (with optional range-checking) */
    data_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::get(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get pointer (with optional range-checking) */
    data_t *get_ptr(size_t i) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"
		 +" in ovector_subvector_reverse_tlate::get_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data;
      }
#endif
      return vparent_t::data+(vparent_t::size-1-i)*vparent_t::stride;
    }
    
    /** \brief Get pointer (with optional range-checking) */
    const data_t *get_const_ptr(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds in "
		 +"ovector_subvector_reverse_tlate::get_const_ptr(). Size: "+
		 itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return (const data_t *)vparent_t::data;
      }
#endif
      return (const data_t *)(vparent_t::data+
			      (vparent_t::size-1-i)*vparent_t::stride);
    }
    
    /** \brief Set (with optional range-checking) */
    int set(size_t i, data_t val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err_ret((((std::string)"Array index ")+itos(i)+" out of bounds "
		     +" in ovector_subvector_reverse_tlate::get_const_ptr(). "+
		     "Size: "+itos(vparent_t::size)+
		     " (index should be less than size).").c_str(),gsl_index);
      }
#endif
      vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride]=val;
      return 0;
    }

    /** \brief Set all of the value to be the value \c val */
    int set_all(double val) {
      for(size_t i=0;i<vparent_t::size;i++) {
	vparent_t::data[i*vparent_t::stride]=val;
      }
      return 0;
    }

  };

  /** 
      \brief Reversed view of a const subvector


  */
  template<class data_t, class vparent_t, class block_t> 
    class ovector_const_subvector_reverse_tlate : 
  public ovector_view_tlate<data_t,vparent_t,block_t> {

  public:

    /** \brief Create a vector from \c dat with size \c n and stride \c s 
     */
    ovector_const_subvector_reverse_tlate
      (const ovector_view_tlate<data_t,vparent_t,block_t> &v, size_t offset,
       size_t n) {
      if (offset+n-1<v.size()) {
	vparent_t::block=NULL;
	vparent_t::data=v.data+offset*v.stride();
	vparent_t::size=n;
	vparent_t::stride=v.stride();
	vparent_t::owner=0;
      } else {
	set_err("Subvector failed in ovector_sub_reverse().",gsl_einval);
      }
    }

    /// \name Get and set methods
    //@{
    /** 
	\brief Array-like indexing 
    */
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator[](size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"+
		 " in ovector_const_subvector_reverse_tlate::operator[]."+
		 " Size: "+itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const data_t &operator()(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"+
		 " in ovector_const_subvector_reverse_tlate::operator()."+
		 " Size: "+itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get (with optional range-checking) */
    data_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"+
		 " in ovector_const_subvector_reverse_tlate::get()."+
		 " Size: "+itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return vparent_t::data[0];
      }
#endif
      return vparent_t::data[(vparent_t::size-1-i)*vparent_t::stride];
    }
    
    /** \brief Get pointer (with optional range-checking) */
    const data_t *get_const_ptr(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=vparent_t::size) {
	set_err((((std::string)"Array index ")+itos(i)+" out of bounds"+
		 " in ovector_const_subvector_reverse_tlate::get_const_ptr()."+
		 " Size: "+itos(vparent_t::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return (const data_t *)vparent_t::data;
      }
#endif
      return (const data_t *)(vparent_t::data+
			      (vparent_t::size-1-i)*vparent_t::stride);
    }
    
  };

  /// ovector typedef
  typedef ovector_tlate<double,gsl_vector,gsl_block> ovector;
  /// ovector_view typedef
#ifdef DOXYGENP
  typedef ovector_view_tlate<data_t,vparent_t,block_t> ovector_view;
#else
  typedef ovector_view_tlate<double,gsl_vector,gsl_block> ovector_view;
#endif
  /// ovector_array typedef
  typedef ovector_array_tlate<double,gsl_vector,gsl_block> ovector_array;
  /// ovector_array_stride typedef
  typedef ovector_array_stride_tlate<double,gsl_vector,gsl_block> 
    ovector_array_stride;
  /// ovector_subvector typedef
  typedef ovector_subvector_tlate<double,gsl_vector,gsl_block> 
    ovector_subvector;
  /// ovector_const_array typedef
  typedef ovector_const_array_tlate<double,gsl_vector,gsl_block> 
    ovector_const_array;
  /// ovector_const_array_stride typedef
  typedef ovector_const_array_stride_tlate<double,gsl_vector,gsl_block> 
    ovector_const_array_stride;
  /// ovector_const_subvector typedef
  typedef ovector_const_subvector_tlate<double,gsl_vector,gsl_block> 
    ovector_const_subvector;
  /// ovector_reverse typedef
  typedef ovector_reverse_tlate<double,gsl_vector,gsl_block> 
    ovector_reverse;
  /// ovector_const_reverse typedef
  typedef ovector_const_reverse_tlate<double,gsl_vector,gsl_block> 
    ovector_const_reverse;
  /// ovector_subvector_reverse typedef
  typedef ovector_subvector_reverse_tlate<double,gsl_vector,gsl_block> 
    ovector_subvector_reverse;
  /// ovector_const_subvector_reverse typedef
  typedef ovector_const_subvector_reverse_tlate<double,gsl_vector,gsl_block> 
    ovector_const_subvector_reverse;
  
  /// ovector_int typedef
  typedef ovector_tlate<int,gsl_vector_int,gsl_block_int> ovector_int;
  /// ovector_int_view typedef
  typedef ovector_view_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_view;
  /// ovector_int_array typedef
  typedef ovector_array_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_array;
  /// ovector_int_array_stride typedef
  typedef ovector_array_stride_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_array_stride;
  /// ovector_int_subvector typedef
  typedef ovector_subvector_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_subvector;
  /// ovector_int_const_array typedef
  typedef ovector_const_array_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_const_array;
  /// ovector_int_const_array_stride typedef
  typedef ovector_const_array_stride_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_const_array_stride;
  /// ovector_int_const_subvector typedef
  typedef ovector_const_subvector_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_const_subvector;
  /// ovector_int_reverse typedef
  typedef ovector_reverse_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_reverse;
  /// ovector_int_const_reverse typedef
  typedef ovector_const_reverse_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_const_reverse;
  /// ovector_int_subvector_reverse typedef
  typedef ovector_subvector_reverse_tlate<int,gsl_vector_int,gsl_block_int> 
    ovector_int_subvector_reverse;
  /// ovector_int_const_subvector_reverse typedef
  typedef ovector_const_subvector_reverse_tlate<int,gsl_vector_int,
    gsl_block_int> ovector_int_const_subvector_reverse;

  /** 
      \brief A operator for naive vector output

      This outputs all of the vector elements. All of these are
      separated by one space character, though no trailing space or \c
      endl is sent to the output.
  */
  template<class data_t, class vparent_t, class block_t> 
    std::ostream &operator<<
    (std::ostream &os, 
     const ovector_view_tlate<data_t,vparent_t,block_t> &v) {
    for(size_t i=0;i<v.size()-1;i++) {
      os << v[i] << ' ';
    }
    os << v[v.size()-1];
    return os;
  }
  
  /** \brief A simple class to provide an \c allocate() function
      for \ref ovector
      

  */
  class ovector_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(ovector &o, int i) { o.allocate(i); }
    /// Free memory
    void free(ovector &o) { o.free(); }
  };

  /** \brief A simple class to provide an \c allocate() function
      for \ref ovector_int
  */
  class ovector_int_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(ovector_int &o, int i) { o.allocate(i); }
    /// Free memory
    void free(ovector_int &o) { o.free(); }
  };

  /** \brief A vector where the memory allocation is performed in 
      the constructor


  */
#ifdef DOXYGENP
  template<size_t N=0> class ofvector : 
  public ovector_tlate<data_t,vparent_t,block_t>
#else
    template<size_t N=0> class ofvector : 
  public ovector_tlate<double,gsl_vector,gsl_block>
#endif
    {
      public:
      ofvector() : ovector_tlate<double,gsl_vector,gsl_block>(N) {
      }
    };
  
#ifndef DOXYGENP
}
#endif

#endif

