/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SEARCH_VEC_H
#define O2SCL_SEARCH_VEC_H

#include <iostream>
#include <string>
#include <gsl/gsl_spline.h>
#include <o2scl/err_hnd.h>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>
/* \file search_vec.h
   \brief File for definitions for \ref search_vec
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Searching class for monotonic data

      A searching class for monotonic vectors. A caching system
      similar to \c gsl_interp_accel is used.
      
      For normal usage, just use find(). If you happen to know in
      advance that the vector is increasing or decreasing, then you
      can use find_inc() or find_dec() instead. 

      \todo The documentation here is still kind of unclear. 
  */
  template<class vec_t> class search_vec {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Storage for the most recent index
    size_t cache;

#endif

  public:

    /** \brief Search an increasing or decreasing vector
     */
    size_t find(const double x0, size_t n, const vec_t &x) {
      if (x[0]<x[n-1]) return find_inc(x0,n,x);
      return find_dec(x0,n,x);
    }

    /** \brief Search part of a increasing vector
     */
    size_t find_inc(const double x0, size_t n, const vec_t &x) {
      size_t x_index=cache;

      // Make sure the cached value is not outside the array.
      if (x_index>n-1) x_index=n/2;
  
      if (x0<x[x_index]) {
	cache=bsearch_inc(x0,x,0,x_index);
      } else if (x0>x[x_index+1]) {
	cache=bsearch_inc(x0,x,x_index,n-1);
      }
  
      return cache;
    }
    
    /** \brief Search part of a decreasing vector
     */
    size_t find_dec(const double x0, size_t n, const vec_t &x) {
      size_t x_index=cache;
      
      // Make sure the cached value is not outside the array.
      if (x_index>n-1) x_index=n/2;
  
      if (x0>x[x_index]) {
	cache=bsearch_dec(x0,x,0,x_index);
      } else if (x0<x[x_index+1]) {
	cache=bsearch_dec(x0,x,x_index,n-1);
      }
  
      return cache;
    }

    /** 
	\brief Find the index of x0 in the ordered array \c x 

	This returns the index i for which x[i] is as close
	as possible to x0 if x[i] is either increasing or decreasing.

	If some of the values in the ovector are not finite, then
	the output of this function is not defined.

	If x[i] is non-monotonic, consider using
	ovector_view_tlate::lookup() or uvector_view_tlate::lookup()
	instead of this function.
	
    */
    size_t ordered_lookup(const double x0, size_t n, const vec_t &x) {
      size_t row;
      err_hnd->reset();

      if (x[0]<=x[n-1]) {

	// Increasing case

	if (x0>=x[n-1]) {
	  row=n-1;
	} else { 
	  row=find_inc(x0,n,x);
	  if (row<n-1 && fabs(x[row+1]-x0)<fabs(x[row]-x0)) row++;
	}
    
      } else {

	// Decreasing case
    
	if (x0<=x[n-1]) {
	  row=n-1;
	} else {
	  row=find_dec(x0,n,x);
	  if (row<n-1 && fabs(x[row+1]-x0)<fabs(x[row]-x0)) row++;
	}
      }

      return row;
    }

    /** 
	\brief Find the interval containing x0 in the ordered array \c x 

	This returns the index i for which x[i]<=x0<x[i+1]. 

	If the array is increasing and x0<x[0], then 0 is returned.
	If the array is increasing and x0>x[n-1], then nvar-1 is
	returned (this behavior is slightly different from GSL).  The
	decreasing case is handled analogously.

	If some of the values in the vector are not finite, then the
	output of this function is not defined.

	If x[i] is non-monotonic, consider using
	ovector_view_tlate::lookup() or uvector_view_tlate::lookup()
	instead of this function.
    */
    size_t ordered_interval(const double x0, size_t n, const vec_t &x) {
      size_t row;
      err_hnd->reset();

      if (x[0]<=x[n-1]) {

	// Increasing case

	if (x0>=x[n-1]) row=n-1;
	else row=find_inc(x0,n,x);

      } else {
	// Decreasing case
    
	if (x0<=x[n-1]) row=n-1;
	else row=find_dec(x0,n,x);
      }

      return row;
    }

    /** 
	\brief Binary search a part of an increasing vector

	This function performs a binary search of between
	<tt>x[lo]</tt> and <tt>x[hi-1]</tt>. It returns 
	- \c lo if \c x0 < <tt>x[lo+1]</tt>
	- \c i if <tt>x[i]</tt> <= \c x0 < <tt>x[i+1]</tt> 
	for \c lo <= \c i < \c hi
	- \c hi-1 if \c x0 >= \c <tt>x[hi-1]</tt>
    
	The cache is not used for this function.
     */
    size_t bsearch_inc(const double x0, const vec_t &x,
		       size_t lo, size_t hi) const {
      while (hi>lo+1) {
	size_t i=(hi+lo)/2;
	if (x[i]>x0) {
	  hi=i;
	} else {
	  lo=i;
	}
      }
  
      return lo;
    }

    /** 
	\brief Binary search a part of an decreasing vector
	
	This function performs a binary search of between
	<tt>x[lo]</tt> and <tt>x[hi-1]</tt>. It returns 
	- \c lo if \c x0 > <tt>x[lo+1]</tt>
	- \c i if <tt>x[i]</tt> >= \c x0 > <tt>x[i+1]</tt> 
	for \c lo <= \c i < \c hi
	- \c hi-1 if \c x0 <= \c <tt>x[hi-1]</tt>
    
	The cache is not used for this function.
     */
    size_t bsearch_dec(const double x0, const vec_t &x,
		       size_t lo, size_t hi) const {
      while (hi>lo+1) {
	size_t i=(hi+lo)/2;
	if (x[i]<x0) {
	  hi=i;
	} else {
	  lo=i;
	}
      }
  
      return lo;
    }

  };
  
#ifndef DOXYGENP
}
#endif

#endif
