/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/string_conv.h>
#include <o2scl/err_hnd.h>

using namespace std;
using namespace o2scl;

std::string o2scl::double_to_ieee_string(double *x) {
    
  gsl_ieee_double_rep r;
    
  gsl_ieee_double_to_rep(x, &r);

  std::string s;
    
  char sign=' ';
  if (r.sign==1) sign='-';
    
  switch (r.type)
    {
    case GSL_IEEE_TYPE_NAN:
      s="NaN";
      break ;
    case GSL_IEEE_TYPE_INF:
      s=sign;
      s+="Inf";
      break ;
    case GSL_IEEE_TYPE_NORMAL:
      s=sign;
      s+="1.";
      s+=r.mantissa;
      s+="*2^";
      s+=itos(r.exponent);
      break ;
    case GSL_IEEE_TYPE_DENORMAL:
      s=sign;
      s+="0.";
      s+=r.mantissa;
      s+="*2^";
      s+=itos(r.exponent+1);
      break ;
    case GSL_IEEE_TYPE_ZERO:
      s=sign;
      s+="0";
      break ;
    default:
      s="[non-standard IEEE float]";
    }
  return s;
}
  
string o2scl::itos(int x) {
  string ret;
  ostringstream strout;
  strout.setf(ios::scientific);

  strout << x;

  ret=strout.str();
  return ret;
}

string o2scl::ptos(void *p) {
  string ret;
  ostringstream strout;
  strout.setf(ios::scientific);

  strout << p;

  ret=strout.str();
  return ret;
}

std::string o2scl::dtos(double x, std::ostream &format) {
  string ret;
  ostringstream strout;
  
  strout.precision(format.precision());
  strout.fill(format.fill());
  strout.flags(format.flags());

  strout << x;
    
  ret=strout.str();
  return ret;
}

size_t o2scl::size_of_exponent(double x) {
  string ret;
  ostringstream strout;
  
  strout.setf(ios::scientific);

  strout << x;
    
  ret=strout.str();
  return ret.length()-ret.find('e')-2;
}
  
string o2scl::dtos(double x, int prec, bool auto_prec) {
  string ret;
  ostringstream strout;

  if (prec!=0) {
    if (!auto_prec) strout.setf(ios::scientific);
    strout.precision(prec);
  }

  strout << x;

  ret=strout.str();
  return ret;
}

int o2scl::stoi(string s) {
  int ret;
  istringstream ins(s);
  if (ins >> ret) {
    return ret;
  }
  set_err("Conversion from integer to string failed in o2scl::stoi().",1);
  return 0;
}

double o2scl::stod(string s) {
  double ret;
  istringstream ins(s);
  if (ins >> ret) {
    return ret;
  } 
  set_err("Conversion from string to double failed in o2scl::stod().",1);
  return 0.0;
}

std::string o2scl::double_to_latex(double x, int sigfigs, int ex_min, 
				 int ex_max) {
  if (x==0.0) return "0";
  int sign=1;
  if (x<0) {
    x=-x; 
    sign=-1;
  }
  std::string s;
  int expo=((int)log10(x)), itmp;
  double mant=x/pow(10.0,expo);
  if (mant<1.0) {
    mant*=10.0;
    expo-=1;
  }
  itmp=(int)(mant*pow(10.0,sigfigs)+0.5);
  mant=((double)itmp)/pow(10.0,sigfigs);
  if (expo<=ex_max && expo>=ex_min) {
    int prec=sigfigs;
    if (prec<0) prec=0;
    s=dtos(mant*pow(10.0,expo),prec,true);
  } else {
    s=dtos(mant,sigfigs,true)+" $\\times 10^{"+itos(expo)+"}$";
  }
  if (sign==-1) s=((string)"$-$")+s;
  return s;
}

std::string o2scl::double_to_html(double x, int sigfigs, int ex_min, 
				int ex_max) {
  if (x==0.0) return "0";
  int sign=1;
  if (x<0) {
    x=-x; 
    sign=-1;
  }
  std::string s;
  int expo=((int)log10(x)), itmp;
  double mant=x/pow(10.0,expo);
  if (mant<1.0) {
    mant*=10.0;
    expo-=1;
  }
  itmp=(int)(mant*pow(10.0,sigfigs)+0.5);
  mant=((double)itmp)/pow(10.0,sigfigs);
  if (expo<=ex_max && expo>=ex_min) {
    int prec=(sigfigs);
    if (prec<0) prec=0;
    s=dtos(mant*pow(10.0,expo),prec,true);
  } else {
    s=dtos(mant,(sigfigs),true)+" &times; 10<sup>"+itos(expo)+"</sup>";
  }
  if (sign==-1) s='-'+s;
  return s;
}
