/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TENSOR_H
#define O2SCL_TENSOR_H

/** \file tensor.h
    \brief File for definitions of tensors
*/

#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_ieee_utils.h>

#include <o2scl/err_hnd.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/smart_interp.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Tensor class with arbitrary dimensions

      \todo More complete testing.
      \todo Add const get functions for const references

      \future Could implement arithmetic operators + and - and some
      different products.
      \future Add slicing to get \ref ovector or \ref omatrix objects
  */
  class tensor {

#ifndef DOXYGEN_INTERNAL

  protected:

    // The data
    double *data;

    /// A rank-sized array of the sizes of each dimension
    size_t *size;

    /// Rank
    size_t rk;

#endif

  public:
    
    /// Create an empty tensor with zero rank
    tensor() {
      rk=0;
      data=0;
      size=0;
    }
    
    /** 
	\brief Create a tensor of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor, and will allocate no memory.
    */
    tensor(size_t rank, size_t *dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  rk=0;
	  set_err("Requested zero size in tensor::tensor(size_t, size_t *)",
		  gsl_einval);
	}
      }
      rk=rank;
      size=new size_t[rk];
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	tot*=size[i];
      }
      data=new double[tot];
    }

    virtual ~tensor() {
      if (rk>0) {
	delete[] size;
	delete[] data;
	rk=0;
      }
    }

    /// Set the element indexed by \c index to value \c val
    virtual int set(size_t *index, double val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	set_err("Empty tensor in set().",gsl_einval);
	return 0;
      }
      if (index[0]>=size[0]) {
	set_err("Index greater than size in set().",gsl_index);
	return 0;
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  set_err("Index greater than size in get().",gsl_index);
	  return 0;
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      data[ix]=val;
      return 0;
    }

    /// Get the element indexed by \c index
    virtual double get(size_t *index) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	set_err("Empty tensor in get().",gsl_einval);
	return 0.0;
      }
      if (index[0]>=size[0]) {
	set_err("Index greater than size in get().",gsl_index);
	return 0.0;
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  set_err("Index greater than size in get().",gsl_index);
	  return 0.0;
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      return data[ix];
    }
    
    /** 
	\brief Fix all but one index to create a vector

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix, and returns the
	corresponding vector, whose length is equal to the size of the
	tensor in that index. The value <tt>index[ix]</tt> is ignored.

	For example, for a rank 3 tensor allocated with
	\code
	tensor t;
	size_t dim[3]={3,4,5};
	t.tensor_allocate(3,dim);
	\endcode
	the following code
	\code
	size_t index[3]={1,0,3};
	ovector_view v=t.vector_slice(index,1);
	\endcode
	Gives a vector \c v of length 4 which refers to the values
	<tt>t(1,0,3)</tt>, <tt>t(1,1,3)</tt>, <tt>t(1,2,3)</tt>, and
	<tt>t(1,3,3)</tt>.
    */
    ovector_view vector_slice(size_t ix, size_t *index) {
      size_t start=index[0];
      if (ix==0) start=0;
      for(size_t i=1;i<rk;i++) {
	start*=size[i];
	if (i!=ix) start+=index[i];
      }
      size_t stride=1;
      for(size_t i=ix+1;i<rk;i++) stride*=size[i];
      return ovector_array_stride(size[ix],data+start,stride);
    }

    /** 
	\brief Fix all but two indices to create a matrix

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix and the last index, and
	returns the corresponding matrix, whose size is equal to the
	size of the tensor in the two indices which are not fixed.
    */
    omatrix_view matrix_slice(size_t *index, size_t ix) {
      size_t start=index[0], tot=size[0];
      if (ix==0) start=0;
      for(size_t i=1;i<rk;i++) {
	tot*=size[i];
	start*=size[i];
	if (i!=ix) start+=index[i];
      }
      size_t tda=1;
      if (ix==rk-1) {
	tda=0;
      } else {
	for(size_t i=ix+1;i<rk;i++) tda*=size[i];
      }
      return omatrix_array(tot,data,start,tda,size[ix],size[rk-1]);
    }

    /// Return the rank of the tensor
    virtual int get_rank() { return rk; }

    /** 
	\brief Allocate space for a 
	tensor of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation and previously specified grid data will 
	be lost.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor is left
	unmodified and no deallocation is performed.
    */
    virtual int tensor_allocate(size_t rank, size_t *dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  set_err_ret
	    ("Requested zero size in tensor::tensor(size_t, size_t *)",
	     gsl_einval);
	}
      }
      if (rk>0) tensor_free();
      rk=rank;
      size=new size_t[rk];
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	tot*=size[i];
      }
      data=new double[tot];
      return 0;
    }

    /// Free allocated space (also sets rank to zero)
    virtual int tensor_free() {
      if (rk>0) {
	delete[] size;
	delete[] data;
	rk=0;
      }
      return 0;
    }

    /// Returns the size of the ith index
    virtual size_t get_size(size_t i) { 
      if (i<rk) {
	return size[i]; 
      }
      set_err_ret("Requested index exceeding rank in get_size()",
		  gsl_einval);
    };

    /// Returns the size of the tensor
    virtual size_t total_size() { 
      if (rk==0) return 0;
      size_t tot=1;
      for(size_t i=0;i<rk;i++) tot*=size[i];
      return tot;
    }
    
    /// Pack the indices into a single array index
    size_t pack_indices(size_t *index) {
      if (rk==0) {
	set_err("Empty tensor in pack_indices().",gsl_einval);
	return 0;
      }
      if (index[0]>=size[0]) {
	set_err("Index greater than size in pack_indices().",gsl_index);
	return 0;
      }
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	if (index[i]>=size[i]) {
	  set_err("Index greater than size in pack_indices().",gsl_index);
	  return 0;
	}
	ix*=size[i];
	ix+=index[i];
      }
      return ix;
    }
    
    /// Unpack the single array index into indices
    int unpack_indices(size_t ix, size_t *index) {
      if (ix>total_size()) {
	set_err("Index greater than total size in unpack_indices().",
		gsl_index);
	return 0;
      }
      size_t ix2, sub_size;
      for(size_t i=0;i<rk;i++) {
	if (i==rk-1) {
	  index[i]=ix;
	} else {
	  sub_size=1;
	  for(size_t j=i+1;j<rk;j++) sub_size*=size[j];
	  index[i]=ix/sub_size;
	  // (Remember we're doing integer arithmetic here.)
	  ix-=sub_size*(ix/sub_size);
	}
      }
      return 0;
    }

  };

  /** 
      \brief Tensor class with arbitrary dimensions

      This tensor class allows one to assign the indexes to 
      numerical scales, so that n-dimensional interpolation can
      be performed. To set the grid, use set_grid() and then
      interpolation can be done using interpolate().
      
      \future Only allocate space for grid if it is set
      \future Could implement arithmetic operators + and - and some
      different products.
  */
  template<template<class c_t> class base_interp_t=cspline_interp> 
    class tensor_grid : public tensor {

#ifndef DOXYGEN_INTERNAL

    protected:

    /// A rank-sized set of arrays for the grid points
    double **grd;

    /// If true, the grid has been set by the user
    bool grid_set;
    
#endif

    public:
    
    /// Create an empty tensor with zero rank
    tensor_grid() : tensor() {
      grd=0;
      grid_set=false;
    }
    
    /** 
	\brief Create a tensor of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor, and will allocate no memory.
    */
    tensor_grid(size_t rank, size_t *dim) : tensor(rank,dim) {
      grid_set=false;
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  rk=0;
	  set_err("Requested zero size in tensor::tensor(size_t, size_t *)",
		  gsl_einval);
	}
      }
      grd=new double *[rk];
      for(size_t i=0;i<rk;i++) {
	grd[i]=new double[size[i]];
      }
    }

    virtual ~tensor_grid() {
      if (rk>0) {
	for(size_t i=0;i<rk;i++) delete[] grd[i];
	delete[] size;
	delete[] data;
	rk=0;
      }
    }

    /// Set the element closest to grid point \c grdp to value \c val
    virtual int set_vals(double *grdp, double val) {

      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      data[ix]=val;

      return 0;
    }

    /// Set the element closest to grid point \c grdp to value \c val
    virtual int set_vals(double *grdp, double val, double *closest) {

      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      data[ix]=val;

      return 0;
    }

    /// Get the element closest to grid point \c grdp to value \c val
    virtual double get_vals(double *grdp, double val) {
      
      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);

      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      return data[ix];
    }

    /// Get the element closest to grid point \c grdp to value \c val
    virtual double get_vals(double *grdp, double val, double *closest) {

      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }

      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      return data[ix];
    }

    /** 
	\brief Set the grid

	The parameter \c grid must define the grid, so that \c
	val[i][j] is the jth grid point for the ith index.  The size
	of array \c grid[i] should be given by \c dim[i] where \c dim
	was the argument given in the constructor or to the function
	\ref tensor_allocate().
    */
    virtual int set_grid(double **val) {
      for(size_t i=0;i<rk;i++) {
	for(size_t j=0;j<size[i];j++) {
	  grd[i][j]=val[i][j];
	}
      }
      grid_set=true;
      return 0;
    }

    /** 
	\brief Allocate space for a tensor of rank \c rank with sizes
	given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation and previously specified grid data will 
	be lost.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor is left
	unmodified and no deallocation is performed.
    */
    virtual int tensor_allocate(size_t rank, size_t *dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  set_err_ret
	  ("Requested zero size in tensor::tensor(size_t, size_t *)",
	   gsl_einval);
	}
      }
      if (rk>0) tensor_free();
      rk=rank;
      size=new size_t[rk];
      grd=new double *[rk];
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	grd[i]=new double[size[i]];
	tot*=size[i];
      }
      data=new double[tot];
      return 0;
    }

    /// Free allocated space (also sets rank to zero)
    virtual int tensor_free() {
      if (rk>0) {
	for(size_t i=0;i<rk;i++) delete[] grd[i];
	delete[] size;
	delete[] grd;
	delete[] data;
	rk=0;
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val
    virtual size_t lookup_grid(size_t i, double val) {
      if (i<rk && grid_set) {
	size_t best=0;
	double min=fabs(grd[i][0]-val);
	for(size_t j=0;j<size[i];j++) {
	  if (fabs(grd[i][j]-val)<min) {
	    best=j;
	    min=fabs(grd[i][j]-val);
	  }
	}
	return best;
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val
    virtual double get_grid(size_t i, size_t j) {
      if (i<rk && grid_set) {
	return grd[i][j];
      }
      return 0;
    }

    /// Lookup indices for grid closest to \c val
    virtual int lookup_grid(double* vals, size_t *indices) {
      for(size_t k=0;k<rk;k++) {
	indices[k]=0;
	double min=fabs(grd[k][0]-vals[k]);
	for(size_t j=0;j<size[k];j++) {
	  if (fabs(grd[k][j]-vals[k])<min) {
	    indices[k]=j;
	    min=fabs(grd[k][j]-vals[k]);
	  }
	}
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val, returning the grid point
    virtual size_t lookup_grid_val(size_t i, double val, double &val2) {
      if (i<rk && grid_set) {
	size_t best=0;
	double min=fabs(grd[i][0]-val);
	val2=grd[i][0];
	for(size_t j=0;j<size[i];j++) {
	  if (fabs(grd[i][j]-val)<min) {
	    best=j;
	    min=fabs(grd[i][j]-val);
	    val2=grd[i][j];
	  }
	}
	return best;
      }
      return 0;
    }

    /** 
	\brief Interpolate values \c vals into the tensor, 
	returning the result
	
	This is a quick and dirty implementation of n-dimensional
	interpolation by recursive application of the 1-dimensional
	routine from \ref smart_interp_vec, using the base
	interpolation object specified in the template parameter \c
	base_interp_t. This will be slow for sufficiently large data
	sets.

	\future It should be straightforward to improve the scaling
	of this algorithm significantly by creating a "window" of
	local points around the point of interest. This could
	be done easily by constructing an initial subtensor.
    */
    virtual double interpolate(double *vals) {
      typedef smart_interp_vec<double *,array_const_subvector,
      double *, pointer_alloc<double> > interp_t;
      
      if (rk==1) {
	
	sma_interp_vec<double *> si(size[0],grd[0],data);
	return si.interp(vals[0]);

      } else {
	
	// Get total number of interpolations at this level
	size_t ss=1;
	for(size_t i=1;i<rk;i++) ss*=size[i];

	// Create space for y vectors and interpolators
	double **yvec=new double *[ss];
	interp_t **si=new interp_t *[ss];
	base_interp_t<double *> **it1=new base_interp_t<double *> *[ss];
	base_interp_t<array_const_subvector> **it2=
	new base_interp_t<array_const_subvector> *[ss];
	for(size_t i=0;i<ss;i++) yvec[i]=new double[size[0]];
	
	// Create space for interpolation results
	tensor_grid tdat;
	tdat.tensor_allocate(rk-1,size+1);

	// Set grid for temporary tensor
	tdat.set_grid(grd+1);

	// Create starting coordinate and counter
	size_t *co=new size_t[rk];
	for(size_t i=0;i<rk;i++) co[i]=0;
	size_t cnt=0;

	// Loop over every interpolation
	bool done=false;
	while(done==false) {

	  // Fill yvector with the appropriate data
	  for(size_t i=0;i<size[0];i++) {
	    co[0]=i;
	    yvec[cnt][i]=get(co);
	  }
	  
	  it1[cnt]=new base_interp_t<double *>;
	  it2[cnt]=new base_interp_t<array_const_subvector>;
	  si[cnt]=new interp_t(*it1[cnt],*it2[cnt],size[0],grd[0],yvec[cnt]);
	  
	  tdat.set(co+1,si[cnt]->interp(vals[0]));

	  // Go to next interpolation
	  cnt++;
	  co[rk-1]++;
	  // carry if necessary
	  for(int j=rk-1;j>0;j--) {
	    if (co[j]>=size[j]) {
	      co[j]=0;
	      co[j-1]++;
	    }
	  }

	  // Test if done
	  if (cnt==ss) done=true;

	  // End of while loop
	}
      
	// Now call the next level of interpolation
	double res=tdat.interpolate(vals+1);
	
	tdat.tensor_free();
	for(size_t i=0;i<ss;i++) {
	  delete[] yvec[i];
	  delete si[i];
	  delete it1[i];
	  delete it2[i];
	}
	delete[] co;
	delete[] si;
	delete[] it1;
	delete[] it2;
	delete[] yvec;

	return res;
      }
    }
  };

  /** 
      \brief Rank 1 tensor
  */
  class tensor1 : public tensor {
  public:
    
    /// Create an empty tensor
    tensor1() : tensor() {}

      /// Create a rank 1 tensory of size \c sz
      tensor1(size_t sz) : tensor(1,&sz) {}
	
    /// Get the element indexed by \c index
	virtual double get(size_t *index) { 
	  return tensor::get(index); 
	}

	/// Set the element indexed by \c index to value \c val
	virtual int set(size_t *index, double val) 
	{ return tensor::set(index,val); }
	
	/// Get the element indexed by \c ix
	virtual double get(size_t ix) { return tensor::get(&ix); }
	
	/// Set the element indexed by \c index to value \c val
	virtual int set(size_t index, double val) 
	{ return tensor::set(&index,val); }
    
	/// Get an element using array-like indexing
	virtual double &operator[](size_t ix) { return this->data[ix]; }
    
	/// Get an element using operator()
	virtual double &operator()(size_t ix) { return this->data[ix]; }
  };
  
  /**
     \brief Rank 2 tensor
  */
  class tensor2 : public tensor {
  public:

    /// Create an empty tensor
    tensor2() : tensor() {}

      /// Create a rank 2 tensor of size \c (sz,sz2)
      tensor2(size_t sz, size_t sz2) : tensor() {
	this->rk=2;
	this->size=new size_t[2];
	this->size[0]=sz;
	this->size[1]=sz2;
	size_t tot=sz*sz2;
	this->data=new double[tot];
      }
	
    /// Get the element indexed by \c index
	virtual double get(size_t *index) { 
	  return tensor::get(index); 
	}

	/// Set the element indexed by \c index to value \c val
	virtual int set(size_t *index, double val) 
	{ return tensor::set(index,val); }
	

	/// Get the element indexed by \c (ix1,ix2)
	virtual double get(size_t ix1, size_t ix2) { 
	  size_t sz[2]={ix1,ix2};
	  return tensor::get(sz); 
	}

	/// Set the element indexed by \c (ix1,ix2) to value \c val
	virtual int set(size_t ix1, size_t ix2, double val) {
	  size_t sz[2]={ix1,ix2};
	  return tensor::set(sz,val); 
	}

	/// Get the element indexed by \c (ix1,ix2)
	virtual double &operator()(size_t ix, size_t iy) 
	{ return this->data[ix*this->size[1]+iy]; }
  };
  
  /**
     \brief Rank 3 tensor
  */
  class tensor3 : public tensor {
  public:

    /// Create an empty tensor
    tensor3() : tensor() {}

      /// Create a rank 3 tensor of size \c (sz,sz2,sz3)
      tensor3(size_t sz, size_t sz2, size_t sz3) : tensor() {
	this->rk=3;
	this->size=new size_t[3];
	this->size[0]=sz;
	this->size[1]=sz2;
	this->size[2]=sz3;
	size_t tot=sz*sz2*sz3;
	this->data=new double[tot];
      }
	
	/// Get the element indexed by \c index
	virtual double get(size_t *index) { 
	  return tensor::get(index); 
	}

	/// Set the element indexed by \c index to value \c val
	virtual int set(size_t *index, double val) 
	{ return tensor::set(index,val); }
	
	/// Get the element indexed by \c (ix1,ix2,ix3)
	virtual double get(size_t ix1, size_t ix2, size_t ix3) { 
	  size_t sz[3]={ix1,ix2,ix3};
	  return tensor::get(sz); 
	}

	/// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
	virtual int set(size_t ix1, size_t ix2, size_t ix3, double val) {
	  size_t sz[3]={ix1,ix2, ix3};
	  return tensor::set(sz,val); 
	}
  };
  
  /**
     \brief Rank 3 tensor with a grid
  */
  template<template<class c_t> class interp_t=cspline_interp> 
    class tensor_grid3 : public tensor_grid<interp_t> {
     
    public:
     
    /// Create an empty tensor
    tensor_grid3() : tensor_grid<interp_t>() {}

    /// Create a rank 3 tensor of size \c (sz,sz2,sz3)
    tensor_grid3(size_t sz, size_t sz2, size_t sz3) : tensor_grid<interp_t>() {
      this->rk=3;
      this->size=new size_t[3];
      this->grd=new double *[3];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->grd[0]=new double[sz];
      this->grd[1]=new double[sz2];
      this->grd[2]=new double[sz3];
      size_t tot=sz*sz2*sz3;
      this->data=new double[tot];
      this->grid_set=false;
    }
   
    virtual ~tensor_grid3() {
      if (this->rk>0) {
	for(size_t i=0;i<this->rk;i++) {
	  delete[] this->grd[i];
	}
	delete[] this->size;
	delete[] this->grd;
	delete[] this->data;
	this->rk=0;
      }
    }
   
	/// Get the element indexed by \c index
    virtual double get(size_t *index) { 
      return tensor_grid<interp_t>::get(index); 
    }

    /// Set the element indexed by \c index to value \c val
    virtual int set(size_t *index, double val) 
    { return tensor_grid<interp_t>::set(index,val); }
	
    /// Get the element indexed by \c (ix1,ix2,ix3)
    virtual double get(size_t ix1, size_t ix2, size_t ix3) { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_grid<interp_t>::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
    virtual int set(size_t ix1, size_t ix2, size_t ix3, double val) {
      size_t sz[3]={ix1,ix2, ix3};
      return tensor_grid<interp_t>::set(sz,val); 
    }
  };
  
  /**
     \brief Rank 4 tensor
  */
  class tensor4 : public tensor {

  public:

    /// Create an empty tensor
    tensor4() : tensor() {}

      /// Create a rank 4 tensor of size \c (sz,sz2,sz3,sz4)
      tensor4(size_t sz, size_t sz2, size_t sz3, size_t sz4) : 
	tensor() {
	this->rk=4;
	this->size=new size_t[4];
	this->size[0]=sz;
	this->size[1]=sz2;
	this->size[2]=sz3;
	this->size[3]=sz4;
	size_t tot=sz*sz2*sz3*sz4;
	this->data=new double[tot];
      }
	
	/// Get the element indexed by \c index
	virtual double get(size_t *index) { 
	  return tensor::get(index); 
	}

	/// Set the element indexed by \c index to value \c val
	virtual int set(size_t *index, double val) 
	{ return tensor::set(index,val); }
	
	/// Get the element indexed by \c (ix1,ix2,ix3,ix4)
	virtual double get(size_t ix1, size_t ix2, size_t ix3, size_t ix4) { 
	  size_t sz[4]={ix1,ix2,ix3,ix4};
	  return tensor::get(sz); 
	}

	/// Set the element indexed by \c (ix1,ix2,ix3,ix4) to value \c val
	virtual int set(size_t ix1, size_t ix2, size_t ix3, size_t ix4, 
			double val) {
	  size_t sz[4]={ix1,ix2,ix3,ix4};
	  return tensor::set(sz,val); 
	}
  };

#ifndef DOXYGENP
}
#endif

#endif



