/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CERN_DERIV_H
#define O2SCL_CERN_DERIV_H

#include <o2scl/deriv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Numerical differentiation routine (CERNLIB)

      This uses Romberg extrapolation to compute the 
      derivative with the finite-differencing formula
      \f[
      f^{\prime}(x) = [f(x+h)-f(x-h)]/(2h)
      \f]

      If \ref root::verbose is greater than zero, then each iteration
      prints out the extrapolation table, and if \ref root::verbose
      is greater than 1, then a keypress is required at the 
      end of each iteration. 

      Based on the CERNLIB routine DERIV(), which was 
      based on \ref Rutishauser63 .
      
      \note Second and third derivatives are computed by naive nested
      applications of the formula for the first derivative and the
      uncertainty for these will likely be underestimated.
      
      \comment
      - Maybe we should consider moving the table size to a template 
      parameter?
      \endcomment
  */
  template<class param_t, class func_t>
    class cern_deriv : public deriv<param_t,func_t> {

    public:
  
    /// A scaling factor (default 1.0)
    double delta;
  
    /// Extrapolation tolerance (default is \f$ 5 \times 10^{14} \f$)
    double eps;

    cern_deriv() {

      dx[0]=0.0256;
      dx[1]=0.0192;
      dx[2]=0.0128;
      dx[3]=0.0096;
      dx[4]=0.0064;
      dx[5]=0.0048;
      dx[6]=0.0032;
      dx[7]=0.0024;
      dx[8]=0.0016;
      dx[9]=0.0012;
  
      w[1][1]=1.3333333333333333;
      w[3][1]=1.0666666666666667;
      w[5][1]=1.0158730158730159;
      w[7][1]=1.0039215686274510;
  
      w[2][1]=3.3333333333333333e-1;
      w[4][1]=6.6666666666666667e-2;
      w[6][1]=1.5873015873015873e-2;
      w[8][1]=3.9215686274509804e-3;

      w[0][2]=2.2857142857142857;
      w[2][2]=1.1636363636363636;
      w[4][2]=1.0364372469635628;
      w[6][2]=1.0088669950738916;
      w[8][2]=1.0022021042329337;

      w[1][2]=1.2857142857142857;
      w[3][2]=1.6363636363636364e-1;
      w[5][2]=3.6437246963562753e-2;
      w[7][2]=8.8669950738916256e-3;
      w[9][2]=2.2021042329336922e-3;
  
      w[0][3]=1.8000000000000000;
      w[2][3]=1.1250000000000000;
      w[4][3]=1.0285714285714286;
      w[6][3]=1.0069930069930070;
      w[8][3]=1.0017391304347826;
  
      w[1][3]=8.0000000000000000e-1;
      w[3][3]=1.2500000000000000e-1;
      w[5][3]=2.8571428571428571e-2;
      w[7][3]=6.9930069930069930e-3;
      w[9][3]=1.7391304347826087e-3;
  
      delta=1.0;
      eps=5.0e-14;
    }
    
    virtual int calc_err(double x, param_t &pa, func_t &func,
			 double &dfdx, double &err) {
      
      double t[10][10], a[10], del, hh;
      bool lev[10]={1,0,1,0,1,0,1,0,1,0}, lmt;
      int is, k, m;

      err_hnd->reset();

      del=10.0*fabs(delta);
      is=10;
  
      do {
	is--;
	del=del/10.0;

	if (is==0 || x+del*dx[9]==x) {
	  delta=del;
	  dfdx=0.0;
	  err=0.0;
	  set_err_ret
	    ("Calculation of derivative failed in cern_deriv::calc().",
	     gsl_efailed);
	}
    
	for(k=0;k<=9;k++) {
	  hh=del*dx[k];
	  t[k][0]=(func(x+hh,pa)-func(x-hh,pa))/(hh+hh);
	  a[k]=t[k][0];
	}
    
	if (a[0]>=a[9]) {
	  for(k=0;k<=9;k++) a[k]=-a[k];
	}
    
	lmt=true;
	for(k=1;k<=9;k++) {
	  hh=a[k-1]-a[k];
	  lmt=(lmt && (hh<=0.0 || fabs(hh)<=eps*fabs(a[k])));
	}
	
	if (this->verbose>0) {
	  std::cout << "cern_deriv, iteration: " << 10-is << std::endl;
	  std::cout << "(hh, a[k], derivative) list: " 
		    << std::endl;
	  for(k=1;k<=9;k++) {
	    std::cout << a[k-1]-a[k] << " " << eps*fabs(a[k]) << " "
		      << t[k][0] << std::endl;
	  }
	  std::cout << "Converged: " << lmt << std::endl;
	  if (this->verbose>1) {
	    char ch;
	    std::cin >> ch;
	  }
	}

      } while (lmt==false);
  
      for(m=1;m<=9;m++) {
	for(k=0;k<=9-m;k++) {
	  if (lev[m]) {
	    t[k][m]=w[m-1][1]*t[k+1][m-1]-w[m][1]*t[k][m-1];
	  } else if (lev[k]) {
	    t[k][m]=w[m-1][2]*t[k+1][m-1]-w[m][2]*t[k][m-1];
	  } else {
	    t[k][m]=w[m-1][3]*t[k+1][m-1]-w[m][3]*t[k][m-1];
	  }
	}
      }
      dfdx=t[0][9];
      if (dfdx!=0.0) err=(dfdx-t[0][8])/dfdx;
      else err=0.0;
      delta=del;
  
      return 0;
    }
 
    virtual int calc_err_int
      (double x, typename deriv<param_t,func_t>::dpars &pa, 
       funct<typename deriv<param_t,func_t>::dpars> &func, 
       double &dfdx, double &err) {
      
      double t[10][10], a[10], del, hh;
      bool lev[10]={1,0,1,0,1,0,1,0,1,0}, lmt;
      int is, k, m;

      err_hnd->reset();

      del=10.0*fabs(delta);
      is=10;
  
      do {
	is--;
	del=del/10.0;
	if (is==0 || x+del*dx[9]==x) {
	  delta=del;
	  set_err("Calculation of derivative failed in cern_deriv::calc().",
		  gsl_efailed);
	  dfdx=0.0;
	  err=0.0;
	  return 1;
	}
    
	for(k=0;k<=9;k++) {
	  hh=del*dx[k];
	  t[k][0]=(func(x+hh,pa)-func(x-hh,pa))/(hh+hh);
	  a[k]=t[k][0];
	}
    
	if (a[0]>=a[9]) {
	  for(k=0;k<=9;k++) a[k]=-a[k];
	}
    
	lmt=true;
	for(k=1;k<=9;k++) {
	  hh=a[k-1]-a[k];
	  lmt=(lmt && (hh<=0.0 || fabs(hh)<=eps*fabs(a[k])));
	}

      } while (lmt==false);
  
      for(m=1;m<=9;m++) {
	for(k=0;k<=9-m;k++) {
	  if (lev[m]) {
	    t[k][m]=w[m-1][1]*t[k+1][m-1]-w[m][1]*t[k][m-1];
	  } else if (lev[k]) {
	    t[k][m]=w[m-1][2]*t[k+1][m-1]-w[m][2]*t[k][m-1];
	  } else {
	    t[k][m]=w[m-1][3]*t[k+1][m-1]-w[m][3]*t[k][m-1];
	  }
	}
      }
      dfdx=t[0][9];
      if (dfdx!=0.0) err=(dfdx-t[0][8])/dfdx;
      else err=0.0;
      delta=del;
  
      return 0;
    }
 
    /// Return string denoting type ("cern_deriv")
    virtual const char *type() { return "cern_deriv"; }

    protected:

#ifndef DOXYGEN_INTERNAL
    
    /// \name Storage for the fixed coefficients
    //@{
    double dx[10];
    double w[10][4];
    //@}

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif


