/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/hadronic_eos.h>

using namespace std;
using namespace o2scl;

hadronic_eos::hadronic_eos() {

  sat_deriv=&def_deriv;
  sat_deriv2=&def_deriv2;
  def_deriv.h=1.0e-3;
  def_deriv2.h=1.0e-3;
  
  def_neutron.init(939.0/o2scl_const::hc_mev_fm,2.0);
  def_proton.init(939.0/o2scl_const::hc_mev_fm,2.0);
  def_neutron.non_interacting=false;
  def_proton.non_interacting=false;
  neutron=&def_neutron;
  proton=&def_proton;
  
  eos_mroot=&def_mroot;
  sat_root=&def_sat_root;
  
  provides=0;
}

int hadronic_eos::calc_e(fermion &n, fermion &p, thermo &th) {
  int ret;

  if (provides & provides_calc_temp_e > 0) {
    
    return calc_temp_e(n,p,0.0,th);
    
  } else if ((provides & provides_calc_p > 0) || 
	     (provides & provides_calc_temp_p > 0)) {
    
    set_n_and_p(n,p);
    set_thermo(th);
    
    denpar dp={n.n,p.n,0.0};
    
    ovector mu(2);
    mu[0]=n.mu;
    mu[1]=p.mu;
    
    if (mu[1]<n.ms) mu[0]=n.ms+1.0e-4;
    if (mu[2]<p.ms) mu[1]=p.ms+1.0e-4;
    
    void *vpx=(void *)&dp;
    mm_funct_mfptr<hadronic_eos,void *> 
      fmf(this,&hadronic_eos::nuc_matter_p);
    ret=eos_mroot->msolve(2,mu,vpx,fmf);
    
    if (ret!=0) {
      add_err_ret("Solver failed in hadronic_eos::calc_p().",ret);
    }
    
    th=*eos_thermo;
    
  } else {
    
    set_err_ret("No EOS provided in hadronic_eos::calc_e().",gsl_nobase);

  }


  return 0;
}

int hadronic_eos::calc_temp_e(fermion &n, fermion &p, 
			      const double T, thermo &th) {
  int ret;

  if (provides & provides_calc_temp_e > 0) {
    
    return calc_temp_e(n,p,0.0,th);
    
  } else if ((provides & provides_calc_p > 0) || 
	     (provides & provides_calc_temp_p > 0)) {
    
    set_n_and_p(n,p);
    set_thermo(th);
    
    denpar dp={n.n,p.n,0.0};
    
    ovector mu(2);
    mu[0]=n.mu;
    mu[1]=p.mu;
    
    if (mu[1]<n.ms) mu[0]=n.ms+1.0e-4;
    if (mu[2]<p.ms) mu[1]=p.ms+1.0e-4;
    
    void *vpx=(void *)&dp;
    mm_funct_mfptr<hadronic_eos,void *> 
      fmf(this,&hadronic_eos::nuc_matter_p);
    ret=eos_mroot->msolve(2,mu,vpx,fmf);
    
    if (ret!=0) {
      add_err_ret("Solver failed in hadronic_eos::calc_p().",ret);
    }
    
    th=*eos_thermo;

  } else {

    set_err_ret("No EOS provided in hadronic_eos::calc_e().",gsl_nobase);

  }


  return 0;
}

int hadronic_eos::calc_p(fermion &n, fermion &p, thermo &th) {
  int ret;
  
  if (provides & provides_calc_temp_p > 0) {
    
    return calc_temp_p(n,p,0.0,th);
    
  } else if ((provides & provides_calc_e > 0) || 
	     (provides & provides_calc_temp_e > 0)) {
    
    set_n_and_p(n,p);
    set_thermo(th);
    
    mupar mp={n.mu,p.mu,0.0};
    
    ovector x(2);
    x[0]=n.n;
    x[1]=p.n;
    
    void *vpx=(void *)&mp;
    mm_funct_mfptr<hadronic_eos,void *> 
      fmf(this,&hadronic_eos::nuc_matter_e);
    ret=eos_mroot->msolve(2,x,vpx,fmf);
    if (ret!=0) {
      add_err_ret("Solver failed in hadronic_eos::calc_p().",ret);
    }
    
    th=*eos_thermo;

  } else {
    set_err_ret("No EOS provided in hadronic_eos::calc_p().",gsl_nobase);
  }

  return 0;
}

int hadronic_eos::calc_temp_p(fermion &n, fermion &p, double T, thermo &th) {
  int ret;
  
  if ((T<=0.0 && provides & provides_calc_e > 0) || 
      (provides & provides_calc_temp_e > 0)) {
    
    set_n_and_p(n,p);
    set_thermo(th);
    
    mupar mp={n.mu,p.mu,T};
    
    ovector x(2);
    x[0]=n.n;
    x[1]=p.n;

    void *vpx=(void *)&mp;
    mm_funct_mfptr<hadronic_eos,void *> 
      fmf(this,&hadronic_eos::nuc_matter_e);
    ret=eos_mroot->msolve(2,x,vpx,fmf);
    if (ret!=0) {
      add_err_ret("Solver failed in hadronic_eos::calc_temp_p().",ret);
    }
    
    th=*eos_thermo;

  } else {
    set_err_ret("No EOS provided in hadronic_eos::calc_temp_p().",gsl_nobase);
  }

  return 0;
}

double hadronic_eos::fcomp(const double nb) {
  double lcomp, err;
  int ret;
  
  funct_mfptr_noerr<hadronic_eos,void *> 
    fmn(this,&hadronic_eos::calc_pressure);
  void *vpx=(void *)this;
  ret=sat_deriv->calc_err(nb,vpx,fmn,lcomp,err);
  if (ret!=0) {
    add_err("Derivative failed in hadronic_eos::fcomp().",gsl_efailed);
  }

  lcomp*=9.0;
  
  return lcomp;
}

double hadronic_eos::feoa(const double nb, const double pf) {
  double leoa;

  n_baryon=nb;
  proton->n=n_baryon*pf;
  neutron->n=n_baryon-proton->n;
  int ret=calc_e(*neutron,*proton,*eos_thermo);
  if (ret!=0) {
    add_err("Function calc_e() failed in hadronic_eos::feoa().",gsl_efailed);
  }
  
  leoa=eos_thermo->ed/n_baryon-(1.0-pf)*neutron->m-pf*proton->m;
  
  return leoa;
}

double hadronic_eos::fesym(const double nb, const double pf) {
  double lesym, err;
  int ret;

  n_baryon=nb;

  funct_mfptr_noerr<hadronic_eos,void *> 
    fmn(this,&hadronic_eos::calc_edensity);
  void *vpx=(void *)this;
  ret=sat_deriv->calc2_err(1.0-2.0*pf,vpx,fmn,lesym,err);
  lesym/=2.0*n_baryon;
  if (ret!=0) {
    add_err("Derivative failed in hadronic_eos::fesym().",gsl_efailed);
  }
  
  return lesym;
}

double hadronic_eos::calc_esym(double lnb, bool &alt) {
  if (alt) return fesym_diff(lnb);
  else return fesym(lnb);
}

double hadronic_eos::fesym_slope(const double nb, bool altsym) {
  double lslope, err;
  int ret;
  
  funct_mfptr_noerr<hadronic_eos,bool> fmn(this,&hadronic_eos::calc_esym);
  
  ret=sat_deriv2->calc_err(nb,altsym,fmn,lslope,err);
  lslope*=3.0*n_baryon;
  if (ret!=0) {
    add_err("Derivative failed in hadronic_eos::fesym().",gsl_efailed);
  }
  
  return lslope;
}

double hadronic_eos::fesym_diff(const double nb) {
  double eoa_neut, eoa_nuc;
  int ret1, ret2;

  neutron->n=nb;
  proton->n=0.0;
  ret1=calc_e(*neutron,*proton,*eos_thermo);
  eoa_neut=eos_thermo->ed/nb-neutron->m;
  
  neutron->n=nb/2.0;
  proton->n=nb/2.0;
  ret2=calc_e(*neutron,*proton,*eos_thermo);
  eoa_nuc=eos_thermo->ed/nb-(neutron->m+proton->m)/2.0;
  
  if (ret1!=0 || ret2!=0) {
    add_err("Function calc_e() failed in hadronic_eos::fesym_diff().",
	    gsl_efailed);
  }

  return eoa_neut-eoa_nuc;
}

double hadronic_eos::fkprime(const double nb) {
  double lkprime, err;
  int ret;

  funct_mfptr_noerr<hadronic_eos,void *> 
    fmn(this,&hadronic_eos::calc_press_on2);
  void *vpx=(void *)this;
  ret=sat_deriv->calc2_err(nb,vpx,fmn,lkprime,err);
  lkprime*=27.0*nb*nb*nb;
  if (ret!=0) {
    add_err("Second derivative failed in hadronic_eos::fkprime().",
	    gsl_efailed);
  }
  
  return lkprime;
}

double hadronic_eos::fmsom(const double nb, const double pf) {
  
  n_baryon=nb;
  proton->n=n_baryon*pf;
  neutron->n=n_baryon-proton->n;
  int ret=calc_e(*neutron,*proton,*eos_thermo);
  if (ret!=0) {
    add_err("Function calc_e() failed in hadronic_eos::fmsom().",
	    gsl_efailed);
  }

  return neutron->ms/neutron->m;
}

double hadronic_eos::fn0(double protfrac, double &leoa) {
  double x, y;
  int ret;
  
  proton_frac=protfrac;

  x=0.16;
  
  void *vpx=(void *)this;
  funct_mfptr<hadronic_eos,void *> 
    fmf(this,&hadronic_eos::saturation_matter_e);
  
  fmf(x,y,vpx);
  
  ret=sat_root->solve(x,vpx,fmf);
  saturation_matter_e(x,y,vpx);
  
  if (ret!=0) {
    add_err("Solver failed in hadronic_eos::fn0().",ret);
  }
  
  leoa=eos_thermo->ed/x-(neutron->m+proton->m)/2.0;
  
  return x;
}

double hadronic_eos::fsprime(const double nb, const double pf) {
  double lsprime,err;
  int ret;
  
  n_baryon=nb;
  proton_frac=pf;
  
  funct_mfptr_noerr<hadronic_eos,void *> 
    fmn(this,&hadronic_eos::calc_esym);
  void *vpx=(void *)this;
  ret=sat_deriv->calc_err(nb,vpx,fmn,lsprime,err);
  lsprime*=nb;
  
  if (ret!=0) {
    add_err("Derivative failed in hadronic_eos::fsprime().",gsl_efailed);
  }
  
  return lsprime;
}

int hadronic_eos::saturation() {
  n0=fn0(0.5,eoa);
  comp=fcomp(n0);
  esym=fesym(n0);
  msom=fmsom(n0);
  kprime=fkprime(n0);
  sprime=fsprime(n0);

  return gsl_success;
}

int hadronic_eos::gradient_qij(fermion &n, fermion &p, thermo &th,
			       double &qnn, double &qnp, double &qpp, 
			       double &dqnndnn, double &dqnndnp,
			       double &dqnpdnn, double &dqnpdnp,
			       double &dqppdnn, double &dqppdnp) {
  double nn=n.n, np=p.n, t1=0.0, t2=0.0, barn=nn+np, den;

  void *vpx=0;

  t1=t1_fun(barn,vpx);
  t2=t2_fun(barn,vpx);
    
  funct_mfptr_noerr<hadronic_eos,void *> t1fun(this,&hadronic_eos::t1_fun);
  funct_mfptr_noerr<hadronic_eos,void *> t2fun(this,&hadronic_eos::t2_fun);

  set_n_and_p(n,p);
  set_thermo(th);
  double dt1=sat_deriv->calc(barn,vpx,t1fun);
  double ddt1=sat_deriv->calc2(barn,vpx,t1fun);
  double dt2=sat_deriv->calc(barn,vpx,t2fun);
  
  qnn=0.0625*(3.0*t1-3.0*t2+2.0*(2.0*barn-nn)*dt1);
  qnp=0.0625*(6.0*t1-2.0*t2+3*barn*dt1);
  qpp=0.0625*(3.0*t1-3.0*t2+2.0*(2.0*barn-np)*dt1);
  
  dqnndnn=0.0625*(3.0*dt1-3.0*dt2+2.0*dt1+2.0*(2.0*barn-nn)*ddt1);
  dqnndnp=0.0625*(3.0*dt1-3.0*dt2+4.0*dt1+2.0*(2.0*barn-nn)*ddt1);
  dqnpdnn=0.0625*(6.0*dt1-2.0*dt2+3.0*barn*ddt1+3.0*dt1);
  dqnpdnp=0.0625*(6.0*dt1-2.0*dt2+3.0*barn*ddt1+3.0*dt1);
  dqppdnn=0.0625*(3.0*dt1-3.0*dt2+4.0*dt1+2.0*(2.0*barn-np)*ddt1);
  dqppdnp=0.0625*(3.0*dt1-3.0*dt2+2.0*dt1+2.0*(2.0*barn-np)*ddt1);

  return 0;
}

double hadronic_eos::t1_fun(double barn, void *&vp) {
  double xp=proton->n/(neutron->n+proton->n);
  neutron->n=(1.0-xp)*barn;
  proton->n=xp*barn;
  calc_e(*neutron,*proton,*eos_thermo);
  double den=neutron->m*proton->m*neutron->ms*proton->ms*barn*
    (neutron->n-proton->n);
  return ((proton->m-proton->ms)*neutron->m*neutron->ms*(neutron->n+2*barn)+
	  (neutron->ms-neutron->m)*proton->ms*proton->m*
	  (proton->n+2*barn))/den;
}

double hadronic_eos::t2_fun(double barn, void *&vp) {
  double xp=proton->n/(neutron->n+proton->n);
  neutron->n=(1.0-xp)*barn;
  proton->n=xp*barn;
  calc_e(*neutron,*proton,*eos_thermo);
  double den=neutron->m*proton->m*neutron->ms*proton->ms*barn*
    (neutron->n-proton->n);
  return ((proton->ms-proton->m)*neutron->m*neutron->ms*(2*barn-neutron->n)+
	  (neutron->m-neutron->ms)*proton->m*proton->ms*
	  (2*barn-proton->n))/den;
}

double hadronic_eos::calc_pressure(double nb, void *&pa) {
  
  neutron->n=nb/2.0;
  proton->n=nb/2.0;
  
  calc_e(*neutron,*proton,*eos_thermo);
  
  return eos_thermo->pr;
}

double hadronic_eos::calc_press_on2(double nb, void *&pa) {
  
  neutron->n=nb/2.0;
  proton->n=nb/2.0;

  calc_e(*neutron,*proton,*eos_thermo);

  return eos_thermo->pr/nb/nb;
}

double hadronic_eos::calc_edensity(double delta, void *&pa) {
  double pf=(1.0-delta)/2.0;
  
  neutron->n=(1.0-pf)*n_baryon;
  proton->n=pf*n_baryon;
  
  calc_e(*neutron,*proton,*eos_thermo);

  return eos_thermo->ed;
}

double hadronic_eos::calc_esym(double nb, void *&pa) {
  return fesym(nb,proton_frac);
}

int hadronic_eos::saturation_matter_e(double x, double &y, void *&pa) {

  neutron->n=x*(1.0-proton_frac);
  proton->n=x*proton_frac;
  
  int ret=calc_e(*neutron,*proton,*eos_thermo);
  if (ret!=0) {
    add_err("Function calc_e() failed in hadronic_eos::saturation_matter_e",
	    gsl_efailed);
  }
  
  y=eos_thermo->pr;

  return gsl_success;
}

int hadronic_eos::nuc_matter_e(size_t nv, const ovector_view &x, 
			       ovector_view &y, 
				 void *&pa) {
  mupar *mp=(mupar *)pa;
  
  neutron->n=x[0];
  proton->n=x[1];
  
  if (!finite(neutron->n) || !finite(proton->n)) {
    cout << "Density problem in heos::nuc_matter_e()." << endl;
    exit(-1);
  }

  int ret;
  if (mp->T>0.0) {
    ret=calc_temp_e(*neutron,*proton,mp->T,*eos_thermo);
  } else {
    ret=calc_e(*neutron,*proton,*eos_thermo);
  }

  if (ret!=0) {
    add_err("Function calc_e() failed in hadronic_eos::nuc_matter_e().",
	    gsl_efailed);
  }

  y[0]=neutron->mu-mp->mun;
  y[1]=proton->mu-mp->mup;
  
  if (!finite(neutron->mu) || !finite(proton->mu)) {
    cout << "mu problem in heos::nuc_matter_e()." << endl;
    exit(-1);
  }

  return ret;
}

int hadronic_eos::nuc_matter_p(size_t nv, const ovector_view &x, 
			       ovector_view &y, 
			       void *&pa) {
  
  denpar *dp=(denpar *)pa;
  
  neutron->mu=x[0];
  proton->mu=x[1];

  int ret;
  if (dp->T>0.0) {
    ret=calc_temp_p(*neutron,*proton,dp->T,*eos_thermo);
  } else {
    ret=calc_p(*neutron,*proton,*eos_thermo);
  }

  if (ret!=0) {
    add_err("calc_p() failed in hadronic_eos::nuc_matter_p().",ret);
  }

  y[0]=neutron->n-dp->nn;
  y[1]=proton->n-dp->np;
  
  return ret;
}

int hadronic_eos::set_n_and_p(fermion &n, fermion &p) {
  neutron=&n;
  proton=&p;
  return gsl_success;
}

int hadronic_eos::set_mroot(mroot<void *,mm_funct<void *> > &mr) {
  eos_mroot=&mr;
  return gsl_success;
}

int hadronic_eos::set_sat_root(root<void *,funct<void *> > &mr) {
  sat_root=&mr;
  return gsl_success;
}

int hadronic_eos::set_sat_deriv(deriv<void *,funct<void *> > &de) {
  sat_deriv=&de;
  return gsl_success;
}

int hadronic_eos::set_sat_deriv2(deriv<bool,funct<bool> > &de) {
  sat_deriv2=&de;
  return gsl_success;
}

template<> int io_tlate<hadronic_eos>::input
(cinput *co, in_file_format *ins, hadronic_eos *he) {
  /*
  std::string t;
  co->object_in("eos",ins,he,t);
  bool sw;
  string tmp;
  ins->bool_in(sw,"mroot");
  if (sw) pointer_in(co,ins,((void **)(&he->eos_mroot)),tmp);
  ins->bool_in(sw,"root");
  if (sw) pointer_in(co,ins,((void **)(&he->sat_root)),tmp);
  ins->bool_in(sw,"deriv");
  if (sw) pointer_in(co,ins,((void **)(&he->sat_deriv)),tmp);
  ins->bool_in(sw,"n_and_p");
  if (sw) {
    pointer_in(co,ins,((void **)(&he->neutron)),tmp);
    pointer_in(co,ins,((void **)(&he->proton)),tmp);
  }
  ins->double_in(he->eoa,"eoa");
  ins->double_in(he->n0,"n0");
  ins->double_in(he->comp,"comp");
  ins->double_in(he->esym,"esym");
  ins->double_in(he->msom,"msom");
  ins->double_in(he->kprime,"kprime");
  ins->double_in(he->sprime,"sprime");
  */
  return 0;
}

template<> int io_tlate<hadronic_eos>::output
(coutput *co, out_file_format *outs, hadronic_eos *he) {
  /*
  std::string t;
  co->object_out("eos",outs,he,0,0,t);
  if (he->eos_mroot!=&he->def_mroot) {
    outs->bool_out(true,"mroot");
    //pointer_out(co,outs,he->eos_mroot,he->eos_mroot->type());
  } else {
    outs->bool_out(false,"mroot");
  }
  if (he->sat_root!=&he->def_sat_root) {
    outs->bool_out(true,"sat_root");
    //pointer_out(co,outs,he->sat_root,he->sat_root->type());
  } else {
    outs->bool_out(false,"sat_root");
  }
  if (he->sat_deriv!=&he->def_deriv) {
    outs->bool_out(true,"deriv");
    //pointer_out(co,outs,he->sat_mroot,he->sat_mroot->type());
  } else {
    outs->bool_out(false,"deriv");
  }
  if (he->neutron!=&he->def_neutron || he->proton!=&he->def_proton) {
    outs->bool_out(true,"n_and_p");
    pointer_out(co,outs,he->neutron,he->neutron->type());
    pointer_out(co,outs,he->proton,he->proton->type());
  } else {
    outs->bool_out(false,"n_and_p");
  }
  outs->double_out(he->eoa,"eoa");
  outs->double_out(he->n0,"n0");
  outs->double_out(he->comp,"comp");
  outs->double_out(he->esym,"esym");
  outs->double_out(he->msom,"msom");
  outs->double_out(he->kprime,"kprime");
  outs->double_out(he->sprime,"sprime");
  */
  return 0;
}

template<> const char *io_tlate<hadronic_eos>::type()
{ return "hadronic_eos"; }

