/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/tov_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

const double xmev_kg=1.782661731e-30; 
const double mev_per_fm3_msun_km3=xmev_kg/gsl_mks::solar_mass*1.0e54; 

tov_interp_eos::tov_interp_eos() {

  eos_read=false;
  ldeos=false;
  verbose=1;
  ldpath=lib_settings.get_data_dir()+"nvbps.eos";

  cole=0;
  colp=0;
  colnb=0;

  ldcole=0;
  ldcolp=0;
  ldcolnb=0;
  
  eld=0.0;
  presld=0.0;
  nbld=0.0;

  efactor=1.0;
  pfactor=1.0;
  nfactor=1.0;

  ldeos_read=false;
  pwidth=1.0;
}

int tov_interp_eos::set_units(double s_efactor, double s_pfactor, 
			      double s_nfactor) {
  efactor=s_efactor;
  pfactor=s_pfactor;
  nfactor=s_nfactor;
  return gsl_success;
}

int tov_interp_eos::set_units(std::string leunits, 
			      std::string lpunits, std::string lnunits) {
  eunits=leunits;
  punits=lpunits;
  nunits=lnunits;

  if (eunits=="g/cm^3") {
    efactor=1.0e12/gsl_mks::solar_mass;
  } else if (eunits=="erg/cm^3") {
    efactor=1.0e2*xmev_kg/gsl_mks::solar_mass/ev_mks;
  } else if (eunits=="MeV/fm^3") {
    efactor=mev_per_fm3_msun_km3;
  } else if (eunits=="Msun/kg^3") {
    efactor=1.0;
  } else if (eunits=="fm^-4") {
    efactor=hc_mev_fm*mev_per_fm3_msun_km3;
  } else if (eunits=="") {
    efactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known energy and pressure units are:" << endl;
      cout << "'g/cm^3'" << endl;
      cout << "'erg/cm^3'" << endl;
      cout << "'MeV/fm^3'" << endl;
      cout << "'fm^-4'" << endl;
      cout << "'Msun/kg^3' (solar masses/kg^3)" << endl;
      cout << "eunits='" << eunits << "' was not understood." << endl;
    }
    set_err_ret("Incorrect energy density units in tov_solve::set_units().",
		gsl_efailed);
  }
  
  if (punits=="g/cm^3") {
    pfactor=1.0e12/gsl_mks::solar_mass;
  } else if (punits=="erg/cm^3") {
    pfactor=1.0e2*xmev_kg/gsl_mks::solar_mass/ev_mks;
  } else if (punits=="MeV/fm^3") {
    pfactor=mev_per_fm3_msun_km3;
  } else if (punits=="Msun/kg^3") {
    pfactor=1.0;
  } else if (punits=="fm^-4") {
    pfactor=hc_mev_fm*mev_per_fm3_msun_km3;
  } else if (punits=="") {
    pfactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known energy and pressure units are:" << endl;
      cout << "'g/cm^3'" << endl;
      cout << "'erg/cm^3'" << endl;
      cout << "'MeV/fm^3'" << endl;
      cout << "'fm^-4'" << endl;
      cout << "'Msun/kg^3' (solar masses/kg^3)" << endl;
      cout << "eunits='" << punits << "' was not understood." << endl;
      cout << "Exiting." << endl;
    }
    set_err_ret("Incorrect pressure units in tov_solve::set_units().",
		gsl_efailed);
  }

  if (nunits=="cm^-3") {
    nfactor=1.0e-39;
  } else if (nunits=="m^-3") {
    nfactor=1.0e-42;
  } else if (nunits=="fm^-3") {
    nfactor=1.0;
  } else if (nunits=="") {
    nfactor=1.0;
  } else {
    if (verbose>0) {
      cout << "Known baryon density units are:" << endl;
      cout << "'m^-3' (baryons/m^3)" << endl;
      cout << "'cm^-3' (baryons/cm^3)" << endl;
      cout << "'fm^-3' (baryons/fm^3)" << endl;
      cout << "nunits='" << nunits << "' was not understood." << endl;
      cout << "Exiting." << endl;
    }
    set_err_ret("Incorrect baryon density units in tov_solve::set_units().",
		gsl_efailed);
  }
  
  return gsl_success;
}

void tov_interp_eos::interp(const ovector_view &x, const ovector_view &y, 
			    double xx, double &yy, int n1, int n2) {
  int lo,hi,k;
  
  lo=n1;
  hi=n2;
  while (hi-lo > 1) {
    k=(hi+lo) >> 1;
    if (x[k] > xx) hi=k;
    else lo=k;
  }
  if (x[hi]==x[lo]) yy=y[hi];
  else yy=y[lo]+(xx-x[lo])/(x[hi]-x[lo])*(y[hi]-y[lo]);
  
  return;
}

int tov_interp_eos::check_eos() {

  // ---------------------------------------------------------------
  // Check EOS file to ensure that the file is increasing everywhere

  if (verbose>1) {
    
    if (false) {
      
      cout << "Checking EOS file...";
      bool success=true;
      for(int i=1;i<nfile && success==true;i++) {
	if ((*eost)[colp][i]<(*eost)[colp][i-1]) {
	  cout << endl << "Pressure decreases from " << (*eost)[colp][i-1] 
	       << " to " << (*eost)[colp][i] << " at row " << i << endl;
	  success=false;
	} else if ((*eost)[cole][i]<(*eost)[cole][i-1]) {
	  cout << endl << "Energy density decreases from " 
	       << (*eost)[cole][i-1] << " to " << (*eost)[cole][i] 
	       << " at row " << i << endl;
	  success=false;
	}
      }
      if (success==true) cout << "Success." << endl;
    }
    
    double dtemp;
    cout << "Energy, pressure, and baryon density\n"
	 << "given eos near transition density:" << endl;
    interp((*eost)[colnb],(*eost)[cole],0.08,dtemp,0,eost->get_nlines()-1);
    cout << dtemp << " ";
    interp((*eost)[colnb],(*eost)[colp],0.08,dtemp,0,eost->get_nlines()-1);
    cout << dtemp << " " << 0.08 << endl;
    
  }

  return gsl_success;
}

int tov_interp_eos::get_names(size_t &np, std::vector<std::string> &pnames) {
  np=0;
  for(int i=0;i<eost->get_ncolumns();i++) {
    if (i!=cole && i!=colp && i!=colnb) {
      np++;
      pnames.push_back(eost->get_column_name(i));
    }
  }
  return 0;
}

int tov_interp_eos::read_table_file(std::string eosfn, string s_cole, 
				    string s_colp, string s_colnb) {
  
  // ---------------------------------------------------------------
  // Read EOS file
  
  if (verbose>1) cout << "Reading EOS file." << endl;
  
  file_detect fd;
  in_file_format *iff=fd.open(eosfn.c_str());
  if (iff==0) {
    set_err_ret
      ("Failure opening file in tov_interp_eos::read_table_file().",
       gsl_efailed);
  }
  void *vp=0;
  int sz, sz2, ret;
  std::string tname;
  collection co;
  ret=co.in_one(iff,"table",tname,vp,sz,sz2);
  fd.close();
  if (ret!=0) {
    set_err_ret
      ("Failure finding table in tov_interp_eos::read_table_file().",
       gsl_efailed);
  }
  eost=(table *)vp;
  
  nfile=eost->get_nlines();
  if (verbose>1) cout << "Lines read from EOS file: " << nfile << endl;
  
  // ---------------------------------------------------------------
  // Look for energy density, pressure, and baryon density columns

  eost->lookup_column(s_cole,cole);
  eost->lookup_column(s_colp,colp);
  if (s_colnb!="") {
    eost->lookup_column(s_colnb,colnb);
    baryon_column=true;
  } else {
    baryon_column=false;
  }
  
  // ---------------------------------------------------------------

  eos_read=true;

  ret=check_eos();
  if (ret!=0) {
    set_err_ret
      ("Function check_eos() failed in tov_interp_eos::read_table_file().",
       gsl_efailed);
  }
  return gsl_success;
}

int tov_interp_eos::read_table(table &eosat, string s_cole, string s_colp, 
			       string s_colnb) {
  
  eost=&eosat;
  
  nfile=eost->get_nlines();
  if (verbose>0) cout << "Lines read from EOS file: " << nfile << endl;
  
  // ---------------------------------------------------------------
  // Look for energy density, pressure, and baryon density columns

  eost->lookup_column(s_cole,cole);
  eost->lookup_column(s_colp,colp);
  if (s_colnb!="") {
    eost->lookup_column(s_colnb,colnb);
    baryon_column=true;
  } else {
    baryon_column=false;
  }

  // ---------------------------------------------------------------
  
  eos_read=true;
  
  int ret=check_eos();
  
  if (ret!=0) {
    set_err_ret("Function check_eos() failed in tov_interp_eos::read_table().",
		gsl_efailed);
  }
  return gsl_success;
}

int tov_interp_eos::get_transition(double &plow, double &ptrans, 
				   double &phi) {
  plow=presld/pfactor;
  ptrans=prest/pfactor;
  phi=(*eost)[colp][0];
  return 0;
}

int tov_interp_eos::set_transition(double p, double wid) {
  prest=p*pfactor;
  pwidth=wid;
  return 0;
}

int tov_interp_eos::set_low_density_eos(bool s_ldeos, std::string s_ldpath,
					int s_ldcole, int s_ldcolp, 
					int s_ldcolnb) {
  
  if (s_ldpath.length()==0) {
    s_ldpath=lib_settings.get_data_dir()+"nvbps.eos";
  }

  if (s_ldeos==false) {
    ldeos=false;
    return gsl_success;
  }

  // ---------------------------------------------------------------
  // Read file

  // If there's already an ldeos that was read, then delete
  // the old one
  if (ldeos_read) {
    delete ld_eos;
  }
  
  ldeos=true;
  ldpath=s_ldpath;
  ldcole=s_ldcole;
  ldcolp=s_ldcolp;
  ldcolnb=s_ldcolnb;
  
  if (verbose>1) cout << "Reading low-density EOS file: " << ldpath << endl;
  
  file_detect fd;
  in_file_format *iff=fd.open(ldpath.c_str());
  if (iff==0) {
    set_err_ret
      ("Failed to open file in tov_interp_eos::set_low_density_eos()",
       gsl_efailed);
  }
  void *vp=0;
  int sz, sz2, ret;
  std::string tname;
  collection co;
  ret=co.in_one(iff,"table",tname,vp,sz,sz2);
  fd.close();
  if (ret!=0) {
    set_err_ret
      ("Failed to find table in tov_interp_eos::set_low_density_eos()",
       gsl_efailed);
  }
  ld_eos=(table *)vp;
  
  ld_eos->lookup_column("pr",ldcolp);
  ld_eos->lookup_column("ed",ldcole);
  ld_eos->lookup_column("nb",ldcolnb);
  eld=(*ld_eos)[ldcole][ld_eos->get_nlines()-1];
  presld=(*ld_eos)[ldcolp][ld_eos->get_nlines()-1];
  nbld=(*ld_eos)[ldcolnb][ld_eos->get_nlines()-1];
  prest=presld;
  
  if (verbose>1) {
    cout << "Largest energy, pressure, and baryon density of LD eos:" 
	 << endl;
    cout << eld << " " << presld << " " << nbld << endl;
  }
  
  ldeos_read=true;
  
  return gsl_success;
}
    
int tov_interp_eos::get_eden(double pres, double &ed, double &nb) {
  
  if (ldeos==true && pres*pfactor<=prest/pwidth) {

    // low density

    interp((*ld_eos)[ldcolp],(*ld_eos)[ldcole],pres*pfactor,ed,0,
	   ld_eos->get_nlines()-1);
    ed/=efactor;
    if (this->baryon_column) {
      interp((*ld_eos)[ldcolp],(*ld_eos)[ldcolnb],pres*pfactor,nb,0,
	     ld_eos->get_nlines()-1);
      nb/=nfactor;
    }

  } else if (ldeos==true && pres*pfactor>prest/pwidth && 
	     pres*pfactor<prest*pwidth) {

    // transition
    
    size_t ilo, ihi;
    ihi=eost->lookup(colp,pres);
    ilo=ld_eos->lookup(ldcolp,pres*pfactor);
    
    double ed1=(*ld_eos)[ldcole][ilo]/efactor;
    double ed2=(*eost)[cole][ihi];
    double chi=(pres*pfactor-prest/pwidth)/(prest*pwidth-prest/pwidth);
    ed=(1.0-chi)*ed1+chi*ed2;
    
    if (this->baryon_column) {
      double nb1=(*ld_eos)[ldcolnb][ilo]/nfactor;
      double nb2=(*eost)[colnb][ihi];
      nb=(1.0-chi)*nb1+chi*nb2;
    }


  } else {

    // high density
    
    interp((*eost)[colp],(*eost)[cole],pres,ed,0,eost->get_nlines()-1);
    if (this->baryon_column) {
      interp((*eost)[colp],(*eost)[colnb],pres,nb,0,eost->get_nlines()-1);
    }

  }
  
  return 0;
}

int tov_interp_eos::get_eden_user(double pres, double &ed, double &nb) {
  
  interp((*eost)[colp],(*eost)[cole],pres,ed,0,eost->get_nlines()-1);
  interp((*eost)[colp],(*eost)[colnb],pres,nb,0,eost->get_nlines()-1);

  return 0;
}

int tov_interp_eos::get_eden_ld(double pres, double &ed, double &nb) {
  
  interp((*ld_eos)[ldcolp],(*ld_eos)[ldcole],
	 pres*pfactor,ed,0,ld_eos->get_nlines()-1);
  ed/=efactor;
  interp((*ld_eos)[ldcolp],(*ld_eos)[ldcolnb],pres*pfactor,nb,0,
	 ld_eos->get_nlines()-1);
  nb/=nfactor;
  
  return 0;
}

int tov_interp_eos::get_aux(double P, size_t &np, ovector_view &auxp) {
  np=0;
  for(int i=0;i<eost->get_ncolumns();i++) {
    if (i!=cole && i!=colp && i!=colnb) {
      if (ldeos==true && P<=presld) {
	auxp[np]=0.0;
      } else {
	interp((*eost)[colp],(*eost)[i],P,auxp[np],0,eost->get_nlines()-1);
      } 
      np++;
    }
  }
  return 0;
}

