/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_COMP_GEN_INTE_H
#define O2SCL_COMP_GEN_INTE_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/funct.h>
#include <o2scl/inte.h>
#include <o2scl/gen_inte.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Naive generalized multi-dimensional integration 

      Naively combine several one-dimensional integration objects from
      class inte in order to perform a multi-dimensional integration.
      The integration routines are specified in the function
      set_ptrs().

      The integration routines are called in order of their
      specification in the list <tt>inte **ip</tt>. For the example of
      a two-dimensional integration <tt>ip[0]</tt> is called first
      with limits <tt>a[0]</tt> and <tt>b[0]</tt> and performs the
      integral of the integral given by <tt>ip[1]</tt> of the function
      from <tt>a[1]</tt> to <tt>b[1]</tt>, both of which may depend
      explicitly on <tt>x[0]</tt>. The integral performed is:
      \f[ 
      \int_{x_0=a_0}^{x_0=b_0} f(x_0) \int_{x_1=a_1(x_0)}^{x_1=b_1(x_0)} 
      f(x_0, x_1) ...
      \int_{x_{\mathrm{n}-1}=a_{\mathrm{n}-1}(x_0,x_1,..,x_{\mathrm{n}-2})}^
      {x_{\mathrm{n}-1}=b_{\mathrm{n}-1}(x_0,x_1,..,x_{\mathrm{n}-2})} 
      f(x_0,x_1,...,x_{\mathrm{n-1}})~d x_{\mathrm{n}-1}~...~d x_1~d x_0
      \f]

      See the discussion about the functions \c func, \c lower and \c
      upper in the documentation for the class gen_inte.

      No error estimate is performed. Error estimation for multiple
      dimension integrals is provided by the Monte Carlo integration
      classes (see \ref mcarlo_inte).
  */
  template<class param_t, class func_t, class lfunc_t=func_t, 
    class ufunc_t=func_t, class vec_t=ovector_view, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc> 
    class comp_gen_inte : 
  public gen_inte<param_t,func_t,lfunc_t,ufunc_t,vec_t> {

    public:

    comp_gen_inte() {
      
      ptrs_set=false;
      
      fmn=new funct_mfptr_noerr<comp_gen_inte<param_t,func_t,lfunc_t,
      ufunc_t,vec_t,alloc_vec_t,alloc_t>,od_parms>
      (this,&comp_gen_inte<param_t,func_t,lfunc_t,ufunc_t,vec_t,
       alloc_vec_t,alloc_t>::odfunc);

      ndim=0;
    }
    
    virtual ~comp_gen_inte() {
      delete fmn;
    }
    
    /** 
	\brief Parameters to send to the 1-d integration functions
	
	For basic usage, the class-user needs this type 
	to specify the parameter type for 1-d integration objects.
    */
    typedef struct {
      /// The independent variable vector
      alloc_vec_t *cx;
      /// The function specifying the lower limits
      lfunc_t *lower;
      /// The function specifying the upper limits
      ufunc_t *upper;
      /// The function to be integrated
      func_t *func;
      /// The number of dimensions
      int ndim;
      /// The present recursion level
      int idim;
      /// The user-specified parameter
      param_t *vp;
    } od_parms;
  
    /** 
	\brief Designate the pointers to the integration routines
	
	The user can, in principle, specify the one instance of an
	\ref inte object for several of the pointers, but this is
	discouraged as most inte objects cannot be used this way.
	This function will not warn you if some of the pointers
	specified in \c ip refer to the same object.

	If more 1-d integration routines than necessary are given,
	the extras will be unused.
    */
    int set_ptrs(inte<od_parms,funct<od_parms> > **ip, int n) {
      if (n<1) {
	set_err_ret
	  ("Number of dimensions less than 1 in set_ptrs().",gsl_einval);
      }
      ptrs=ip;
      ndim=n;
      ptrs_set=true;
      return 0;
    }

    /** \brief Integrate function \c func from 
	\f$ {\ell}_i=f_i(x_i) \f$ to \f$ u_i=g_i(x_i) \f$ for 
	\f$ 0<i<\mathrm{n}-1 \f$.
    */
    virtual double ginteg(func_t &func, size_t n, func_t &lower, 
			  func_t &upper, param_t &pa) {
      
      if (ptrs_set==false || n>ndim || n==0) {
	set_err("Pointers not set with proper dimension in ginteg().",
		gsl_einval);
	return 0.0;
      }
      
      alloc_vec_t cx;
      ao.allocate(cx,n);
      
      od_parms op={&cx,&lower,&upper,&func,n,0,&pa};
      
      double res=ptrs[0]->integ(*fmn,lower(0,cx,pa),upper(0,cx,pa),op);
      
      ao.free(cx);
      
      return res;
    }
  
    /// Return string denoting type ("comp_gen_inte")
    virtual const char *type() { return "comp_gen_inte"; }

#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// Memory allocator for objects of type \c alloc_vec_t 
    alloc_t ao;
    
    /// The function to send to the integrators
    funct_mfptr_noerr<comp_gen_inte<param_t,func_t,lfunc_t,ufunc_t,
    vec_t,alloc_vec_t,alloc_t>,od_parms> *fmn;
    
    /// The one-dimensional integration function
    double odfunc(double x, od_parms &od) {

      double res;

      (*od.cx)[od.idim]=x;
      if (od.idim==od.ndim-1) {
	(*od.func)(od.ndim,(*od.cx),res,*od.vp);
      } else {
	od_parms od2={od.cx,od.lower,od.upper,od.func,
		      od.ndim,od.idim+1,od.vp};
	(*od.func)(od.idim+1,*od.cx,res,*od.vp);
	res*=ptrs[od.idim]->integ
	  (*fmn,(*od.lower)(od.idim+1,*od.cx,*od.vp),
	   (*od.upper)(od.idim+1,*od.cx,*od.vp),od2);
      }
      return res;
    }
  
    /// The number of dimensions
    size_t ndim;
    
    /// Pointers to the integration objects
    inte<od_parms,funct<od_parms> > **ptrs;

    /// True if the integration objects have been specified
    bool ptrs_set;

#endif


  };

#ifndef DOXYGENP
}
#endif

#endif

