/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_GSL_RKCK_H
#define O2SCL_GSL_RKCK_H

#include <o2scl/err_hnd.h>
#include <o2scl/odestep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Cash-Karp embedded Runge-Kutta ODE stepper (GSL)
   */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc> class gsl_rkck : 
  public odestep<param_t,func_t,vec_t> {

  protected:
  
    /// \name Storage for the intermediate steps
    //@{
    alloc_vec_t k2, k3, k4, k5, k6, ytmp;
    //@}
      
    /// Size of allocated vectors
    size_t ndim;

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /** \name Storage for the coefficients
     */
    //@{
    double ah[5], b3[2], b4[3], b5[4], b6[5], ec[7];
    double b21, c1, c3, c4, c6;
    //@}
      
  public:

    gsl_rkck() {
      this->order=5;

      ah[0]=1.0/5.0;
      ah[1]=3.0/10.0;
      ah[2]=3.0/5.0;
      ah[3]=1.0;
      ah[4]=7.0/8.0;
      b3[0]=3.0/40.0;
      b3[1]=9.0/40.0;
      b4[0]=3.0/10.0;
      b4[1]=-9.0/10.0;
      b4[2]=12.0/10.0;
      b5[0]=-11.0/54.0;
      b5[1]=5.0/2.0;
      b5[2]=-70.0/27.0;
      b5[3]=35.0/27.0;
      b6[0]=1631.0/55296.0;
      b6[1]=175.0/512.0;
      b6[2]=575.0/13824.0;
      b6[3]=44275.0/110592.0;
      b6[4]=253.0/4096.0;
      ec[0]=0.0;
      ec[1]=37.0/378.0-2825.0/27648.0;
      ec[2]=0.0;
      ec[3]=250.0/621.0-18575.0/48384.0;
      ec[4]=125.0/594.0-13525.0/55296.0;
      ec[5]=-277.0/14336.0;
      ec[6]=512.0/1771.0-1.0/4.0;

      b21=1.0/5.0;
      c1=37.0/378.0;
      c3=250.0/621.0;
      c4=125.0/594.0;
      c6=512.0/1771.0;
	
      ndim=0;
    }
      
    virtual ~gsl_rkck() {
      if (ndim!=0) {
	ao.free(k2);
	ao.free(k3);
	ao.free(k4);
	ao.free(k5);
	ao.free(k6);
	ao.free(ytmp);
      }
    }

    /** 
	\brief Perform an integration step

	Given initial value of the n-dimensional function in \c y and
	the derivative in \c dydx (which must generally be computed
	beforehand) at the point \c x, take a step of size \c h giving
	the result in \c yout, the uncertainty in \c yerr, and the new
	derivative in \c dydx_out using function \c derivs to
	calculate derivatives. The parameters \c yout and \c y and the
	parameters \c dydx_out and \c dydx may refer to the same
	object.
    */
    virtual int step(double x, double h, size_t n, vec_t &y, vec_t &dydx, 
		     vec_t &yout, vec_t &yerr, vec_t &dydx_out, param_t &pa, 
		     func_t &derivs) {
	
      int ret=0;
      size_t i;
      
      if (ndim!=n) {
	if (ndim>0) {
	  ao.free(k2);
	  ao.free(k3);
	  ao.free(k4);
	  ao.free(k5);
	  ao.free(k6);
	  ao.free(ytmp);
	}
	ao.allocate(k2,n);
	ao.allocate(k3,n);
	ao.allocate(k4,n);
	ao.allocate(k5,n);
	ao.allocate(k6,n);
	ao.allocate(ytmp,n);

	ndim=n;
      }

      for (i=0;i<n;i++) {
	ytmp[i]=y[i]+b21*h*dydx[i];
      }
	
      o2scl::error_update(ret,derivs(x+ah[0]*h,n,ytmp,k2,pa));

      for (i=0;i<n;i++) {
	ytmp[i]=y[i]+h*(b3[0]*dydx[i]+b3[1]*k2[i]);
      }
      
      o2scl::error_update(ret,derivs(x+ah[1]*h,n,ytmp,k3,pa));
      
      for (i=0;i<n;i++) {
	ytmp[i]=y[i]+h*(b4[0]*dydx[i]+b4[1]*k2[i]+b4[2]*k3[i]);
      }

      o2scl::error_update(ret,derivs(x+ah[2]*h,n,ytmp,k4,pa));

      for (i=0;i<n;i++) {
	ytmp[i]=y[i]+h*(b5[0]*dydx[i]+b5[1]*k2[i]+b5[2]*k3[i]+
			b5[3]*k4[i]);
      }
	
      o2scl::error_update(ret,derivs(x+ah[3]*h,n,ytmp,k5,pa));
      
      for (i=0;i<n;i++) {
	ytmp[i]=y[i]+h*(b6[0]*dydx[i]+b6[1]*k2[i]+b6[2]*k3[i]+
			b6[3]*k4[i]+b6[4]*k5[i]);
      }
      
      o2scl::error_update(ret,derivs(x+ah[4]*h,n,ytmp,k6,pa));
      
      for (i=0;i<n;i++) {
	yout[i]=y[i]+h*(c1*dydx[i]+c3*k3[i]+c4*k4[i]+c6*k6[i]);
      }
      
      // We put this before the last function evaluation, in contrast
      // to the GSL version, so that the dydx[i] that appears in the
      // for loop below isn't modified by the subsequent derivative
      // evaluation using dydx_out. (The user could have given the
      // same vector for both)
      for (i=0;i<n;i++) {
	yerr[i]=h*(ec[1]*dydx[i]+ec[3]*k3[i]+ec[4]*k4[i]+ec[5]*k5[i]+
		   ec[6]*k6[i]);
      }

      o2scl::error_update(ret,derivs(x+h,n,yout,dydx_out,pa));
      
      return ret;
    }
    
  };
  
  /** \brief Faster Cash-Karp embedded Runge-Kutta ODE stepper (GSL)

      This a faster version of \ref gsl_rkck, which is a stepper for a
      fixed number of ODEs. It ignores the error values returned by
      the \c derivs argument. The argument \c n to step() should
      always be equal to the template parameter \c N, and the vector
      parameters to step must have space allocated for at least \c N
      elements. No error checking is performed to ensure that this is
      the case.
  */
  template<size_t N, class param_t, class func_t, class vec_t=ovector_view, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc> 
    class gsl_rkck_fast : 
  public odestep<param_t,func_t,vec_t> {
    
  protected:
    
    /// \name Storage for the intermediate steps
    //@{
    alloc_vec_t k2, k3, k4, k5, k6, ytmp;
    //@}
    
    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /** \name Storage for the coefficients
     */
    //@{
    double ah[5], b3[2], b4[3], b5[4], b6[5], ec[7];
    double b21, c1, c3, c4, c6;
    //@}
      
  public:

    gsl_rkck_fast() {
      this->order=5;

      ah[0]=1.0/5.0;
      ah[1]=3.0/10.0;
      ah[2]=3.0/5.0;
      ah[3]=1.0;
      ah[4]=7.0/8.0;
      b3[0]=3.0/40.0;
      b3[1]=9.0/40.0;
      b4[0]=3.0/10.0;
      b4[1]=-9.0/10.0;
      b4[2]=12.0/10.0;
      b5[0]=-11.0/54.0;
      b5[1]=5.0/2.0;
      b5[2]=-70.0/27.0;
      b5[3]=35.0/27.0;
      b6[0]=1631.0/55296.0;
      b6[1]=175.0/512.0;
      b6[2]=575.0/13824.0;
      b6[3]=44275.0/110592.0;
      b6[4]=253.0/4096.0;
      ec[0]=0.0;
      ec[1]=37.0/378.0-2825.0/27648.0;
      ec[2]=0.0;
      ec[3]=250.0/621.0-18575.0/48384.0;
      ec[4]=125.0/594.0-13525.0/55296.0;
      ec[5]=-277.0/14336.0;
      ec[6]=512.0/1771.0-1.0/4.0;

      b21=1.0/5.0;
      c1=37.0/378.0;
      c3=250.0/621.0;
      c4=125.0/594.0;
      c6=512.0/1771.0;

      ao.allocate(k2,N);
      ao.allocate(k3,N);
      ao.allocate(k4,N);
      ao.allocate(k5,N);
      ao.allocate(k6,N);
      ao.allocate(ytmp,N);
	
    }
      
    virtual ~gsl_rkck_fast() {

      ao.free(k2);
      ao.free(k3);
      ao.free(k4);
      ao.free(k5);
      ao.free(k6);
      ao.free(ytmp);
    }

    /** 
	\brief Perform an integration step

	Given initial value of the n-dimensional function in \c y and
	the derivative in \c dydx (which must generally be computed
	beforehand) at the point \c x, take a step of size \c h giving
	the result in \c yout, the uncertainty in \c yerr, and the new
	derivative in \c dydx_out using function \c derivs to
	calculate derivatives. The parameters \c yout and \c y and the
	parameters \c dydx_out and \c dydx may refer to the same
	object.

	\note The value of the parameter \c n should be equal to 
	the template parameter \c N.
    */
    virtual int step(double x, double h, size_t n, vec_t &y, vec_t &dydx, 
		     vec_t &yout, vec_t &yerr, vec_t &dydx_out, param_t &pa, 
		     func_t &derivs) {
      
      size_t i;
      
      for (i=0;i<N;i++) {
	ytmp[i]=y[i]+b21*h*dydx[i];
      }
	
      derivs(x+ah[0]*h,N,ytmp,k2,pa);

      for (i=0;i<N;i++) {
	ytmp[i]=y[i]+h*(b3[0]*dydx[i]+b3[1]*k2[i]);
      }
	
      derivs(x+ah[1]*h,N,ytmp,k3,pa);

      for (i=0;i<N;i++) {
	ytmp[i]=y[i]+h*(b4[0]*dydx[i]+b4[1]*k2[i]+b4[2]*k3[i]);
      }

      derivs(x+ah[2]*h,N,ytmp,k4,pa);

      for (i=0;i<N;i++) {
	ytmp[i]=y[i]+h*(b5[0]*dydx[i]+b5[1]*k2[i]+b5[2]*k3[i]+
			b5[3]*k4[i]);
      }
	
      derivs(x+ah[3]*h,N,ytmp,k5,pa);

      for (i=0;i<N;i++) {
	ytmp[i]=y[i]+h*(b6[0]*dydx[i]+b6[1]*k2[i]+b6[2]*k3[i]+
			b6[3]*k4[i]+b6[4]*k5[i]);
      }
    
      derivs(x+ah[4]*h,N,ytmp,k6,pa);

      for (i=0;i<n;i++) {
	yout[i]=y[i]+h*(c1*dydx[i]+c3*k3[i]+c4*k4[i]+c6*k6[i]);
      }
      
      // We put this before the last function evaluation, in contrast
      // to the GSL version, so that the dydx[i] that appears in the
      // for loop below isn't modified by the subsequent derivative
      // evaluation using dydx_out. (The user could have given the
      // same vector for both)
      for (i=0;i<n;i++) {
	yerr[i]=h*(ec[1]*dydx[i]+ec[3]*k3[i]+ec[4]*k4[i]+ec[5]*k5[i]+
		   ec[6]*k6[i]);
      }
      
      derivs(x+h,n,yout,dydx_out,pa);
      
      return 0;
    }
    
  };

#ifndef DOXYGENP
}
#endif

#endif
