/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_FUNCT_H
#define O2SCL_ODE_FUNCT_H
#include <string>
#include <o2scl/collection.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Ordinary differential equation function base

      This base class provides the basic format for specifying
      ordinary differential equations to integrate with the O2scl ODE
      solvers. Select the appropriate child of this class according
      to the kind of functions which are to be given to the solver.
  */
  template <class param_t, class vec_t=ovector_view>
    class ode_funct {
    public:

    ode_funct() {}

    virtual ~ode_funct() {}

    /** \brief The overloaded operator()
     */
    virtual int operator()(double x, size_t nv, const vec_t &y, 
			   vec_t &dydx, param_t &pa) {
      set_err_ret("Empty ode_funct::operator().",gsl_nobase);
    }
  
#ifndef DOXYGEN_INTERNAL
    private:
    ode_funct(const ode_funct &);
    ode_funct& operator=(const ode_funct&);
#endif
  };

  /** 
      \brief Provide ODE functions in the form of function pointers
  */
  template <class param_t, class vec_t=ovector_view>
    class ode_funct_fptr : public ode_funct<param_t,vec_t> {
    public:

    /** \brief Create an object given a function pointer
     */
    ode_funct_fptr(int (*fp)(double, size_t, const vec_t &, 
			     vec_t &, param_t &)) {
      fptr=fp;
    }

    virtual ~ode_funct_fptr() {}

    /** \brief Compute the \c nv derivatives as a function of the \c nv 
	functions specified in \c y at the point \c x.
    */
    virtual int operator()(double x, size_t nv, const vec_t &y, 
			   vec_t &dydx, param_t &pa) 
    {
      return fptr(x,nv,y,dydx,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    ode_funct_fptr() {};

    /// The function pointer
    int (*fptr)(double x, size_t nv, const vec_t &y, 
		vec_t &dydx, param_t &);

    private:

    ode_funct_fptr(const ode_funct_fptr &);
    ode_funct_fptr& operator=(const ode_funct_fptr&);

#endif

  };

  /** 
      \brief Provide ODE functions in the form of member 
      function pointers
  */
  template <class tclass, class param_t, class vec_t=ovector_view> 
    class ode_funct_mfptr : public ode_funct<param_t,vec_t> {

    public:
    
    /** \brief Create an object given a class and member function pointer
     */
    ode_funct_mfptr
      (tclass *tp, int (tclass::*fp)(double x, size_t nv, 
				     const vec_t &y, 
				     vec_t &dydx, param_t &)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~ode_funct_mfptr() {};
  
    /** \brief Compute the \c nv derivatives as a function of the \c nv 
	functions specified in \c y at the point \c x.
    */
    virtual int operator()(double x, size_t nv, const vec_t &y, 
			   vec_t &dydx, param_t &pa) {
      return (*tptr.*fptr)(x,nv,y,dydx,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

    protected:

    /// The pointer to the member function
    int (tclass::*fptr)(double x, size_t nv, const vec_t &y, 
			vec_t &dydx, param_t &);

    /// The pointer to the class
    tclass *tptr;

    private:

    ode_funct_mfptr(const ode_funct_mfptr &);
    ode_funct_mfptr& operator=(const ode_funct_mfptr&);

#endif

  };

  // ---------------------------------------------------------------------
  
  /** 
      \brief Ordinary differential equation function base for arrays
      
      This base class provides the basic format for specifying
      ordinary differential equations to integrate with the O2scl ODE
      solvers. Select the appropriate child of this class according
      to the kind of functions which are to be given to the solver.
  */
  template <class param_t, size_t nv> class ode_vfunct {

    public:

    ode_vfunct() {}

    virtual ~ode_vfunct() {}

    /** \brief Compute the \c nv derivatives as a function of the \c nv 
	functions specified in \c y at the point \c x.
    */
    virtual int operator()(double x, size_t nvar, const double y[nv], 
			   double dydx[nv], param_t &pa) {
      set_err_ret("Empty ode_vfunct::operator().",gsl_nobase);
    }
  
#ifndef DOXYGEN_INTERNAL
    private:
    ode_vfunct(const ode_vfunct &);
    ode_vfunct& operator=(const ode_vfunct&);
#endif
  };

  /** 
      \brief Function pointer to a function
  */
  template <class param_t, size_t nv>
    class ode_vfunct_fptr : public ode_vfunct<param_t,nv> {
    public:
    
    /** \brief Specify the function pointer
     */
    ode_vfunct_fptr(int (*fp)(double, size_t, const double y[nv], 
			      double dydx[nv], param_t &)) {
      fptr=fp;
    }

    virtual ~ode_vfunct_fptr() {}

    /** \brief The overloaded operator()
     */
    virtual int operator()(double x, size_t nvar, const double y[nv], 
			   double dydx[nv], param_t &pa) 
    {
      return fptr(x,nv,y,dydx,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    ode_vfunct_fptr() {};

    /// The function pointer
    int (*fptr)(double x, size_t nvar, const double y[nv], 
		double dydx[nv], param_t &);

    private:

    ode_vfunct_fptr(const ode_vfunct_fptr &);
    ode_vfunct_fptr& operator=(const ode_vfunct_fptr&);

#endif

  };

  /** 
      \brief Provide ODE functions in the form of member 
      function pointers
  */
  template <class tclass, class param_t, size_t nv> 
    class ode_vfunct_mfptr : public ode_vfunct<param_t,nv> {
    public:
    
    /** \brief Specify the member function pointer
     */
    ode_vfunct_mfptr
      (tclass *tp, int (tclass::*fp)(double x, size_t nvar, 
				     const double y[nv], 
				     double dydx[nv], param_t &)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~ode_vfunct_mfptr() {};
  
    /** \brief The overloaded operator()
     */
    virtual int operator()(double x, size_t nvar, const double y[nv], 
			   double dydx[nv], param_t &pa) {
      return (*tptr.*fptr)(x,nv,y,dydx,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

    protected:
  
    /// Pointer to the member function
    int (tclass::*fptr)(double x, size_t nvar, const double y[nv], 
			double dydx[nv], param_t &);

    /// Pointer to the class
    tclass *tptr;

    private:

    ode_vfunct_mfptr(const ode_vfunct_mfptr &);
    ode_vfunct_mfptr& operator=(const ode_vfunct_mfptr&);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
