/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/contour.h>

using namespace o2scl;
using namespace std;

contour::contour() {
  ncurves=0;
  nx=0;
  lev_adjust=1.0e-8;
  verbose=0;
  new_debug=0;
}

contour::~contour() {
  free_memory();
}

int contour::check_data() {
  bool increasing=false;
  if ((*xfun)[0]<(*xfun)[1]) increasing=true;
  for(int j=1;j<nx;j++) {
    if (increasing && (*xfun)[j-1]>=(*xfun)[j]) {
      set_err_ret("'x' array is not monotonic",gsl_einval);
    } else if (!increasing && (*xfun)[j-1]<=(*xfun)[j]) {
      set_err_ret("'x' array is not monotonic",gsl_einval);
    }
  }
  increasing=false;
  if ((*yfun)[0]<(*yfun)[1]) increasing=true;
  for(int j=1;j<ny;j++) {
    if (increasing && (*yfun)[j-1]>=(*yfun)[j]) {
      set_err_ret("'y' array is not monotonic",gsl_einval);
    } else if (!increasing && (*yfun)[j-1]<=(*yfun)[j]) {
      set_err_ret("'y' array is not monotonic",gsl_einval);
    }
  }
  return 0;
}

int contour::free_memory() {
  // Delete the storage space for the data
  if (nx!=0) {
    delete xfun;
    delete yfun;
    for (int i=0;i<ny;i++) {
      delete data[i];
    }
    delete[] data;
    nx=0;
  }
  
  // Delete the storage space for the edges
  for(size_t j=0;j<re.size();j++) {
    delete re[j];
    delete be[j];
    delete redges[j];
    delete bedges[j];
  }
  re.clear();
  be.clear();
  redges.clear();
  bedges.clear();

  // Delete the storage space for the contours
  if (ncurves!=0) {
    csizes.clear();
    vals.clear();
    xc.clear();
    yc.clear();
    ncurves=0;
  }

  return 0;
}

int contour::regrid_data(size_t xfact, size_t yfact) {
  if (nx==0) {
    set_err_ret("Data not set in regrid_data.",gsl_einval);
  }

  int i, j, newx=(nx-1)*xfact+1, newy=(ny-1)*yfact+1;
  
  ovector *xfun_old=xfun, *yfun_old=yfun;
  ovector **data_old=data;

  // Make space for new data
  xfun=new ovector(newx);
  yfun=new ovector(newy);
  data=new ovector *[newy];
  for(i=0;i<newy;i++) data[i]=new ovector(newx);
  
  sm_interp_vec *si;
  linear_interp<ovector_view> it1;
  linear_interp<ovector_const_subvector> it2;

  // Create new xfun
  ovector xidx(nx);
  for(i=0;i<nx;i++) xidx[i]=((double)i);
  si=new sm_interp_vec(it1,it2,nx,xidx,*xfun_old);
  for(i=0;i<newx;i++) {
    (*xfun)[i]=si->interp(((double)i)/(newx-1+xfact)*nx);
  }
  delete si;
  
  // Create new yfun
  ovector yidx(ny);
  for(i=0;i<ny;i++) yidx[i]=((double)i);
  si=new sm_interp_vec(it1,it2,ny,yidx,*yfun_old);
  for(i=0;i<newy;i++) {
    (*yfun)[i]=si->interp(((double)i)/(newy-1+yfact)*ny);
  }
  delete si;
  
  // Make a workspace array with [col][row] instead of [row][col]
  double **bc=new double *[newx];
  for(i=0;i<newx;i++) bc[i]=new double[ny];

  // Perform the first round
  for(i=0;i<newy;i+=yfact) {
    ovector at(nx);
    for(int ij=0;ij<nx;ij++) at[ij]=(*(data_old[i/yfact]))[ij];
    si=new sm_interp_vec(it1,it2,nx,*xfun_old,at);
    for(j=0;j<newx;j++) {
      bc[j][i/yfact]=si->interp((*xfun)[j]);
    }
    delete si;
  }

  // Perform the second round
  for(i=0;i<newx;i++) {
    ovector at2(ny);
    for(int ij=0;ij<ny;ij++) at2[ij]=bc[i][ij];
    si=new sm_interp_vec(it1,it2,ny,*yfun_old,at2);
    for(j=0;j<newy;j++) {
      (*(data[j]))[i]=si->interp((*yfun)[j]);
    }
    delete si;
  }

  // Delete the workspace 
  for(i=0;i<ny;i++) delete[] bc[i];
  delete[] bc;

  // Delete old data
  delete xfun_old;
  delete yfun_old;
  for(i=0;i<ny;i++) delete data_old[i];
  delete[] data_old;
  
  // Modify size
  nx=newx;
  ny=newy;
      
  return 0;
}

int contour::get_contour_size(int index) {
  if (index>=ncurves) {
    set_err("No curve with specified index in get_contour_size().",
	    gsl_efailed);
    return 0;
  }
  return csizes[index];
}

int contour::get_edges(const std::vector<omatrix_int *> *rints,
		       const std::vector<omatrix_int *> *bints,
		       const std::vector<omatrix *> *rpoints,
		       const std::vector<omatrix *> *bpoints) {
  rints=(const std::vector<omatrix_int *> *)&re;
  bints=(const std::vector<omatrix_int *> *)&be;
  rpoints=(const std::vector<omatrix *> *)&redges;
  bpoints=(const std::vector<omatrix *> *)&bedges;
  
  return 0;
}

int contour::get_edges_for_level(size_t nl, omatrix_int &rints, 
				 omatrix_int &bints,
				 omatrix &rpoints, 
				 omatrix &bpoints) {
  
  rints.allocate(re[nl]->rows(),re[nl]->cols());
  bints.allocate(be[nl]->rows(),be[nl]->cols());
  rpoints.allocate(redges[nl]->rows(),redges[nl]->cols());
  bpoints.allocate(bedges[nl]->rows(),bedges[nl]->cols());

  for(size_t j=0;j<re[nl]->rows();j++) {
    for(size_t k=0;k<re[nl]->cols();k++) {
      rints[j][k]=(*re[nl])[j][k];
      rpoints[j][k]=(*redges[nl])[j][k];
    }
  }
  for(size_t j=0;j<be[nl]->rows();j++) {
    for(size_t k=0;k<be[nl]->cols();k++) {
      bints[j][k]=(*be[nl])[j][k];
      bpoints[j][k]=(*bedges[nl])[j][k];
    }
  }

  return 0;
}

int contour::regions(bool larger) {
  /* 
     If the contour is closed, we need to find at least one point
     inside the contour (there is always at least one)
     Then, it is easy to determine if we need to shade the outside
     or the inside of the closed contour. However, its not 
     clear how straightforward it is to find an external line
     to properly shade the outside of a closed contour.
     If the contour is open, there are only six possibilities:
     1) right-bottom
     2) right-top
     3) left-bottom
     4) left-top
     5) left-right
     6) bottom-top

     To find the point inside a closed contour, all we need to
     do is find the point which has both a right edge and
     a bottom edge at the same point. 

     To find an external line, just find the rightmost right
     edge and then form a line from that right point
  */

  bool closed;
  double xfrst, xlast, yfrst, ylast;
  for(int nc=0;nc<ncurves;nc++) {
    xfrst=xc[nc][0];
    xlast=xc[nc][(xc[nc].size()-1)];
    yfrst=yc[nc][0];
    ylast=yc[nc][(yc[nc].size()-1)];
    if (xfrst==xlast && yfrst==ylast) {
      closed=true;
    } else {
      closed=false;
    }
    if (closed==true) {

      double data_inside=0.0;

      for(int j=0;j<nx-1;j++) {
	for(int k=0;k<ny-1;k++) {
	  if ((*rep)[j][k]==contourp && (*bep)[j][k]) {
	    data_inside=((*data)[j])[k];
	    j=nx+2;
	    k=ny+2;
	  }
	}
      }

      if ((larger && data_inside<levels[nc]) ||
	  (!larger && data_inside>levels[nc])) {

	// Need to find an external line
	int jright=0, kright=0;

	for(int j=0;j<nx-1;j++) {
	  for(int k=0;k<ny-1;k++) {
	    if ((*rep)[j][k]==contourp && j>jright) {
	      jright=j;
	      kright=k;
	    }
	  }
	}

      }
    } else {
      /*
      if (fabs(xfrst-(*xfun)[0])<xtol && 
	  fabs(xlast-(*xfun)[nx-1])<xtol) {
	// left-right
      } else if (fabs(xfrst-(*xfun)[nx-1])<xtol && 
		 fabs(xlast-(*xfun)[0])<xtol) {
	// right-left
      } else if (fabs(yfrst-(*yfun)[0])<ytol && 
		 fabs(ylast-(*yfun)[ny-1])<ytol) {
	// top-bottom
      } else if (fabs(yfrst-(*yfun)[ny-1])<ytol && 
		 fabs(ylast-(*yfun)[0])<ytol) {
	// bottom-top
      } else if (fabs(yfrst-(*yfun)[ny-1])<ytol && 
		 fabs(xlast-(*xfun)[nx-1])<xtol) {
	// right-bottom
      } else if (fabs(xfrst-(*xfun)[nx-1])<xtol && 
		 fabs(ylast-(*yfun)[ny-1])<ytol) {
	// bottom-right
      } else if (fabs(yfrst-(*yfun)[ny-1])<ytol && 
		 fabs(xlast-(*xfun)[0])<xtol) {
	// bottom-left
      } else if (fabs(xfrst-(*xfun)[0])<xtol && 
		 fabs(ylast-(*yfun)[ny-1])<ytol) {
	// left-bottom
      } else if (fabs(yfrst-(*yfun)[0])<ytol && 
		 fabs(xlast-(*xfun)[nx-1])<xtol) {
	// right-top
      } else if (fabs(xfrst-(*xfun)[nx-1])<xtol && 
		 fabs(ylast-(*yfun)[0])<ytol) {
	// top-right
      } else if (fabs(yfrst-(*yfun)[0])<ytol && 
		 fabs(xlast-(*xfun)[0])<xtol) {
	// top-left
      } else if (fabs(xfrst-(*xfun)[0])<xtol && 
		 fabs(ylast-(*yfun)[0])<ytol) {
	// left-top
      } else {
	cout << "Error in regions." << endl;
      }
      */
    }
  }
  return 0;
}

int contour::find_next_point_right(int j, int k, int &jnext, int &knext, 
				   int &dir_next, int nsw) {

  double closest=0.0, next;
  bool found=false;
  // down and left
  if (j<nx-1 && (*bep)[k][j]==nsw) {
    closest=sqrt(pow((*redgesp)[k][j]-(*yfun)[k],2.0)+
		 pow((*xfun)[j]-(*bedgesp)[k][j],2.0));
    found=true;
    jnext=j;
    knext=k;
    dir_next=dbottom;
  }
  // down
  if (j<nx-1 && (*rep)[k][j+1]==nsw) {
    next=sqrt(pow((*redgesp)[k][j]-(*redgesp)[k][j+1],2.0)+
	      pow((*xfun)[j]-(*xfun)[j+1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j+1;
      knext=k;
      dir_next=dright;
      closest=next;
    }
  }
  // down and right
  if (j<nx-1 && (*bep)[k+1][j]==nsw) {
    next=sqrt(pow((*redgesp)[k][j]-(*yfun)[k+1],2.0)+
	      pow((*xfun)[j]-(*bedgesp)[k+1][j],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j;
      knext=k+1;
      dir_next=dbottom;
      closest=next;
    }
  }
  // up
  if (j>0 && (*rep)[k][j-1]==nsw) {
    next=sqrt(pow((*redgesp)[k][j]-(*redgesp)[k][j-1],2.0)+
	      pow((*xfun)[j]-(*xfun)[j-1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j-1;
      knext=k;
      dir_next=dright;
      closest=next;
    }
  }
  // up and left
  if (j>0 && (*bep)[k][j-1]==nsw) {
    next=sqrt(pow((*redgesp)[k][j]-(*yfun)[k],2.0)+
	      pow((*xfun)[j]-(*bedgesp)[k][j-1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j-1;
      knext=k;
      dir_next=dbottom;
      closest=next;
    }
  }
  // up and right
  if (j>0 && (*bep)[k+1][j-1]==nsw) {
    next=sqrt(pow((*redgesp)[k][j]-(*yfun)[k+1],2.0)+
	      pow((*xfun)[j]-(*bedgesp)[k+1][j-1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j-1;
      knext=k+1;
      dir_next=dbottom;
      closest=next;
    }
  }
  if (found==true) return efound;
  return enot_found;
}

int contour::find_next_point_bottom(int j, int k, int &jnext, int &knext, 
				    int &dir_next, int nsw) {

  if (new_debug) cout << "fnpb: " << endl;
  double closest=0.0, next;
  bool found=false;
  // right and up
  if (new_debug) if (k<ny-1) cout << "1: " << (*rep)[k][j] << endl;
  if (k<ny-1 && (*rep)[k][j]==nsw) {
    closest=sqrt(pow((*bedgesp)[k][j]-(*xfun)[j],2.0)+
		 pow((*yfun)[k]-(*redgesp)[k][j],2.0));
    found=true;
    jnext=j;
    knext=k;
    dir_next=dright;
  }
  // right and down
  if (new_debug) if (k<ny-1) cout << "2: " << (*rep)[k][j+1] << endl;
  if (k<ny-1 && (*rep)[k][j+1]==nsw) {
    next=sqrt(pow((*bedgesp)[k][j]-(*xfun)[j+1],2.0)+
	      pow((*yfun)[k]-(*redgesp)[k][j+1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j+1;
      knext=k;
      dir_next=dright;
      closest=next;
    }
  }
  // right
  if (new_debug) if (k<ny-1) cout << "3: " << (*bep)[k+1][j] << endl;
  if (k<ny-1 && (*bep)[k+1][j]==nsw) {
    next=sqrt(pow((*bedgesp)[k][j]-(*bedgesp)[k+1][k],2.0)+
	      pow((*yfun)[k]-(*yfun)[k+1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j;
      knext=k+1;
      dir_next=dbottom;
      closest=next;
    }
  }
  // left and up
  if (new_debug) if (k>0) cout << "4: " << (*rep)[k-1][j] << endl;
  if (k>0 && (*rep)[k-1][j]==nsw) {
    next=sqrt(pow((*bedgesp)[k][j]-(*xfun)[j],2.0)+
	      pow((*yfun)[k]-(*redgesp)[k-1][j],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j;
      knext=k-1;
      dir_next=dright;
      closest=next;
    }
  }
  // left and down
  if (new_debug) if (k>0) cout << "5: " << (*rep)[k-1][j+1] << endl;
  if (k>0 && (*rep)[k-1][j+1]==nsw) {
    next=sqrt(pow((*bedgesp)[k][j]-(*xfun)[j+1],2.0)+
	      pow((*yfun)[k]-(*redgesp)[k-1][j+1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j+1;
      knext=k-1;
      dir_next=dright;
      closest=next;
    }
  }
  // left
  if (new_debug) if (k>0) cout << "6: " << (*bep)[k-1][j] << endl;
  if (k>0 && (*bep)[k-1][j]==nsw) {
    next=sqrt(pow((*bedgesp)[k][j]-(*bedgesp)[k-1][j],2.0)+
	      pow((*yfun)[k]-(*yfun)[k-1],2.0));
    if ((found==true && next<closest) || found==false) {
      found=true;
      jnext=j;
      knext=k-1;
      dir_next=dbottom;
      closest=next;
    }
  }
  
  if (new_debug) cout << "fnpb: " << found << endl;
  if (found==true) return efound;
  return enot_found;
}

int contour::find_intersections(size_t ilev, double &level) {
  int j, k;

  // Adjust the specified contour level to ensure none of the data
  // points is exactly on a contour
  bool level_corner;
  do {
    // Look for a match
    level_corner=false;
    for(k=0;k<ny;k++) {
      for(j=0;j<nx;j++) {
	if ((*(data[k]))[j]==level) {
	  level_corner=true;
	}
      }
    }
    // If we found a match, adjust the level
    if (level_corner==true) {
      if (verbose>0) {
	cout << "Found intersection of contour level and corner." << endl;
	cout << "Adjusting level " << level << " to ";
      }
      if (nlev==1) {
	// If there's only one contour level, then make a
	// simple adjustment
	level*=1.0+lev_adjust;
      } else {
	// Otherwise, compute the appropriate adjustment by 
	// finding the closest contour level not equal to the
	// current contour level
	double diff;
	int iclosest=0;
	if (ilev==0) {
	  diff=fabs(level-levels[1]);
	  iclosest=1;
	} else {
	  diff=fabs(level-levels[0]);
	  iclosest=0;
	}
	for(int ik=0;ik<nlev;ik++) {
	  if (ik!=((int)ilev)) {
	    if (fabs(level-levels[ik])<diff) {
	      iclosest=ik;
	      diff=fabs(level-levels[ik]);
	    }
	  }
	}
	level+=fabs(level-levels[iclosest])*lev_adjust;
      }
      if (verbose>0) cout << level << endl;
    }
  } while (level_corner==true);
  
  // Look in the square portion of the grid
  for(k=0;k<ny-1;k++) {
    for(j=0;j<nx-1;j++) {
      (*bep)[k][j]=empty;
      (*rep)[k][j]=empty;
      if (((*(data[k]))[j]-level)*((*(data[k]))[j+1]-level)<0.0) {
	(*bep)[k][j]=edge;
	if (verbose>0) {
	  cout << "Vertical edge for level   " << level << " between (" 
	       << k << "," << j << ") and (" << k << "," 
	       << j+1 << ")" << endl;
	}
      } 
      if (((*(data[k]))[j]-level)*((*(data[k+1]))[j]-level)<0.0) {
	if (verbose>0) {
	  cout << "Horizontal edge for level " << level << " between (" 
	       << k << "," << j << ") and (" << k+1 << "," 
	       << j << ")" << endl;
	}
	(*rep)[k][j]=edge;
      }
    }
  }

  // Look in the extra right edges
  for(k=0;k<ny-1;k++) {
    j=nx-1;
    (*rep)[k][j]=empty;
    if (((*(data[k]))[j]-level)*((*(data[k+1]))[j]-level)<0.0) {
      if (verbose>0) {
	cout << "Horizontal edge for level " << level << " between (" 
	     << k << "," << j << ") and (" << k+1 << "," 
	     << j << ")" << endl;
      }
      (*rep)[k][j]=edge;
    }
  }

  // Look in the extra bottom edges
  for(j=0;j<nx-1;j++) {
    k=ny-1;
    (*bep)[k][j]=empty;
    if (((*(data[k]))[j]-level)*((*(data[k]))[j+1]-level)<0.0) {
      if (verbose>0) {
	cout << "Vertical edge for level   " << level << " between (" 
	     << k << "," << j << ") and (" << k << "," 
	     << j+1 << ")" << endl;
      }
      (*bep)[k][j]=edge;
    } 
  }

  return 0;
}

int contour::right_edges(double level, sm_interp *si) {
  int j,k,m,ileft,iright,nint;

  for(k=0;k<ny-1;k++) {
    for(j=0;j<nx;j++) {
      
      // For each iright edge
      if ((*rep)[k][j]==edge) {
		
	// Find two points for linear interpolation
	ileft=k;
	if (ileft==ny-1) ileft--;
	nint=2;
	
	ovector xi(nint), yi(nint);
	
	xi[0]=(*(data[k]))[j];
	xi[1]=(*(data[k+1]))[j];
	yi[0]=(*yfun)[ileft];
	yi[1]=(*yfun)[ileft+1];
	
	// Do the interpolation and set the edge appropriately
	(*redgesp)[k][j]=si->interp(level,nint,xi,yi);
	
	if (verbose>0) {
	  cout << "Horizontal edge: (" << k << "," << j << ") -> ("
	       << k+1 << "," << j << ")" << endl;
	  cout << " coords: " << (*yfun)[ileft] << " "
	       << (*redgesp)[k][j] << " " << (*yfun)[ileft+1] << endl;
	  cout << "   data: " << (*(data[k]))[j] << " "
	       << level << " " << (*(data[k+1]))[j] << endl;
	}
      }
    }
  }
  return 0;
}

int contour::bottom_edges(double level, sm_interp *si) {
  int j,k,m,ileft,iright,nint;

  for(k=0;k<ny;k++) {
    for(j=0;j<nx-1;j++) {
      
      // For each bottom edge
      if ((*bep)[k][j]==edge) {
	
	// Find two points for linear interpolation
	ileft=j;
	if (ileft==nx-1) ileft--;
	nint=2;
	
	ovector xi(nint), yi(nint);
	
	xi[0]=(*(data[k]))[j];
	xi[1]=(*(data[k]))[j+1];
	yi[0]=(*xfun)[ileft];
	yi[1]=(*xfun)[ileft+1];

	// Do the interpolation and set the edge appropriately
	(*bedgesp)[k][j]=si->interp(level,nint,xi,yi);
	
	if (verbose>0) {
	  cout << "Vertical edge:   (" << k << "," << j << ") -> ("
	       << k+1 << "," << j << ")" << endl;
	  cout << " coords: " << (*xfun)[ileft] << " "
	       << (*bedgesp)[k][j] << " " << (*xfun)[ileft+1] << endl;
	  cout << "   data: " << (*(data[k]))[j] << " "
	       << level << " " << (*(data[k]))[j+1] << endl;
	}
      }
    }
  }
  return 0;
}

int contour::process_line(int j, int k, int dir, 
			  vector<double> &x, vector<double> &y,
			  bool first) {
  int fp, jnext, knext, dir_next;
  bool zero_points;

  // If we've found a new intersection
  if (new_debug) cout << "Looking for second point: " << (dir==dright) << endl;
  if (dir==dright) {
    fp=find_next_point_right(j,k,jnext,knext,dir_next);
  } else {
    fp=find_next_point_bottom(j,k,jnext,knext,dir_next);
  }
  if (fp==enot_found) zero_points=true;
  else zero_points=false;
  if (new_debug) cout << "zero_points: " << zero_points << endl;
  while (fp==efound) {
    j=jnext;
    k=knext;
    dir=dir_next;
    if (dir==dright) {
      if (verbose>0) {
	cout << "(" << (*xfun)[j] << ", " << (*redgesp)[k][j] << ")" << endl;
      }
      if (first) {
	x.push_back((*xfun)[j]);
	y.push_back((*redgesp)[k][j]);
      } else {
	x.insert(x.begin(),(*xfun)[j]);
	y.insert(y.begin(),(*redgesp)[k][j]);
      }
      (*rep)[k][j]=contourp;
      fp=find_next_point_right(j,k,jnext,knext,dir_next);
    } else {
      if (verbose>0) {
	cout << "(" << (*bedgesp)[k][j] << ", " << (*yfun)[k] << ")" << endl;
      }
      if (first) {
	x.push_back((*bedgesp)[k][j]);
	y.push_back((*yfun)[k]);
      } else {
	x.insert(x.begin(),(*bedgesp)[k][j]);
	y.insert(y.begin(),(*yfun)[k]);
      }
      (*bep)[k][j]=contourp;
      fp=find_next_point_bottom(j,k,jnext,knext,dir_next);
    }
  }

  // Set the last point to an endpoint
  if (dir==dright) {
    (*rep)[k][j]=endpoint;
  } else {
    (*bep)[k][j]=endpoint;
  }

  // If this is the second o2sclf of the line, see if we
  // need to connect two endpoints
  if (zero_points==false && first==false) {
    if (dir==dright) {
      fp=find_next_point_right(j,k,jnext,knext,dir_next,endpoint);
    } else {
      fp=find_next_point_bottom(j,k,jnext,knext,dir_next,endpoint);
    }
    if (dir==dright) {
      (*rep)[k][j]=contourp;
    } else {
      (*bep)[k][j]=contourp;
    }
    if (dir_next==dright) {
      (*rep)[knext][jnext]=contourp;
    } else {
      (*bep)[knext][jnext]=contourp;
    }
    if (fp==1) {
      x.push_back(x[0]);
      y.push_back(y[0]);
    }
  }

  if (verbose>0) {
    cout << "Endpoint reached." << endl;
    if (verbose>1) {
      char ch;
      cin >> ch;
    }
  }
  return 0;
}

bool contour::is_point_inside_old(double x1, double y1, const ovector_view &x,
				  const ovector_view &y, double xscale, 
				  double yscale) {
  
  // Find a point outside
  double minx, miny;
  minx=x[0];
  miny=y[0];
  for(size_t j=1;j<x.size();j++) {
    if (x[j]<minx) minx=x[j];
    if (y[j]<miny) miny=y[j];
  }

  // Ensure that the point is truly outside using xscale and yscale
  if (xscale<=0.0) minx-=0.01;
  else minx-=xscale;
  if (yscale<=0.0) miny-=0.01;
  else miny-=yscale;
  
  // Count crossings of line given by (x1,y1) and (minx,miny) 
  // with the contours. Odd means inside, even means outside
  int crossings=0;
  for(size_t j=1;j<x.size();j++) {
    if (lines_cross_old(minx,miny,x1,y1,x[j],y[j],x[j-1],y[j-1])) crossings++;
  }

  if (crossings%2==0) return false;
  return true;
}

bool contour::lines_cross_old(double x1, double y1, double x2, double y2,
			      double x3, double y3, double x4, double y4) {
  // The intersection of the two line segments is (xi,yi). 
  double xi=(x1*(x4*(y3-y2)+x3*(y2-y4))+x2*(x4*(y1-y3)+x3*(y4-y1)))/
    ((x1-x2)*(y3-y4)-(x3-x4)*(y1-y2));
  double yi=(x4*(y1-y2)*y3+x1*y2*y3-x3*y1*y4-x1*y2*y4+x3*y2*y4+x2*y1*(y4-y3))/
    ((x1-x2)*(y3-y4)-(x3-x4)*(y1-y2));
  if (!gsl_finite(xi) && !gsl_finite(yi)) {
    set_err("Given points do not define a line.",gsl_einval);
    return false;
  }
  if (gsl_finite(xi)) {
    // Determine if xi is between the endpoints of both line segments
    if ((xi-x1)*(xi-x2)<=0.0 && (xi-x3)*(xi-x4)<=0.0) {
      // If so, return true
      return true;
    }
  } else {
    // Determine if yi is between the endpoints of both line segments
    if ((yi-y1)*(yi-y2)<=0.0 && (yi-y3)*(yi-y4)<=0.0) {
      // If so, return true
      return true;
    }
  }
  return false;
}

int contour::smooth_contours(size_t nfact) {
  if (nfact<2) return 0;

  o2scl_interp<> hi;
  for(int i=0;i<ncurves;i++) {
    
    // Copy the curve to temporary storage
    ovector xtmp(csizes[i]), ytmp(csizes[i]), ttmp(csizes[i]);
    for(int j=0;j<csizes[i];j++) {
      ttmp[j]=((double)j);
      xtmp[j]=xc[i][j];
      ytmp[j]=yc[i][j];
    }
    
    // Make sure enough storage is reserved for the new contours
    xc[i].resize((csizes[i])*(nfact));
    yc[i].resize((csizes[i])*(nfact));
    
    for(int j=0;j<csizes[i];j++) {

      // Input the given point
      xc[i][j*nfact]=xtmp[j];
      yc[i][j*nfact]=ytmp[j];

      // Compute the remaining points by interpolation
      for(size_t k=1;k<nfact;k++) {
	// Need to check this
	double t=((double)j)+((double)k)/((double)nfact);
	xc[i][j*nfact+k]=hi.interp(t,csizes[i],ttmp,xtmp);
	yc[i][j*nfact+k]=hi.interp(t,csizes[i],ttmp,ytmp);
      }
    }
    
    csizes[i]*=nfact;
  }
  return 0;
}
