/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_PINSIDE_H
#define O2SCL_PINSIDE_H

#include <o2scl/ovector_tlate.h>
#include <o2scl/array.h>
#include <o2scl/test_mgr.h>
#include <o2scl/vec_arith.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Test line intersection and point inside polygon

      This is a fast and dirty implementation of the point inside
      polygon test from Jerome L. Lewis, SIGSCE Bulletin, 34 (2002)
      81.

      Note that an error in that article ("count-" should have been
      "count--") has been corrected here. 
  */
  class pinside {

  protected:
  
    /// Internal point definition
    struct point {
      double x;
      double y;
    };
  
    /// Internal line definition
    struct line {
      point p1;
      point p2;
    };
  
    /// Test if line segments \c P and \c Q intersect
    int intersect(line P, line Q);
  
    /// Test if point \c t is inside polygon \c p of size \c N
    int inside(point t, point p[], int N);
  
  public:

    /** 
	\brief Determine if two line segments intersect

	The function returns 1 if the line segment
	determined by the endpoints \f$ (x_1,y_1) \f$ 
	and \f$ (x_2,y_2) \f$ and the line segment 
	determined by the endpoints \f$ (x_3,y_3) \f$ 
	and \f$ (x_4,y_4) \f$ intersect, and 0 otherwise.
    */
    int intersect(double x1, double y1, double x2, double y2, 
		  double x3, double y3, double x4, double y4) {
      line P={{x1,y1},{x2,y2}};
      line Q={{x3,y3},{x4,y4}};
      return intersect(P,Q);
    }

    /** 
	\brief Determine if point (x,y) is inside a polygon

	This returns 1 if the point (x,y) is inside the polygon
	defined by \c xa and \c ya, and 0 otherwise. 

	Note that if the point (x,y) is exactly on the polygon,
	then the result of this function is not well-defined
	and it will return either 0 or 1. 
    */
    int inside(double x, double y, const o2scl::ovector_view &xa,
	       const o2scl::ovector_view &ya);
  
    /// Perform some simple testing
    int test(test_mgr &t) {
      line tx1={{0,0},{2,2}};
      line ty1={{0,1},{1,0}};
      t.test_gen(intersect(tx1,ty1)==1,"i1");
      line tx2={{0,0},{0.49,0.49}};
      line ty2={{0,1},{1,0}};
      t.test_gen(intersect(tx2,ty2)==0,"i2");
      line tx3={{0,0},{-1,-1}};
      line ty3={{0,1},{1,0}};
      t.test_gen(intersect(tx3,ty3)==0,"i3");

      ovector x, y;
      x.push_back(1); y.push_back(0);
      x.push_back(2); y.push_back(0);
      x.push_back(2); y.push_back(1);
      x.push_back(3); y.push_back(1);
      x.push_back(3); y.push_back(2);
      x.push_back(2); y.push_back(2);
      x.push_back(2); y.push_back(3);
      x.push_back(1); y.push_back(3);
      x.push_back(1); y.push_back(2);
      x.push_back(0); y.push_back(2);
      x.push_back(0); y.push_back(1);
      x.push_back(1); y.push_back(1);

      t.test_gen(inside(1.5,1.5,x,y)==1,"in1");
      t.test_gen(inside(2.01,1.99,x,y)==1,"in2");
      t.test_gen(inside(1.99,2.01,x,y)==1,"in3");
      t.test_gen(inside(1.99,1.99,x,y)==1,"in4");

      t.test_gen(inside(2.01,2.01,x,y)==0,"out1");
      t.test_gen(inside(0.5,0.5,x,y)==0,"out2");
      t.test_gen(inside(2.5,2.5,x,y)==0,"out3");

      t.test_gen(inside(0.5,1.5,x,y)==1,"c1");
      t.test_gen(inside(0.5,2.5,x,y)==0,"c2");
      t.test_gen(inside(1.5,0.5,x,y)==1,"c3");
      t.test_gen(inside(1.5,2.5,x,y)==1,"c4");
      t.test_gen(inside(2.5,0.5,x,y)==0,"c5");
      t.test_gen(inside(2.5,1.5,x,y)==1,"c6");

      return 0;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif
