/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_PLANAR_INTP_H
#define O2SCL_PLANAR_INTP_H

#include <iostream>
#include <string>
#include <cmath>
#include <o2scl/err_hnd.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Interpolate among two independent variables with planes

      This is an analog of 1-d linear interpolation for two
      dimensions.  For a set of data \f${x_i,y_i,f_{j,i}}\f$, values
      for \f$f_j\f$ are predicted given a value of x and y.  (In
      contrast to twod_intp, the data need not be presented in a
      grid.)  This is done by finding the plane that goes through
      three closest points in the data set.

      This procedure will fail if the three closest points are 
      co-linear, and interp() will then call set_err() and
      return zero.
      
      There is no caching so the numeric values of the data may be
      freely changed between calls to interp().

      The vector and matrix types can be any types which have
      suitably defined functions \c operator[]. 

      \future Rewrite so that it never fails unless all the points in
      the data set lie on a line. This would probably demand sorting
      all of the points by distance from desired location.
  */
  template<class vec_t, class mat_t> class planar_intp {
  public:
    
    planar_intp() {
      data_set=false;
    }

    /** \brief Initialize the data for the planar interpolation
     */
    int set_data(size_t n_points, vec_t &x, vec_t &y, size_t n_dat, 
		 mat_t &dat) {
      if (n_points<3) {
	set_err_ret("Must provide at least three points in set_data()",
		    gsl_efailed);
      }
      if (x.size()<n_points) {
	set_err_ret("x too small.",gsl_einval);
      } else if (y.size()<n_points) {
	set_err_ret("y too small.",gsl_einval);
      } 
      np=n_points;
      nd=n_dat;
      ux=&x;
      uy=&y;
      udat=&dat;
      data_set=true;
      return 0;
    }
    
    /** 
	\brief Perform the planar interpolation 

	It is assumed that \c ip is properly allocated beforehand.
    */
    int interp(double x, double y, vec_t &ip) {
      double x1, x2, x3, y1, y2, y3;
      return interp(x,y,ip,x1,y1,x2,y2,x3,y3);
    }

    /** 
	\brief Planar interpolation returning the closest points 

	This function interpolates \c x and \c y into the data
	returning \c ip. It also returns the three closest x- and
	y-values used for computing the plane.
	
	It is assumed that \c ip is properly allocated beforehand.
    */
    int interp(double x, double y, vec_t &ip, double &x1, double &y1, 
	       double &x2, double &y2, double &x3, double &y3) {
      size_t i, j;
      double c1, c2, c3, c4;
      int i1, i2, i3, i4;
      
      if (data_set==false) {
	set_err_ret("Data not set in planar_intp::interp().",gsl_einval);
      }
  
      /// Put in initial points
      i1=0; i2=1; i3=2;
      c1=sqrt(pow(x-(*ux)[0],2.0)+pow(y-(*uy)[0],2.0));
      c2=sqrt(pow(x-(*ux)[1],2.0)+pow(y-(*uy)[1],2.0));
      c3=sqrt(pow(x-(*ux)[2],2.0)+pow(y-(*uy)[2],2.0));

      /// Sort initial points
      if (c2<c1) {
	if (c3<c2) {
	  // 321
	  swap(i1,c1,i3,c3);
	} else if (c3<c1) {
	  // 231
	  swap(i1,c1,i2,c2);
	  swap(i2,c2,i3,c3);
	} else {
	  // 213
	  swap(i1,c1,i2,c2);
	}
      } else {
	if (c3<c1) {
	  // 312
	  swap(i1,c1,i3,c3);
	  swap(i2,c2,i3,c3);
	} else if (c3<c2) {
	  // 132
	  swap(i3,c3,i2,c2);
	}
	// 123
      }

      // Go through remaining points and sort accordingly
      for(j=3;j<np;j++) {
	i4=j;
	c4=sqrt(pow(x-(*ux)[i4],2.0)+pow(y-(*uy)[i4],2.0));
	if (c4<c1) {
	  swap(i4,c4,i3,c3);
	  swap(i3,c3,i2,c2);
	  swap(i2,c2,i1,c1);
	} else if (c4<c2) {
	  swap(i4,c4,i3,c3);
	  swap(i3,c3,i2,c2);
	} else if (c4<c3) {
	  swap(i4,c4,i3,c3);
	}
      }

      // Solve for denominator:
      double a, b, c, den, z1, z2, z3;
      x1=(*ux)[i1];
      x2=(*ux)[i2];
      x3=(*ux)[i3];
      y1=(*uy)[i1];
      y2=(*uy)[i2];
      y3=(*uy)[i3];
      den=(-x2*y1+x3*y1+x1*y2-x3*y2-x1*y3+x2*y3);
  
      if (den==0.0) {

	set_err_ret("Interpolation failed.",gsl_efailed);

      } else {
    
	// Calculate value of function
	for(i=0;i<nd;i++) {
	  z1=(*(udat[i]))[i1];
	  z2=(*(udat[i]))[i2];
	  z3=(*(udat[i]))[i3];
	  a=-(-y2*z1+y3*z1+y1*z2-y3*z2-y1*z3+y2*z3)/den;
	  b=-(x2*z1-x3*z1-x1*z2+x3*z2+x1*z3-x2*z3)/den;
	  c=-(x3*y2*z1-x2*y3*z1-x3*y1*z2+x1*y3*z2+x2*y1*z3-x1*y2*z3)/den;
      
	  ip[i]=a*x+b*y+c;
	}
      }

      return 0;
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The number of points
    size_t np;
    /// The number of functions
    size_t nd;
    /// The x-values
    vec_t *ux;
    /// The y-values
    vec_t *uy;
    /// The data
    mat_t *udat;
    /// True if the data has been specified
    bool data_set;
    
    /// Swap points 1 and 2.
    int swap(int &i1, double &c1, int &i2, double &c2) {
      int t;
      double tc;
      
      t=i1; tc=c1;
      i1=i2; c1=c2;
      i2=t; c2=tc;
      
      return 0;
    }
    
#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif



