/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/fermion.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

fermion::fermion(double mass, double dof) : part(mass,dof) {
  kf=0.0;
  del=0.0;
  name="fermion";
  massless_root=&def_massless_root;
}

int fermion::massless_calc_mu(const double temper) {

  double fm2, fm3;

  if (non_interacting) { nu=mu; }

  fm2=gsl_sf_fermi_dirac_int(2,nu/temper);
  fm3=gsl_sf_fermi_dirac_int(3,nu/temper);
  
  n=g/pi2*pow(temper,3.0)*fm2;
  ed=g*3.0/pi2*pow(temper,4.0)*fm3;
  pr=ed/3.0;
  en=(ed+pr-n*nu)/temper;

  return gsl_success;
}

int fermion::massless_fun(double x, double &y, void *&pa) {
  double fm2, *temper=(double *)pa;
  
  fm2=gsl_sf_fermi_dirac_int(2,x/(*temper));
  y=g*pow(*temper,3.0)*fm2/pi2/n-1.0;
  
  return gsl_success;
}


int fermion::massless_calc_density(const double temper) {
  double x, T=temper;
  int ret;
  
  x=ms+temper;
  void *vp=(void *)&T;
  funct_mfptr<fermion,void *> mf(this,&fermion::massless_fun);
  ret=massless_root->solve(x,vp,mf);
  if (ret!=0) {
    // We set an error, but proceed anyway.
    add_err("Solver failed in massless_calc_density().",ret);
  }
  nu=x;

  massless_calc_mu(temper);

  // If the particle is non-interacting, then need to set
  // mu=nu to get the entropy right
  if (non_interacting) { mu=nu; }

  return ret;
}

int fermion::calc_mu(const double temper) {
  set_err_ret("Missing base function fermion::calc_mu().",gsl_nobase);
}

int fermion::calc_density(const double temper) {
  set_err_ret("Missing base function fermion::calc_density().",gsl_nobase);
}

int fermion::pair_mu(const double temper) {
  set_err_ret("Missing base function fermion::pair_mu().",gsl_nobase);
}

int fermion::pair_density(const double temper) {
  set_err_ret("Missing base function fermion::pair_density().",gsl_nobase);
}

// Zero-temperature fermions:
int fermion::kffromden() {
  kf=cbrt(6.0*pi2/g*n);
  return gsl_success;
}

int fermion::calc_mu_zerot() {
  bool nulessthan0;
  if (non_interacting) { nu=mu; ms=m; }
  
  if (inc_rest_mass) {
    if (nu>ms) {
      nulessthan0=false;
      kf=sqrt(nu*nu-ms*ms);
    } else {
      nulessthan0=false;
      kf=0.0;
    }
  } else {
    double nupm=nu+m;
    if ((nupm)>ms) {
      nulessthan0=false;
      kf=sqrt(nupm*nupm-ms*ms);
    } else {
      nulessthan0=false;
      kf=0.0;
    }
  }
  n=g/6.0/pi2*pow(kf,3.0);
  sden();
  eden();
  pres();
  en=0.0;
  if (!inc_rest_mass) ed-=n*m;
  if (nulessthan0==true) {
    n*=-1.0;
    kf*=-1.0;
  }

  return gsl_success;
}

int fermion::calc_density_zerot() {
  if (non_interacting) { ms=m; }

  kf=cbrt(6.0*pi2/g*n);
  nu=sqrt(kf*kf+ms*ms);
  sden();
  n=g/6.0/pi2*pow(kf,3.0);
  eden();
  pres();
  en=0.0;

  if (!inc_rest_mass) {
    nu-=m;
    ed-=n*m;
  }

  if (non_interacting) { mu=nu; }

  return gsl_success;
}

int fermion::sden() {
  double r, efs;
  if (kf>0.0) {
    if (ms>0.0) {
      efs=sqrt(kf*kf+ms*ms);
      r=(kf+efs)/ms;
    }
  }
  return gsl_success;
}

int fermion::eden() {
  double r,efs;
  if (kf>0.0) {
    if (ms<=0.0) {
      ed=g*(pow(kf,4.0)/8.0/pi2);
    } else {
      efs=sqrt(kf*kf+ms*ms);
      r=(kf+efs)/ms;
      ed=g/16.0/pi2*(2.0*kf*pow(efs,3.0)-kf*efs*ms*ms
		     -pow(ms,4.0)*log(r));
    }
  } else {
    ed=0.0;
  }
  return gsl_success;
}

int fermion::pres() {
  double r,efs;
  if (kf>0.0) {
    if (ms<=0.0) {
      pr=g*(pow(kf,4.0)/24.0/pi2);
    } else {
      efs=sqrt(kf*kf+ms*ms);
      r=(kf+efs)/ms;
      pr=g/48.0/pi2*(2.0*efs*pow(kf,3.0)-3.0*kf*efs*ms*ms
		     +3.0*pow(ms,4.0)*log(r));
    }
  } else {
    pr=0.0;
  }
  return gsl_success;
}

int fermion::massless_pair_mu(const double temper) {
  double pitmu, pitmu2, nu2;

  if (non_interacting) { nu=mu; ms=m; }
  if (nu==0.0) {
    n=0.0;
    ed=g/8.0/pi2*7.0/15.0*pow(pi*temper,4.0);
    pr=ed/3.0;
    en=(ed+pr-n*mu)/temper;
  } else {
    nu2=nu*nu;
    pitmu=pi*temper/nu;
    pitmu2=pitmu*pitmu;
    n=g*nu*nu2/6.0/pi2*(1.0+pitmu2);
    ed=g*nu2*nu2/8.0/pi2*
      (1.0+2.0*pitmu2+7.0/15.0*pitmu2*pitmu2);
    pr=ed/3.0;
    en=(ed+pr-n*nu)/temper;
  }
  return gsl_success;
}

int fermion::massless_pair_density(const double temper) {
  double t2=temper*temper,pitmu,pitmu2,nu2;
  double cbt, alpha, two13, alpha16;

  if (non_interacting) { ms=m; }
  if (n<=0.0) {
    nu=0.0;
    ed=g/8.0/pi2*7.0/15.0*pow(pi*temper,4.0);
    pr=ed/3.0;
  } else {
    alpha=g*g*pi2*t2*t2*t2/243.0/n/n;
    if (alpha>1.0e4) {
      nu=(2.0/3.0/sqrt(alpha)-8.0/81.0/pow(alpha,1.5)+
	  32.0/729.0/pow(alpha,2.5))*pi*temper/sqrt(3.0);
    } else if (alpha<1.0e-4) {
      two13=cbrt(2.0);
      alpha16=pow(alpha,1.0/6.0);
      nu=(two13/alpha16-alpha16/two13+alpha/alpha16/6.0/two13/two13
	  +alpha*alpha16/12.0/two13-alpha*alpha/alpha16/18.0/two13/two13-
	  5.0*alpha*alpha*alpha16/144.0/two13+
	  77.0/2592.0*alpha*alpha*alpha/alpha16/two13/two13)*
	pi*temper/sqrt(3.0);
    } else {
      cbt=pow(-1.0+sqrt(1.0+alpha),1.0/3.0)/pow(alpha,1.0/6.0);
      nu=pi*temper/sqrt(3.0)*(1.0/cbt-cbt);
    }
    pitmu=pi*temper/nu;
    pitmu2=pitmu*pitmu;
    nu2=nu*nu;
    ed=g*nu2*nu2/8.0/pi2*
      (1.0+2.0*pitmu2+7.0/15.0*pitmu2*pitmu2);
    pr=ed/3.0;

    if (!finite(nu)) {
      if (non_interacting) { mu=nu; }
      en=(ed+pr-n*nu)/temper;
      set_err_ret("Internal error in fermion::massless_pair_density().",
		  gsl_efailed);
    }
  }

  if (non_interacting) { mu=nu; }
  en=(ed+pr-n*nu)/temper;

  return gsl_success;
}

template<> int io_tlate<fermion>::input(cinput *co, in_file_format *ins, 
					fermion *f) {
  part *p=(part *)f;
  std::string t;
  co->object_in("part",ins,p,t);
  ins->double_in(f->del,"del");
  ins->double_in(f->kf,"kf");
  return 0;
}

template<> int io_tlate<fermion>::output(coutput *co, out_file_format *outs, 
					 fermion *f) {
  std::string t;
  co->object_out("part",outs,(part *)f,0,0,t);
  outs->double_out(f->kf,"del");
  outs->double_out(f->del,"kf");
  return 0;
}

template<> const char *io_tlate<fermion>::type() { return "fermion"; }

