/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_FERMION_H
#define O2SCL_REL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qagiu.h>
#include <o2scl/gsl_inte_qag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic fermion

      This implements an equation of state for a relativistic fermion
      using direct integration. Define the degeneracy parameter 
      \f[
      \psi=(\nu-m^{*})/T 
      \f] 
      where \f$ \nu \f$ is the effective chemical potential and \f$
      m^{*} \f$ is the effective mass. For \f$ \psi \f$ greater than
      \ref deg_limit (degenerate regime), a finite interval integrator
      is used and for \f$ \psi \f$ less than \ref deg_limit
      (non-degenerate regime), an integrator over the interval from
      \f$ [0,\infty) \f$ is used. Typical choices are Gauss-Legendre
      integration for the degenerate regime and Gauss-Laguerre
      integration for the non-degenerate regime. The upper limit on
      the degenerate integration is given by
      \f[
      \sqrt{(20 T+{\nu})^2-m^{*,2}}
      \f]
      
      The default integrators are gsl_inte_qag and gsl_inte_qagiu.
      
      \note This does not work with inc_rest_mass=false
      
  */
  class rel_fermion : public fermion {

  public:

    /// The critical degeneracy at which to switch integration techniques
    double deg_limit;

    /// Storage for the uncertainty
    fermion unc;

    /** \brief If true, use the present value of the chemical potential as
	a guess for the new chemical potential
    */
    bool guess_from_nu;
    
    /// Create a fermion with mass \c m and degeneracy \c g
    rel_fermion(double m=0.0, double g=0.0);
    virtual ~rel_fermion();
    
    /** 
	\brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(const double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
     */
    virtual int pair_density(const double temper);

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(const double temper);
    
    /// Set integrators
    int set_inte(inte<void *,funct<void *> > &non_it, 
		 inte<void *,funct<void *> > &deg_it);
    
    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<void *,funct<void *> > &rp) {
      density_root=&rp;
      return 0;
    }
    
    /// The default solver for calc_density().
    cern_mroot_root<void *,funct<void *> > def_density_root;

    /// The default integrator for degenerate fermions
    gsl_inte_qag<void *,funct<void *> > def_dit;

    /// The default integrator for non-degenerate fermions
    gsl_inte_qagiu<void *,funct<void *> > def_nit;

    /// Return string denoting type ("rel_fermion")
    virtual const char *type() { return "rel_fermion"; }

  protected:
    
#ifndef DOXYGEN_INTERNAL
    
    friend class io_tlate<rel_fermion>;

    /// The non-degenerate integrator
    inte<void *,funct<void *> > *nit;
    /// The degenerate integrator
    inte<void *,funct<void *> > *dit;
    /// The solver for calc_density()
    root<void *,funct<void *> > *density_root;

    /// The integrand for the density for non-degenerate fermions
    double density_fun(double u, void *&pa);

    /// The integrand for the energy density for non-degenerate fermions
    double energy_fun(double u, void *&pa);

    /// The integrand for the entropy density for non-degenerate fermions
    double entropy_fun(double u, void *&pa);

    /// The integrand for the density for degenerate fermions
    double deg_density_fun(double u, void *&pa);

    /// The integrand for the energy density for degenerate fermions
    double deg_energy_fun(double u, void *&pa);

    /// The integrand for the entropy density for degenerate fermions
    double deg_entropy_fun(double u, void *&pa);

    /// Solve for the chemical potential given the density
    int solve_fun(double x, double &yy, void *&pa);
    int pair_fun(double x, double &yy, void *&pa);
    
#endif

  };

  template<> int io_tlate<rel_fermion>::input
    (cinput *co, in_file_format *ins, rel_fermion *f);
  template<> int io_tlate<rel_fermion>::output
    (coutput *co, out_file_format *outs, rel_fermion *f);
  template<> const char *io_tlate<rel_fermion>::type();
  template<> int io_tlate<rel_fermion>::stat_input
    (cinput *co, in_file_format *ins, rel_fermion *f);
  template<> int io_tlate<rel_fermion>::stat_output
    (coutput *co, out_file_format *outs, rel_fermion *f);

  typedef io_tlate<rel_fermion> rel_fermion_io_type;

#ifndef DOXYGENP
}
#endif

#endif
