/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MROOT_HYBRIDS_H
#define O2SCL_GSL_MROOT_HYBRIDS_H

#include <gsl/gsl_multiroots.h>
#include <gsl/gsl_linalg.h>

#include <o2scl/ovector_tlate.h>
#include <o2scl/gsl_mroot_hybrids_b.h>
#include <o2scl/mroot.h>
#include <o2scl/jacobian.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Multidimensional root-finding algorithm using
      Powell's Hybrid method (GSL)
      
      This is a recasted version of the GSL routines which use a
      modified version of Powell's Hybrid method as implemented in the
      HYBRJ algorithm in MINPACK. Both the scaled and unscaled options
      are available by setting \ref int_scaling (the scaled version is
      the default). If derivatives are not provided, they will be
      computed automatically. This class provides the GSL-like
      interface using allocate(), set() (or set_de() in case where
      derivatives are available), iterate(), and free() and
      higher-level interfaces, msolve() and msolve_de(), which perform
      the solution automatically. Some additional checking is
      performed in case the user calls the functions out of order
      (i.e. set() without allocate()).
      
      The functions msolve() and msolve_de() use the condition \f$
      \sum_i |f_i| < \f$ \ref mroot::tolf to determine if the solver has
      succeeded.
      
      The original GSL algorithm has been modified to shrink the
      stepsize if a proposed step causes the function to return a
      non-zero value. This allows the routine to automatically try to
      avoid regions where the function is not defined. To return to
      the default GSL behavior, set \ref shrink_step to false.
      
      \todo Need to complete generalization to generic matrix
      types
  */
  template<class param_t, class func_t=mm_funct<param_t>, 
    class vec_t=ovector_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc, 
    class jfunc_t=jac_funct<param_t,vec_t,omatrix_view> > 
    class gsl_mroot_hybrids : 
  public mroot<param_t,func_t,vec_t,jfunc_t>, public hybrids_base {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:

    /// Desc
    double enorm_new(const vec_t &ff) {
      double e2 = 0 ;
      size_t i;
      for (i = 0; i < dim ; i++) {
	double fi= ff[i];
	e2 += fi * fi ;
      }
      return sqrt(e2);
    }

    /// Desc
    int compute_trial_step_new(gsl_vector *xl, gsl_vector *dxl, 
			       vec_t &x_trial) {
      size_t i, N = xl->size;
      
      for (i = 0; i < N; i++) {
	double pi = gsl_vector_get (dxl, i);
	double xi = gsl_vector_get (xl, i);
	x_trial[i]=xi+pi;
      }
      return 0;
    }

    /// Desc
    int compute_df_new(const vec_t &f_trial, 
		       const gsl_vector * fl, gsl_vector * dfl) {
      size_t i, n = fl->size;
      
      for (i = 0; i < n; i++) {
	double dfi =f_trial[i] - gsl_vector_get (fl, i);
	gsl_vector_set (dfl, i, dfi);
      }

      return 0;
    }

    /// Pointer to the user-specified Jacobian object
    jfunc_t *jac;
    
    /// Pointer to the automatic Jacobian object
    jacobian<param_t,func_t,vec_t,omatrix_view> *ajac;

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// A structure for \ref gsl_mroot_hybrids
    typedef struct
    {
      size_t iter;
      size_t ncfail;
      size_t ncsuc;
      size_t nslow1;
      size_t nslow2;
      double fnorm;
      double delta;
      gsl_matrix *J;
      gsl_matrix *q;
      gsl_matrix *r;
      gsl_vector *tau;
      gsl_vector *diag;
      gsl_vector *qtf;
      gsl_vector *newton;
      gsl_vector *gradient;
      gsl_vector *df;
      gsl_vector *qtdf;
      gsl_vector *rdx;
      gsl_vector *w;
      gsl_vector *v;
    }
    o2scl_hybrid_state_t;

    /// The present solution
    gsl_vector *x;

    /// The value of the derivative
    gsl_vector *dx;

    /// The solver state
    o2scl_hybrid_state_t *state;

    /// The function parameters
    param_t *params;

    /// The number of equations and unknowns
    size_t dim;

    /// True if the jacobian has been given
    bool jac_given;

    /// Pointer to the user-specified function
    func_t *fnewp;

    /// True if "set" has been called
    bool set_called;

    /// Finish the solution after set() or set_de() has been called
    int solve_set(size_t nn, vec_t &xx, param_t &pa, func_t &ufunc) {
      int iter=0, status;

      do {
	iter++;
	
	status=iterate(xx);
	if (status) break;

	// ------------------------------------------------------
	// The effective equivalent of the statement:
	// 
	// status=gsl_multiroot_test_residual(f,this->tolf);

	double resid=0.0;
	for(size_t i=0;i<nn;i++) {
	  resid+=fabs(gsl_vector_get(f,i));
	}
	if (resid<this->tolf) status=gsl_success;
	else status=gsl_continue;
	
	// ------------------------------------------------------
	
	if (this->verbose>0) {
	  ovector *rx=(ovector *)x;
	  ovector *rf=(ovector *)f;
	  this->print_iter(nn,*rx,*rf,iter,resid,this->tolf,
			   "gsl_mroot_hybrids");
	}
	
      } while (status == GSL_CONTINUE && iter < this->ntrial);
      
      for(size_t i=0;i<nn;i++) {
	xx[i]=gsl_vector_get(x,i);
      }

      this->last_ntrial=iter;

      if (iter>=this->ntrial) {
	set_err_ret("Function msolve() exceeded max number of iterations.",
		    o2scl::gsl_emaxiter);
      }
      
      if (status!=0) {
	add_err_ret("Function iterate() failed in msolve().",
		    gsl_efailed);
      }

      return gsl_success;
    }

#endif

    public:
      
    gsl_mroot_hybrids() {
      shrink_step=true;
      dim=0;
      ajac=&def_jac;
      def_jac.epsrel=GSL_SQRT_DBL_EPSILON;
      int_scaling=true;
      jac_given=false;
      set_called=false;
    }

    virtual ~gsl_mroot_hybrids() {}

    /** \brief If true, iterate() will shrink the step-size automatically if
	the function returns a non-zero value (default true)

	The original GSL behavior can be obtained by setting 
	this to \c false.
    */
    bool shrink_step;

    /// If true, use the internal scaling method (default true)
    bool int_scaling;

    /// Default automatic Jacobian object
    simple_jacobian<param_t,func_t,vec_t,omatrix_view,
    alloc_vec_t,alloc_t> def_jac;

    /// Set the automatic Jacobian object
    virtual int set_jacobian(jacobian<param_t,func_t,vec_t,omatrix_view> &j) {
      ajac=&j;
      return 0;
    }

    /** 
	\brief The value of the function at the present iteration
	
	\comment
	We need this to be public so that the user can see if 
	iterate() is working
	\endcomment
	
    */
    gsl_vector *f;

    /** 
	\brief Perform an iteration
	
	At the end of the iteration, the current value of the solution 
	is stored in \c ux.
    */
    int iterate(vec_t &ux) {
	
      if (!set_called) {
	std::string s2="Function set() not called or most recent call failed";
	s2+="in gsl_mroot_hybrids::iterate().";
	set_err_ret(s2.c_str(),gsl_efailed);
      }

      const double fnorm=state->fnorm;
 
      gsl_matrix *J=state->J;
      gsl_matrix *q=state->q;
      gsl_matrix *r=state->r;
      gsl_vector *tau=state->tau;
      gsl_vector *diag=state->diag;
      gsl_vector *qtf=state->qtf;
      gsl_vector *df=state->df;
      gsl_vector *qtdf=state->qtdf;
      gsl_vector *rdx=state->rdx;
      gsl_vector *w=state->w;
      gsl_vector *v=state->v;
 
      double prered,actred;
      double pnorm,fnorm1,fnorm1p;
      double ratio;
      double p1=0.1,p5=0.5,p001=0.001,p0001=0.0001;
      
      /* Compute qtf=Q^T f */
      
      compute_qtf(q,f,qtf);
  
      /* Compute dogleg step */
      
      dogleg(r,qtf,diag,state->delta,state->newton,state->gradient,dx);
      
      /* Take a trial step */
      
      alloc_vec_t x_trial,f_trial;
      ao.allocate(x_trial,dim);
      ao.allocate(f_trial,dim);

      compute_trial_step_new(x,dx,x_trial);
      
      pnorm=scaled_enorm(diag,dx);
      
      if (state->iter == 1) {
	if (pnorm < state->delta) {
	  state->delta=pnorm;
	}
      } 

      /* Evaluate function at x + p */
      
      int status;

      if (shrink_step==false) {
	  
	// Evaluate the function at the new point, exit if it fails
	  
	status=(*fnewp)(dim,x_trial,f_trial,*params);
	  
	if (status != gsl_success) {
	  ao.free(x_trial);
	  ao.free(f_trial);
	  add_err_ret
	    ("Returned non-zero value in gsl_mroot_hybrids::iterate().",
	     o2scl::gsl_ebadfunc);
	}
    
      } else {
	  
	// Evaluate the function at the new point, try to recover
	// if it fails
	  
	status=(*fnewp)(dim,x_trial,f_trial,*params);
	  
	int bit=0;
	while(status!=0 && bit<20) {
	  for(size_t ib=0;ib<dim;ib++) {
	    x_trial[ib]=(x_trial[ib]+gsl_vector_get(x,ib))/2.0;
	  }
	  status=(*fnewp)(dim,x_trial,f_trial,*params);
	  bit++;
	}
    
	// Exit if we weren't able to find a new good point

	if (status != gsl_success) {
	  ao.free(x_trial);
	  ao.free(f_trial);
	  add_err_ret
	  ("No suitable step found in gsl_mroot_hybrids::iterate().",
	   o2scl::gsl_ebadfunc);
	}
    
      }

      /* Set df=f_trial-f */
      
      compute_df_new(f_trial,f,df);
      
      /* Compute the scaled actual reduction */
  
      fnorm1=enorm_new(f_trial);
  
      actred=compute_actual_reduction(fnorm,fnorm1);
  
      /* Compute rdx=R dx */
  
      compute_rdx(r,dx,rdx);
  
      /* Compute the scaled predicted reduction phi1p=|Q^T f + R dx| */
  
      fnorm1p=enorm_sum(qtf,rdx);
      
      prered=compute_predicted_reduction(fnorm,fnorm1p);
      
      /* Compute the ratio of the actual to predicted reduction */
  
      if (prered > 0) {
	ratio=actred / prered;
      } else {
	ratio=0;
      }
  
      /* Update the step bound */
  
      if (ratio < p1)	{
	state->ncsuc=0;
	state->ncfail++;
	state->delta *= p5;
      } else {
	state->ncfail=0;
	state->ncsuc++;
      
	if (ratio >= p5 || state->ncsuc > 1) {
	  state->delta=GSL_MAX (state->delta,pnorm / p5);
	}
	if (fabs (ratio - 1) <= p1) {
	  state->delta=pnorm / p5;
	}
      }
  
      /* Test for successful iteration */
  
      if (ratio >= p0001) {
	for(size_t i=0;i<dim;i++) {
	  gsl_vector_set(x,i,x_trial[i]);
	  gsl_vector_set(f,i,f_trial[i]);
	}
	state->fnorm=fnorm1;
	state->iter++;
      }
      
      ao.free(x_trial);
      ao.free(f_trial);

      /* Determine the progress of the iteration */
  
      state->nslow1++;
      if (actred >= p001) {
	state->nslow1=0;
      }
  
      if (actred >= p1) {
	state->nslow2=0;
      }
    
      if (state->ncfail == 2) {

	int rx;
	  
	alloc_vec_t bx;
	alloc_vec_t bf;
	ao.allocate(bx,dim);
	ao.allocate(bf,dim);
	for(size_t ikx=0;ikx<dim;ikx++) {
	  bx[ikx]=gsl_vector_get(x,ikx);
	  bf[ikx]=gsl_vector_get(f,ikx);
	}
	omatrix_view *oJ2=(omatrix_view *)J;
	
	if (jac_given) rx=(*jac)(dim,bx,bf,*oJ2,*params);
	else rx=(*ajac)(dim,bx,bf,*oJ2,*params);

	ao.free(bx);
	ao.free(bf);

	if (rx!=0) {
	  add_err_ret("Jacobian failed in iterate().",gsl_efailed);
	}
      
	state->nslow2++;
      
	if (state->iter == 1) {
	  if (int_scaling) {
	    compute_diag (J,diag);
	  }
	  state->delta=compute_delta (diag, x);
	} else {
	  if (int_scaling) {
	    update_diag (J,diag);
	  }
	}
      
	/* Factorize J into QR decomposition */
      
	gsl_linalg_QR_decomp (J,tau);
	gsl_linalg_QR_unpack (J,tau,q,r);
	
	return gsl_success;
      }
  
      /* Compute qtdf=Q^T df,w=(Q^T df - R dx)/|dx|, v=D^2 dx/|dx| */
  
      compute_qtf(q,df,qtdf);
  
      compute_wv(qtdf,rdx,dx,diag,pnorm,w,v);
  
      /* Rank-1 update of the jacobian Q'R'=Q(R + w v^T) */
  
      gsl_linalg_QR_update(q,r,w,v);
  
      /* No progress as measured by jacobian evaluations */

      if (state->nslow2 == 5) {
	set_err_ret("No progress in Jacobian in iterate().",
		    gsl_enoprogj);
      }
  
      /* No progress as measured by function evaluations */

      if (state->nslow1 == 10) {
	set_err_ret("No progress in function in iterate().",
		    gsl_enoprog);
      }

      for(size_t i=0;i<dim;i++) ux[i]=gsl_vector_get(x,i);
  
      return gsl_success;
    }
      
    /// Allocate the memory
    int allocate(size_t n) {
      int status;

      if (dim>0) free();
  
      x=gsl_vector_calloc(n);
      if (x == 0) {
	set_err_ret("failed to allocate space for x",gsl_enomem);
      }
      
      f=gsl_vector_calloc(n);
      if (f == 0) {
	gsl_vector_free(x);
	set_err_ret("failed to allocate space for f",gsl_enomem);
      }
      dx=gsl_vector_calloc(n);
      
      if (dx == 0) {
	gsl_vector_free(x);
	gsl_vector_free(f);
	set_err_ret("failed to allocate space for dx",gsl_enomem);
      }
  
      state=(o2scl_hybrid_state_t *)
      malloc(gsl_multiroot_fsolver_hybrids->size);
  
      if (state == 0) {
	gsl_vector_free(dx);
	gsl_vector_free(x);
	gsl_vector_free(f);
	set_err_ret("failed to allocate space for multiroot solver state",
		    o2scl::gsl_enomem);
      }
	
      status=(gsl_multiroot_fsolver_hybrids->alloc)(state,n);
	
      if (status != gsl_success) {
	(gsl_multiroot_fsolver_hybrids->free)(state);
	std::free(state);
	gsl_vector_free(dx);
	gsl_vector_free(x);
	gsl_vector_free(f);
	set_err_ret("Solver allocation failed.",status);
      }
      
      dim=n;

      return o2scl::gsl_success;
    }
    
    /// Free the allocated memory
    int free() {
      if (dim>0) {
	(gsl_multiroot_fsolver_hybrids->free)(state);
	std::free(state);
	gsl_vector_free(dx);
	gsl_vector_free(x);
	gsl_vector_free(f);
	dim=0;
      }
      return o2scl::gsl_success;
    }
      
    /// Return the type,\c "gsl_mroot_hybrids".
    virtual const char *type() { return "gsl_mroot_hybrids"; }
    
    /** \brief Solve \c func with derivatives \c dfunc using \c x as 
	an initial guess,returning \c x.

    */
    virtual int msolve_de(size_t nn, vec_t &xx, param_t &pa, func_t &ufunc,
			  jfunc_t &dfunc) {
      int status;
  
      status=set_de(nn,xx,ufunc,dfunc,pa);
      if (status!=0) {
	add_err_ret("Function set() failed in msolve().",gsl_efailed);
      }
      
      return solve_set(nn,xx,pa,ufunc);
    }

    /// Solve \c ufunc using \c xx as an initial guess,returning \c xx.
    virtual int msolve(size_t nn, vec_t &xx, param_t &pa, func_t &ufunc) {
      
      int status;

      status=set(nn,xx,ufunc,pa);
      if (status!=0) {
	add_err_ret("Function set() failed in msolve().",gsl_efailed);
      }
      
      return solve_set(nn,xx,pa,ufunc);
    }
      
    /** \brief Set the function, the parameters, and the initial guess 
     */
    int set(size_t nn, vec_t &ax, func_t &ufunc, param_t &pa) {

      int status;
  
      if (nn!=dim) { 
	status=allocate(nn);
	if (status!=0) {
	  add_err_ret("Function allocate() failed in msolve().",gsl_efailed);
	}
      }
      
      fnewp=&ufunc;

      // Specify function for automatic Jacobian
      ajac->set_function(ufunc);
      
      if (dim==0) {
	set_err_ret("Memory has not been allocated. Use allocate().",
		    o2scl::gsl_ebadlen);
      }
      
      params=&pa;
  
      // We can't use gsl_vector_memcpy here,since ax may have a
      // different size than dim.
      for(size_t i=0;i<dim;i++) gsl_vector_set(x,i,ax[i]);
	
      o2scl_hybrid_state_t *hstate=(o2scl_hybrid_state_t *)state;
	
      gsl_matrix *J=state->J;
      gsl_matrix *q=state->q;
      gsl_matrix *r=state->r;
      gsl_vector *tau=state->tau;
      gsl_vector *diag=state->diag;
 
      alloc_vec_t aaf;
      ao.allocate(aaf,dim);
      for(size_t ikx=0;ikx<dim;ikx++) {
	aaf[ikx]=gsl_vector_get(f,ikx);
      }
      omatrix_view *oJ=(omatrix_view *)J;

      status=ufunc(dim,ax,aaf,*params);
      if (status!=0) {
	set_err_ret("Function returned non-zero in set().",gsl_ebadfunc);
      }

      if (jac_given) status=(*jac)(dim,ax,aaf,*oJ,*params);
      else status=(*ajac)(dim,ax,aaf,*oJ,*params);
      if (status!=0) {
	add_err_ret("Jacobian failed in set().",gsl_efailed);
      }

      for(size_t ikx=0;ikx<dim;ikx++) {
	gsl_vector_set(f,ikx,aaf[ikx]);
      }
      ao.free(aaf);

      state->iter=1;
      state->fnorm=enorm (f);
      state->ncfail=0;
      state->ncsuc=0;
      state->nslow1=0;
      state->nslow2=0;
  
      gsl_vector_set_all(dx,0.0);
  
      /* Store column norms in diag */
  
      if (int_scaling) {
	compute_diag(J,diag);
      } else {
	gsl_vector_set_all(diag,1.0);
      }
	
      /* Set delta to factor |D x| or to factor if |D x| is zero */
	
      state->delta=compute_delta(diag,x);
  
      /* Factorize J into QR decomposition */
  
      status=gsl_linalg_QR_decomp(J,tau);
      if (status!=0) {
	set_err_ret("QR decomposition failed in set().",gsl_efailed);
      }
      
      status=gsl_linalg_QR_unpack(J,tau,q,r);
      if (status!=0) {
	set_err_ret("QR unpacking failed in set().",gsl_efailed);
      }
      
      set_called=true;
      jac_given=false;

      return gsl_success;
    }

    /** \brief Set the function, the Jacobian, the parameters,
	and the initial guess 
    */
    int set_de(size_t nn, vec_t &ax, func_t &ufunc, jfunc_t &dfunc, 
	       param_t &pa) {

      fnewp=&ufunc;
      jac=&dfunc;

      /// Make sure set() uses the right Jacobian
      jac_given=true;
      
      int ret=set(nn,ax,ufunc,pa);

      /// Reset jac_given since set() will set it back to false
      jac_given=true;
      set_called=true;

      return ret;
    }
    
  };
  
#ifndef DOXYGENP
}
#endif

#endif
