/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <cmath>
#include <o2scl/test_mgr.h>
#include <o2scl/mm_funct.h>
#include <o2scl/gsl_mroot_hybrids.h>

int gfn(size_t nv, const o2scl::ovector_view &x, 
	o2scl::ovector_view &y, void *&pa) {
  y[0]=sin(x[1]-0.2);
  y[1]=sin(x[0]-0.25);
  return 0;
}

class cl {
public:
  int mfn(size_t nv, const o2scl::ovector_view &x, o2scl::ovector_view &y, 
	  void *&pa) {
    y[0]=sin(x[1]-0.2);
    y[1]=sin(x[0]-0.25);
    return 0;
  }
  int mfnd(size_t nv, o2scl::ovector_view &x, o2scl::ovector_view &y,
	   o2scl::omatrix_view &j, void *&pa) {
    j[0][0]=0.0;
    j[0][1]=cos(x[1]-0.2);
    j[1][0]=cos(x[0]-0.25);
    j[1][1]=0.0;
    return 0;
  }

  int mfna(size_t nv, const double x[2], double y[2], void *&pa) {
    y[0]=sin(x[1]-0.2);
    y[1]=sin(x[0]-0.25);
    return 0;
  }
};

using namespace std;
using namespace o2scl;

int main(void) {
  cl acl;
  ovector x(2);
  double xa[2];
  int i;
  void *vp=NULL;
  size_t tmp;
  int r1, r2, r3;
  bool done;
  test_mgr t;

  t.set_output_level(2);

  // 1 - Normal execution using a member function
  mm_funct_mfptr<cl,void *,ovector_view> fmf(&acl,&cl::mfn);
  gsl_mroot_hybrids<void *> cr1;
    
  x[0]=0.5;
  x[1]=0.5;
  cr1.msolve(2,x,vp,fmf);
  t.test_rel(x[0],0.25,1.0e-6,"1a");
  t.test_rel(x[1],0.2,1.0e-6,"1b");

  // 2 - Using the set(), iterate() interface
  gsl_mroot_hybrids<void *> cr1b;
  
  x[0]=0.5;
  x[1]=0.5;
  cr1b.allocate(2);
  cr1b.set(2,x,fmf,vp);
  done=false;
  do {
    r3=cr1b.iterate(x);
    double resid=fabs(gsl_vector_get(cr1b.f,0))+
      fabs(gsl_vector_get(cr1b.f,1));
    if (resid<cr1b.tolf || r3>0) done=true;
  } while (done==false);
  cr1b.free();
  t.test_rel(x[0],0.25,1.0e-6,"2a");
  t.test_rel(x[1],0.2,1.0e-6,"2b");

  // 3 - Having specified the Jacobian
  jac_funct_mfptr<cl,void *,ovector_view,omatrix_view> fmfd(&acl,&cl::mfnd);
    
  x[0]=0.5;
  x[1]=0.5;
  cr1.msolve_de(2,x,vp,fmf,fmfd);
  t.test_rel(x[0],0.25,1.0e-6,"3a");
  t.test_rel(x[1],0.2,1.0e-6,"3b");

  // 4 - Using the set_de(), iterate() interface
  gsl_mroot_hybrids<void *> cr1c;

  x[0]=0.5;
  x[1]=0.5;
  cr1c.allocate(2);
  cr1c.set_de(2,x,fmf,fmfd,vp);
  done=false;
  do {
    r3=cr1c.iterate(x);
    double resid=fabs(gsl_vector_get(cr1c.f,0))+
      fabs(gsl_vector_get(cr1c.f,1));
    if (resid<cr1c.tolf || r3>0) done=true;
  } while (done==false);
  cr1c.free();
  t.test_rel(x[0],0.25,1.0e-6,"4a");
  t.test_rel(x[1],0.2,1.0e-6,"4b");

  // 5 - Using arrays instead of ovectors
  mm_vfunct_mfptr<cl,void *,2> fmf2(&acl,&cl::mfna);
  gsl_mroot_hybrids<void *,mm_vfunct_mfptr<cl,void *,2>,double[2],
    double[2],array_alloc<double[2]> > cr2;
    
  xa[0]=0.5;
  xa[1]=0.5;
  cr2.msolve(2,xa,vp,fmf2);
  t.test_rel(xa[0],0.25,1.0e-6,"5a");
  t.test_rel(xa[1],0.2,1.0e-6,"5b");
  
  // 6 - Using a global function pointer directly
  typedef int (*gfnt)(size_t, const ovector_view &, ovector_view &, 
		      void *&);
  gsl_mroot_hybrids<void *,gfnt,ovector_view> cr4;
  gfnt gfnv=&gfn;

  x[0]=0.5;
  x[1]=0.5;
  cr4.msolve(2,x,vp,gfnv);
  t.test_rel(x[0],0.25,1.0e-6,"6a");
  t.test_rel(x[1],0.2,1.0e-6,"6b");
  
  t.report();
  return 0;
}
